/* ========================================================================
   Source File: d6math3d.cpp
   Description: Yet another cheesy 3-vector, 3x3-matrix library
   ======================================================================== */

/* ========================================================================
   Explicit Dependencies
   ======================================================================== */
#include <assert.h>
#include "math3d.h"


//*** We're intentionally converting doubles to floats all over...
#pragma warning ( disable : 4136 )
#pragma warning ( disable : 4244 )

/*----------------------------------------------------------------------------

matrix multiplication.

*/

matrix_3x3 operator*( matrix_3x3 const &Multiplicand,
		matrix_3x3 const &Multiplier )
{
	matrix_3x3 ReturnMatrix;

	for(int i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			real Value = 0;
		  
			for(int k = 0;k < 3;k++)
			{
				Value += Multiplicand.GetElement(i,k) *
						Multiplier.GetElement(k,j);
			}

			ReturnMatrix.SetElement(i,j,Value);
		}
	}

	return ReturnMatrix;
}

vector_3 operator*( matrix_3x3 const &Multiplicand,
		vector_3 const &Multiplier )
{
	vector_3 ReturnPoint;

	for(int i = 0;i < 3;i++)
	{
		real Value = 0;
	  
		for(int k = 0;k < 3;k++)
		{
			Value += Multiplicand.GetElement(i,k) *
					Multiplier.GetElement(k);
		}

		ReturnPoint.SetElement(i,Value);
	}

	return ReturnPoint;
}


/*----------------------------------------------------------------------------

constructors.

*/

matrix_3x3::matrix_3x3( void )
{
	for(int Counter = 0;Counter < 9;Counter++)
	{
		aElements[0][Counter] = 0;
	}

	aElements[0][0] = aElements[1][1] = aElements[2][2] = 1;
}

void matrix_3x3::InitWith( vector_3 const &Vector, real Radians )
{
	real Sin = sin(Radians), Cos = cos(Radians);

	real Vx = Vector.GetX(), Vx2 = Vx * Vx;
	real Vy = Vector.GetY(), Vy2 = Vy * Vy;
	real Vz = Vector.GetZ(), Vz2 = Vz * Vz;

	aElements[0][0] = Vx2 + Cos * (1 - Vx2);
	aElements[0][1] = Vx * Vy * (1 - Cos) - Vz * Sin;
	aElements[0][2] = Vz * Vx * (1 - Cos) + Vy * Sin;
	
	aElements[1][0] = Vx * Vy * (1 - Cos) + Vz * Sin;
	aElements[1][1] = Vy2 + Cos * (1 - Vy2);
	aElements[1][2] = Vy * Vz * (1 - Cos) - Vx * Sin;
	
	aElements[2][0] = Vz * Vx * (1 - Cos) - Vy * Sin;
	aElements[2][1] = Vy * Vz * (1 - Cos) + Vx * Sin;
	aElements[2][2] = Vz2 + Cos * (1 - Vz2);
}

void matrix_3x3::InitWith( vector_3 const &XAxis, vector_3 const &YAxis,
						   vector_3 const &ZAxis )
{
	// put in rotation

	int Counter;
	for(Counter = 0;Counter < 3;Counter++)
	{
		SetElement(0,Counter,XAxis.GetElement(Counter));
		SetElement(1,Counter,YAxis.GetElement(Counter));
		SetElement(2,Counter,ZAxis.GetElement(Counter));
	}
}	
	
/*----------------------------------------------------------------------------

Rotations.

*/

matrix_3x3 &matrix_3x3::ConcatenateXRotation( real Radians )
{
	real Temp10, Temp11, Temp12;
	real Temp20, Temp21, Temp22;

	real Sin = sin(Radians), Cos = cos(Radians);

	Temp10 = aElements[1][0] * Cos + aElements[2][0] * -Sin;
	Temp11 = aElements[1][1] * Cos + aElements[2][1] * -Sin;
	Temp12 = aElements[1][2] * Cos + aElements[2][2] * -Sin;

	Temp20 = aElements[1][0] * Sin + aElements[2][0] * Cos;
	Temp21 = aElements[1][1] * Sin + aElements[2][1] * Cos;
	Temp22 = aElements[1][2] * Sin + aElements[2][2] * Cos;

	aElements[1][0] = Temp10;
	aElements[1][1] = Temp11;
	aElements[1][2] = Temp12;
	aElements[2][0] = Temp20;
	aElements[2][1] = Temp21;
	aElements[2][2] = Temp22;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateYRotation( real Radians )
{
	real Temp00, Temp01, Temp02;
	real Temp20, Temp21, Temp22;

	real Sin = sin(Radians), Cos = cos(Radians);

	Temp00 = aElements[0][0] * Cos + aElements[2][0] * Sin;
	Temp01 = aElements[0][1] * Cos + aElements[2][1] * Sin;
	Temp02 = aElements[0][2] * Cos + aElements[2][2] * Sin;

	Temp20 = aElements[0][0] * -Sin + aElements[2][0] * Cos;
	Temp21 = aElements[0][1] * -Sin + aElements[2][1] * Cos;
	Temp22 = aElements[0][2] * -Sin + aElements[2][2] * Cos;

	aElements[0][0] = Temp00;
	aElements[0][1] = Temp01;
	aElements[0][2] = Temp02;
	aElements[2][0] = Temp20;
	aElements[2][1] = Temp21;
	aElements[2][2] = Temp22;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateZRotation( real Radians )
{
	real Temp00, Temp01, Temp02;
	real Temp10, Temp11, Temp12;

	real Sin = sin(Radians), Cos = cos(Radians);

	Temp00 = aElements[0][0] * Cos + aElements[1][0] * -Sin;
	Temp01 = aElements[0][1] * Cos + aElements[1][1] * -Sin;
	Temp02 = aElements[0][2] * Cos + aElements[1][2] * -Sin;

	Temp10 = aElements[0][0] * Sin + aElements[1][0] * Cos;
	Temp11 = aElements[0][1] * Sin + aElements[1][1] * Cos;
	Temp12 = aElements[0][2] * Sin + aElements[1][2] * Cos;

	aElements[0][0] = Temp00;
	aElements[0][1] = Temp01;
	aElements[0][2] = Temp02;
	aElements[1][0] = Temp10;
	aElements[1][1] = Temp11;
	aElements[1][2] = Temp12;

	return *this;
}

/*----------------------------------------------------------------------------

Scales.

*/

matrix_3x3 &matrix_3x3::ConcatenateXScale( real Distance )
{
	aElements[0][0] = aElements[0][0] * Distance;
	aElements[0][1] = aElements[0][1] * Distance;
	aElements[0][2] = aElements[0][2] * Distance;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateYScale( real Distance )
{
	aElements[1][0] = aElements[1][0] * Distance;
	aElements[1][1] = aElements[1][1] * Distance;
	aElements[1][2] = aElements[1][2] * Distance;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateZScale( real Distance )
{
	aElements[2][0] = aElements[2][0] * Distance;
	aElements[2][1] = aElements[2][1] * Distance;
	aElements[2][2] = aElements[2][2] * Distance;

	return *this;
}

/*----------------------------------------------------------------------------

vector length and normalize.

*/

real vector_3::Length( void ) const
{
	return sqrt(GetX()*GetX() + GetY()*GetY() + GetZ()*GetZ());
}

vector_3 &vector_3::Normalize( void )
{
	real MyLength = Length();

	if(MyLength != 0)
	{
		SetX(GetX() / MyLength);
		SetY(GetY() / MyLength);
		SetZ(GetZ() / MyLength);
	}

	return *this;
}
  
real AngleBetween( vector_3 const &Operand1, vector_3 const &Operand2 )
{
	vector_3 Unit1 = Operand1, Unit2 = Operand2;

	Unit1.Normalize();
	Unit2.Normalize();
	
	return acos(DotProduct(Unit1,Unit2));
}

real AngleBetweenUnit( vector_3 const &Unit1, vector_3 const &Unit2 )
{
	return acos(DotProduct(Unit1,Unit2));
}

void GetDextralCoordinateSystem( vector_3 const &InputX,
				vector_3 &OutputY, vector_3 &OutputZ )
{
	// use algorithm from Shabana, p. 39

	vector_3 Perp(1,1,1);

	real Max = 0;
	int Index = 0;

	for(int i = 0;i < 3;i++)
	{
		real Value = fabs(InputX.GetElement(i));
		if(Value > Max)
		{
			Max = Value;
			Index = i;
		}
	}

	Perp.SetElement(Index,0);

	OutputY = CrossProduct(Perp,InputX).Normalize();
	OutputZ = CrossProduct(InputX,OutputY).Normalize();

	assert(fabs(InputX.Length() - 1) < Epsilon);
	assert(fabs(OutputY.Length() - 1) < Epsilon);
	assert(fabs(OutputZ.Length() - 1) < Epsilon);
	assert(fabs(DotProduct(InputX,OutputY)) < Epsilon);
	assert(fabs(DotProduct(InputX,OutputZ)) < Epsilon);
	assert(fabs(DotProduct(OutputZ,OutputY)) < Epsilon);
}


void OrthonormalizeOrientation( matrix_3x3 &Orientation )
{
	vector_3 X(Orientation(0,0),Orientation(1,0),Orientation(2,0));
	vector_3 Y(Orientation(0,1),Orientation(1,1),Orientation(2,1));
	vector_3 Z;

	X.Normalize();
	Z = CrossProduct(X,Y).Normalize();
	Y = CrossProduct(Z,X).Normalize();

	Orientation(0,0) = X(0); Orientation(0,1) = Y(0); Orientation(0,2) = Z(0);
	Orientation(1,0) = X(1); Orientation(1,1) = Y(1); Orientation(1,2) = Z(1);
	Orientation(2,0) = X(2); Orientation(2,1) = Y(2); Orientation(2,2) = Z(2);
}

real RayIntersectsPlane(vector_3 const &ray_origin,
						vector_3 const &ray_direction,
						vector_3 const &plane_normal,
						real const plane_d,
						real &vd)
{
	real t;

	// Determine the angle of incidence between the ray and the plane
	vd = DotProduct(plane_normal, ray_direction);
	
	if(fabs(vd) > Epsilon)
	{
		// The rey interesects the plane
		// Compute the intersection distance, t
		real vn = DotProduct(plane_normal, ray_origin) + plane_d;
		t = -vn / vd;
	}
	else
	{
		// The ray is parallel, intersects only at infinity
		t = REAL_MAX;
	}

	return(t);
}

bool RayIntersectsBoundingBox(vector_3 const &ray_origin,
							  vector_3 const &ray_direction,
							  vector_3 const &box_origin,
							  matrix_3x3 const &box_orientation,
							  real const width,
							  real const height,
							  real const depth,
							  real &t, vector_3 &normal)
{
	real const dimension_array[3] = {width/2, height/2, depth/2};

	// Begin with the whole line
	real t_near = -REAL_MAX;
	real t_far = REAL_MAX;
	
	for(unsigned int counter = 0; counter < 6; counter++)
	{
		unsigned int dimension_index = counter / 2;
		real sign = (counter % 2) ? 1 : -1;
		real dimension = dimension_array[dimension_index];

		// Determine the plane made by this part of the bounding box
		vector_3 plane_normal(box_orientation.GetElement(0, dimension_index),
							  box_orientation.GetElement(1, dimension_index),
							  box_orientation.GetElement(2, dimension_index));
		plane_normal = plane_normal * sign;
		real plane_d = dimension - DotProduct(box_origin,
											  plane_normal);

		// Intersect the line with the plane
		real vd;
		real t_test = RayIntersectsPlane(ray_origin, ray_direction,
										 plane_normal, plane_d, vd);

		// Clip the line to the plane
		if(vd < 0)
		{
			// Plane is a back-face, affects t_far
			if(t_test < t_far)
			{
				t_far = t_test;
			}
		}
		else
		{
			// Plane is a front-face, affects t_near
			if(t_test > t_near)
			{
				normal = -plane_normal;
				t_near = t_test;
			}
		}
	}

	// If we have any line left, we've intersected
	t = t_near;
	return((t_far - t_near) > Epsilon);
}

/*----------------------------------------------------------------------------

Inverse

@todo both of these inverses suck ass!!!

*/

matrix_3x3 Inverse( matrix_3x3 const &A )
{
	matrix_3x3 Inv;

	real Adet = A(0,0) * (A(1,1)*A(2,2) - A(1,2)*A(2,1)) -
				A(0,1) * (A(1,0)*A(2,2) - A(1,2)*A(2,0)) +
				A(0,2) * (A(1,0)*A(2,1) - A(1,1)*A(2,0));

	assert(Adet != 0);
	
	Inv(0,0) = (A(1,1)*A(2,2) - A(1,2)*A(2,1))/Adet;
	Inv(1,0) = -(A(1,0)*A(2,2) - A(1,2)*A(2,0))/Adet;
	Inv(2,0) = (A(1,0)*A(2,1) - A(1,1)*A(2,0))/Adet;

	Inv(0,1) = -(A(0,1)*A(2,2) - A(0,2)*A(2,1))/Adet;
	Inv(1,1) = (A(0,0)*A(2,2) - A(0,2)*A(2,0))/Adet;
	Inv(2,1) = -(A(0,0)*A(2,1) - A(0,1)*A(2,0))/Adet;

	Inv(0,2) = (A(0,1)*A(1,2) - A(0,2)*A(1,1))/Adet;
	Inv(1,2) = -(A(0,0)*A(1,2) - A(0,2)*A(1,0))/Adet;
	Inv(2,2) = (A(0,0)*A(1,1) - A(0,1)*A(1,0))/Adet;

	return Inv;
}

bool Inverse( matrix_3x3 const &A, matrix_3x3 &Inv )
{
	bool Inverted = false;

	real Adet = A(0,0) * (A(1,1)*A(2,2) - A(1,2)*A(2,1)) -
				A(0,1) * (A(1,0)*A(2,2) - A(1,2)*A(2,0)) +
				A(0,2) * (A(1,0)*A(2,1) - A(1,1)*A(2,0));

	// @todo need to do something with the condition number or norms here
	if(Adet != 0)
	{
		Inv(0,0) = (A(1,1)*A(2,2) - A(1,2)*A(2,1))/Adet;
		Inv(1,0) = -(A(1,0)*A(2,2) - A(1,2)*A(2,0))/Adet;
		Inv(2,0) = (A(1,0)*A(2,1) - A(1,1)*A(2,0))/Adet;
	
		Inv(0,1) = -(A(0,1)*A(2,2) - A(0,2)*A(2,1))/Adet;
		Inv(1,1) = (A(0,0)*A(2,2) - A(0,2)*A(2,0))/Adet;
		Inv(2,1) = -(A(0,0)*A(2,1) - A(0,1)*A(2,0))/Adet;
	
		Inv(0,2) = (A(0,1)*A(1,2) - A(0,2)*A(1,1))/Adet;
		Inv(1,2) = -(A(0,0)*A(1,2) - A(0,2)*A(1,0))/Adet;
		Inv(2,2) = (A(0,0)*A(1,1) - A(0,1)*A(1,0))/Adet;

		Inverted = true;
	}

	return Inverted;
}


//*** Turn on float conversion warning
#pragma warning ( default : 4136 )
#pragma warning ( disable : 4514 ) // unused inline function

