#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "animation.h"
#include "sprite_texture.h"
#include "image_operations.h"

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

void texMake32(int w, int h, uint8 *data, int texture_id,uint WrapMode);

void Animation_Texture_Set::init() {
    sprite_texture_index = 0;
}

void Animation_Mode::init() {
    frames_per_second = 10;
    frame_index_times_256 = 0;
    mode_to_transition_into = NULL;
    mode_is_wrapping_animation = 1;
    my_index = -1;
}

void Animation_Manager::init() {
    num_animation_modes = 0;
    num_texture_sets = 0;
    texture_width = PACKED_TEXTURE_WIDTH;

    my_sprite_texture = CreateSpriteTexture(texture_width, 256);

    uint id;
    glGenTextures(1, &id);
    packed_sprite_texture_handle = id;
    packed_sprite_texture_is_dirty = true;

    Animation_Texture_Set *set;
    Animation_Mode *mode;

    // default dude animation (doom pink pig dude)
    set = load_texture_set_wad_style("sprites/doom95/sprites/sarg", 'A', 4, 8, true, 1, 1);

    mode = &animation_modes[0];
    mode->init();
    mode->texture_set = set;
    num_animation_modes++;

}

void Animation_Manager::update() {
    double now = g_server_time;

    int i;
    for (i = 0; i < num_animation_modes; i++) {
        Animation_Mode *mode = &animation_modes[i];
        if (mode->frames_per_second == 0) continue;

        int packing_index = mode->texture_set->sprite_texture_index;
        int FrameCount = g_animation_manager->my_sprite_texture->Animations[packing_index].FrameCount;
        float frame_number = now * mode->frames_per_second * 1.0;
        float frame_number_times_256 = frame_number * 256;

        int if256 = (int)(frame_number_times_256 * 0.5f);
        
        mode->frame_index_times_256 = if256 % (FrameCount << 8);
    }
}

void Animation_Manager::ensure_texture_is_ready() {
    if (packed_sprite_texture_is_dirty) {
        packed_sprite_texture_is_dirty = false;
        texMake32(texture_width, texture_width, 
                  (unsigned char *)my_sprite_texture->Pixels,
                  packed_sprite_texture_handle,GL_CLAMP_TO_EDGE);

//        DebugDumpSpriteTexture(*my_sprite_texture, "debug_sprite_texture.bmp");
    }
}

Animation_Mode *Animation_Manager::get_next_animation_mode() {
    if (num_animation_modes == MAX_ANIMATION_MODES) return NULL;

    Animation_Mode *result = &animation_modes[num_animation_modes];
    result->init();
    result->texture_set = &texture_sets[0];
    result->my_index = num_animation_modes;

    num_animation_modes++;
    return result;
}

Animation_Texture_Set *Animation_Manager::load_texture_set_wad_style(char *prefix, char letter, int nframes, int nviews, float scale_x, float scale_y, bool loop) {
    if (scale_y == 0) scale_y = scale_x;
    Animation_Texture_Set *set = &texture_sets[num_texture_sets];
    set->sprite_texture_index = my_sprite_texture->AnimationCount;

    AddAnimationToSpriteTexture(*my_sprite_texture, prefix, letter, nframes, nviews, scale_x, scale_y, loop);

    // Here, we assume success.

    packed_sprite_texture_is_dirty = true;

    num_texture_sets++;
    return set;
}

void
gluBuild2DMipmapsThatDontSuckCompleteAss(
    GLenum target,
    GLint components,
    GLint width,
    GLint height,
    GLenum format,
    GLenum type,
    const void *data)
{
    // TODO: We could just call ConvertPixelFormat here
    // if we had a GL texture type lookup table.  But we don't.
    assert(components == 4);
    assert(type == GL_UNSIGNED_BYTE);

    int32 LevelIndex = 0;
    glTexImage2D(GL_TEXTURE_2D, LevelIndex++,
                 4, width, height, 0,
                 GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
    
    void *TempBuffer = malloc(width*height);
    assert(TempBuffer);
    if(TempBuffer)
    {
        int32 destwidth = width / 2;
        int32 destheight = height / 2;
        while((destwidth > 0) && (destheight > 0))        
        {
            ScaleImage(CubicPixelFilter,
                       width, height,
                       width*4, (uint8 *)data,
                       destwidth, destheight,
                       destwidth*4, (uint8 *)TempBuffer);
            glTexImage2D(GL_TEXTURE_2D, LevelIndex++,
                         4, destwidth, destheight, 0,
                         GL_BGRA_EXT, GL_UNSIGNED_BYTE, TempBuffer);
            
            destwidth /= 2;
            destheight /= 2;
        }
    }
}


