#include <stdlib.h>
#include <assert.h>
#include "animation_public.h"
#include "animation.h"
#include "globals.h"
#include "guys.h"
#include "sprite_texture.h"

Animation_Mode *animation_mode_create() {
    return g_animation_manager->get_next_animation_mode();
}

void animation_mode_set_texture(Animation_Mode *mode, 
                                Animation_Texture_Set *set) {
    mode->texture_set = set;
}

void animation_mode_set_speed(Animation_Mode *mode, float frames_per_second) {
    mode->frames_per_second = frames_per_second;
}


// Use this to say which animation mode a particular dude should be in.
void animation_dude_set_mode(Guy *guy, Animation_Mode *mode) {
    static uint32 fast_rand=0;
    if (guy->animation_mode_index == mode->my_index) return;

    guy->animation_mode_index = mode->my_index;
    int k = mode->texture_set->sprite_texture_index;
    int Frames = g_animation_manager->my_sprite_texture->Animations[k].FrameCount;
    guy->animation_timing_offset = (fast_rand % (Frames << 8)) >> 3;
    fast_rand = (fast_rand + 0x5bcf) & 0xffff;
}

int animation_dude_set_mode_oneshot(Guy *guy, Animation_Mode *mode) {
    guy->animation_mode_index = mode->my_index;
    int n = mode->frame_index_times_256;
    int k = mode->texture_set->sprite_texture_index;
    int Frames = g_animation_manager->my_sprite_texture->Animations[k].FrameCount;
    assert(Frames * 256 - n >= 0);
    guy->animation_timing_offset = (Frames*256 - n) >> 3;
    return FLOAT_TIME_TO_INT_TIME(((Frames - 1) / mode->frames_per_second));
}

Animation_Mode *animation_dude_get_mode(Guy *guy) {
    return &g_animation_manager->animation_modes[guy->animation_mode_index];
}

Animation_Texture_Set *animation_texture_set_create_doom_style(char *prefix, char letter, int nframes, int nviews, bool loop, float x_scale, float y_scale) {
    return g_animation_manager->load_texture_set_wad_style(prefix, letter, nframes, nviews, x_scale, y_scale, loop);
}

Animation_Mode *animation_create(
                      char *prefix, char letter,
                      int nframes, int nviews,
                      bool loop,
                      float x_scale, float y_scale)
{
   Animation_Texture_Set *set = animation_texture_set_create_doom_style(prefix, letter, nframes, nviews, loop, x_scale, y_scale);
   if (set == NULL) return NULL;
   Animation_Mode *mode = animation_mode_create();
   animation_mode_set_texture(mode, set);
   return mode;
}
