#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <math.h>

#include "globals.h"
#include "buildings.h"
#include "hf.h"


Building building[MAX_BUILDINGS];
int num_buildings;

// favor short shifts by using uint8 not uint32
uint8 building_blocked[BUILDING_X*(BUILDING_Y+2)/8 + 2];

#define BUILDING_NUMBER(x,y)   ((y)*BUILDING_X + (x))
#define BUILDING_INDEX(x,y)    ((BUILDING_NUMBER(x,y) >> 3) + (BUILDING_X >> 3) + 1)
#define BUILDING_BIT(x,y)      (1 << (BUILDING_NUMBER(x,y) & 7))

#define BUILDING_TEST(x,y)          (building_blocked[BUILDING_INDEX(x,y)] & BUILDING_BIT(x,y))
#define BUILDING_SET(x,y)           (building_blocked[BUILDING_INDEX(x,y)] |= BUILDING_BIT(x,y))

void makeSolid(int x0, int y0, int x1, int y1)
{
   for (int y=y0; y < y1; ++y)
      for (int x=x0; x < x1; ++x)
         BUILDING_SET(x,y);
}

void buildingCreate(float x0, float y0, float x1, float y1, float h, float r, float g, float b)
{
   int n = num_buildings++;
   int gx0,gy0,gx1,gy1;
   gx0 = floor(x0 / BUILDING_SCALE_X + 0.5);
   gy0 = floor(y0 / BUILDING_SCALE_Y + 0.5);
   gx1 = floor(x1 / BUILDING_SCALE_X + 0.5);
   gy1 = floor(y1 / BUILDING_SCALE_Y + 0.5);

   building[n].x0 = gx0 * BUILDING_SCALE_X + 2;
   building[n].y0 = gy0 * BUILDING_SCALE_Y + 2;
   building[n].x1 = gx1 * BUILDING_SCALE_X - 2;
   building[n].y1 = gy1 * BUILDING_SCALE_Y - 2;
   building[n].h  = h;
   building[n].r  = r;
   building[n].g  = g;
   building[n].b  = b;
   makeSolid(gx0,gy0, gx1,gy1);
}
void buildingCreate(float x0, float y0, float x1, float y1, float h)
{
    buildingCreate(x0,y0,x1,y1,h,GenerateUnitRandomReal(),GenerateUnitRandomReal(),GenerateUnitRandomReal());
}

void buildingsInit(void)
{
   int x_max = (HF_X-1)*HF_SCALE / BUILDING_SCALE_X;
   int y_max = (HF_Y-1)*HF_SCALE / BUILDING_SCALE_Y;
   for(int x=-1; x <= BUILDING_X; ++x) {
      BUILDING_SET(x,-1);
      BUILDING_SET(x,y_max);
   }
   for(int y=-1; y <= BUILDING_Y; ++y) {
      BUILDING_SET(-1,y);
      BUILDING_SET(x_max,y);
   }
}

static void drawBox(float x0, float y0, float x1, float y1, float h, float r, float g, float b)
{
   float base_z = -200.0f;
   glBegin(GL_QUADS);
      glColor3f(r*0.8, g*0.8, b*0.8);
      glVertex3f(x0, y0, base_z);
      glVertex3f(x1, y0, base_z);
      glVertex3f(x1, y0, h);
      glVertex3f(x0, y0, h);

      glVertex3f(x1, y1, base_z);
      glVertex3f(x0, y1, base_z);
      glVertex3f(x0, y1, h);
      glVertex3f(x1, y1, h);


      glColor3f(r*0.9, g*0.9, b*0.9);
      glVertex3f(x1, y0, base_z);
      glVertex3f(x1, y1, base_z);
      glVertex3f(x1, y1, h);
      glVertex3f(x1, y0, h);

      glVertex3f(x0, y1, base_z);
      glVertex3f(x0, y0, base_z);
      glVertex3f(x0, y0, h);
      glVertex3f(x0, y1, h);

      glColor3f(r, g, b);
      glVertex3f(x0, y0, h);
      glVertex3f(x1, y0, h);
      glVertex3f(x1, y1, h);
      glVertex3f(x0, y1, h);
   glEnd();
}

void drawBuilding(int n)
{
   Building *b = &building[n];
   float x0 = b->x0, y0 = b->y0, x1 = b->x1, y1 = b->y1, h = b->h;
   drawBox(x0,y0,x1,y1,h,b->r,b->g,b->b);
}

void drawBuildings( void )
{
   for (int i=0; i < num_buildings; ++i)
      drawBuilding(i);
}    
bool inBuilding(float x, float y)
{
   int gx = FLOAT2INT_BUILDING_X(x);
   int gy = FLOAT2INT_BUILDING_Y(y);
   if (BUILDING_TEST(gx,gy))
      return TRUE;
   return FALSE;
}

