#if !defined(CAMERA_H)
/* ========================================================================
   $RCSfile: camera.h,v $
   $Date: 2002/03/17 08:18:35 $
   $Revision: 1.4.20.2 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#if !defined(GLOBALS_H)
#include "globals.h"
#endif

struct camera
{
    // Aspect ratios
    real32 WpOverHp; // Physical aspect ratio
    real32 WrOverHr; // Resolution aspect ratio
    real32 WwOverHw; // Window aspect ratio

    // View parameters
    real32 FOVY;
    real32 NearClipPlane;
    real32 FarClipPlane;

    // Location parameters
    real32 Orientation[4];
    real32 Position[3];
    real32 EAR[3]; // Elevation/Azimuth/Roll
    real32 Offset[3];

    real32 View4x4[4][4];
    real32 InverseView4x4[4][4];
    real32 Projection4x4[4][4];
    real32 InverseProjection4x4[4][4];
};

void InitializeDefaultCamera(camera &Camera);
void SetCameraAspectRatios(camera &Camera,
                           real32 PhysicalAspectRatio,
                           real32 ScreenWidth, real32 ScreenHeight,
                           real32 WindowWidth, real32 WindowHeight);
void MoveCameraRelative(camera &Camera, real32 X, real32 Y, real32 Z);
void BuildCameraMatrices(camera &Camera);

void GetCameraLocation(camera &Camera, real32 *Result);
void WindowSpaceToWorldSpace(camera &Camera,
                             real32 Width, real32 Height,
                             real32 const *ScreenSpacePoint,
                             real32 *Result);

void GetPickingRay(camera &Camera, real32 Width, real32 Height,
                   real32 MouseX, real32 MouseY,
                   real32 *Origin, real32 *Normal);

#define StandardTVPhysicalAspectRatio 1.33f
#define StandardFilmPhysicalAspectRatio 1.85f

void MouseZoomCamera(camera &Camera, float dZ);
void MouseOrbitCamera(camera &Camera, float dX, float dY);
void MouseRotateCameraRelative(camera &Camera, float dX, float dY);
void MouseMoveCameraRelative(camera &Camera,
                             float dX, float dY, float dZ);
void MouseMoveCameraPlaneRelative(camera &Camera,
                                  float dXPlane, float dYPlane);
void MouseMoveCameraAbsolute(camera &Camera,
                             float dX, float dY, float dZ);

// if you want the camera pos & dir...
extern vec3f	viewpoint_position;
extern vec3f	view_direction;

#define CAMERA_H
#endif
