// derJamminIndieGameenHausen - Irrascible God Bowling

// general configuration

// Tech
//    How do i bounce off buildings?
//
// Game, probably
//    better ball physics - spin in air, figure out slope bug
//    many scoring and so on tweaks
//    find edge of map

#define GAME_SOUNDS
//#define POWER_BAR


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include <math.h>

// engine stuff
#include "globals.h"
#include "sound/sound_public.h"
#include "sprite_texture.h"
#include "memory_utils.h"
#include "value_map.h"
#include "system.h"

#include "camera.h"
#include "app_camera.h"

#include "font.h"

#include "math3d.h"

// game stuff
#include "guys.h"
#include "hf.h"
#include "buildings.h"
#include "animation.h"
#include "animation_public.h"

#include "sound/ds.h" // @todo take this out when debugging is done
#include "sound/sound_manager.h" // @todo take this out when debugging is done
#include "bmp.h"
#include "net.h"

// allocated guy game_data bytes
#define TYPE(g)  (*(uint8*)((g)->game_data+0))  // 0   uint8


void duel_render();
void DuelTick(float time, float dt);
int  DuelChar(int charcode);
void duel_init();

#if 1   // change this to 1 to enable (make sure you change game.cpp as well)
// define system hooks and hook ourselves in
system_hooks SystemHooks = {
    duel_init,
    duel_render,
    DuelTick,
    DuelChar,
};
#endif

void duel_new_game();
void MyPostNetworkMessages( );
void MyProcessNetworkMessage( int player, Input_t input, int x, int y, int z );


void    draw_sprite(int texid, int x0, int y0, int x1, int y1);


struct sprite {
	unsigned int	m_texid;	// opengl texture id
	int		m_width;
	int		m_height;
	float	m_aspect;	// width/height

	sprite() : m_texid(0) {}
	
	void	clear()
	{
		if (m_texid) {
			glDeleteTextures(1, &m_texid);
			m_texid = 0;
		}
	}

	void load(const char* filename)
	{
		clear();

		bmp*        image = ReadBMP(filename, 0xFF);
		assert(image);
		m_texid = texMake32(image->Width, image->Height, (unsigned char*) image->Pixels, GL_CLAMP);
		m_width = image->Width;
		m_height = image->Height;
		assert(m_height > 0);
		m_aspect = float(m_width) / m_height;

		FreeBMP(image);
	}

	void	draw(int x, int y, float scale)
	{
		float	pix_half_width = m_width * scale * 0.5;
		float	pix_half_height = m_height * scale * 0.5;

		draw_sprite(m_texid, x - pix_half_width, y - pix_half_height,
					x + pix_half_width, y + pix_half_height);
	}
};


void	set_speed(Guy* g, float speed)
// scale the guy's velocity.
{
	if (g->vel * g->vel < 0.001) {
		g->vel = vec3f(speed, 0, 0);
	}
	else {
		vec3f_norm(&g->vel, &g->vel);
		g->vel *= speed;
	}
	setSpeed(g, speed);
}


void	set_direction_and_speed(Guy* g, const vec3f& dir, float speed)
// Normalize dir, and enforce the spec'd speed.
{
	vec3f	d = dir;
	d.z = 0;
	if (d * d < 0.001) {
		d = vec3f(1, 0, 0);
	}
	else {
		vec3f_norm(&d, &d);
	}

	setDirectionAndSpeed(g, d.x, d.y, speed);
	setVelocity(g, d.x * speed, d.y * speed, 0);
}



//
// Duel game state
//

struct duel {
    float       m_game_time;
    float       m_mouse_speed_x;
    float       m_mouse_speed_y;
    float       m_mouse_buttons;

    Guy*        m_hunter_guy;
    vec3f       m_hunter_pos;
    float       m_hunter_vel_z;
    float       m_hunter_angle;
	float		m_hunter_odometer;
	int			m_hunter_variant;
	sprite		m_hunter_sprite;

	float		m_ping_timer;

	sprite		m_hunter_pistol;

	sprite		m_muzzle_flash;
	float		m_muzzle_flash_timer;

    Guy*        m_hunted_guy;
	int			m_hunted_variant;
	sprite		m_hunted_sprite;
	float		m_hunted_footstep_timer;
	float		m_hunted_footstep_period;
	float		m_hunted_distance;
	float		m_hunted_angle;	// only on (hunted) client side.

    bool        m_first_person;
    bool        m_player_is_hunter;

	enum state {
		ATTRACT,
		GAME_START,
		PLAYING,
		HUNTER_WINS,
		HUNTER_LOSES,
	} m_game_state;

	float		m_timer;	// generic countdown timer for timed state transitions.

	sprite		m_intro;
	sprite		m_fugitive_label;
	sprite		m_failure_label;
	sprite		m_success_label;
	sprite		m_time_label;

    void        reset()
    // Reset game state.
    {
        m_game_time = 180.f;

        m_hunter_pos.x = 100.f;
        m_hunter_pos.y = 100.f;
        m_hunter_pos.z = 0.0f;
        m_hunter_vel_z = 0;
        m_hunter_angle = 0.f;
		m_hunter_odometer = 0;
		m_muzzle_flash_timer = 0;

        m_first_person = false;
        m_player_is_hunter = true;

        m_hunted_guy = NULL;
		m_hunted_footstep_period = 0;
		m_hunted_footstep_timer = 0;
		m_hunted_angle = 0;

		m_game_state = ATTRACT;
		m_timer = 0;
    }

	vec3f	hunter_direction()
	{
		return vec3f(cosf(m_hunter_angle), sinf(m_hunter_angle), 0);
	}

    duel()
    {
        reset();
    }
} duel;


void	setup_camera();


//=====================================================================================
//	GetMouseInput
//=====================================================================================
static void GetMouseInput(void)
{
	extern HWND WindowHandle;
	if (GetFocus() != WindowHandle) {
		duel.m_mouse_speed_x = 0;
		duel.m_mouse_speed_y = 0;
		return;
	}

	POINT	Point;
	float	dx, dy, cx, cy;
	int32	x, y;

	GetCursorPos(&Point);
  
	x = Point.x;
	y = Point.y;

	cx = (float) WindowWidth / 2.0f;
	cy = (float) WindowHeight / 2.0f;

	dx = ((float)(cx - Point.x) / 350.0f);
	dy = ((float)(cy - Point.y) / 350.0f);

	SetCursorPos((int)cx, (int)cy);

	duel.m_mouse_speed_x = dx;
	duel.m_mouse_speed_y = dy;
}

//=====================================================================================
//	IsKeyDown
//=====================================================================================
static bool IsKeyDown(int KeyCode)
{
	extern HWND WindowHandle;
	if (GetFocus() == WindowHandle)
		if (GetAsyncKeyState(KeyCode) & 0x8000)
			return true;

	return false;
}

//=====================================================================================
//	NewKeyDown
//=====================================================================================
static bool NewKeyDown(int KeyCode)
{
	extern HWND WindowHandle;
	if (GetFocus() == WindowHandle)
		if (GetAsyncKeyState(KeyCode) & 1)
			return true;

	return false;
}




static float hf_old[HF_Y][HF_X];

Animation_Texture_Set *prophet_texture_set;
Animation_Mode *prophet_mode;

#define RANDGUYS 1
#if RANDGUYS
// these all don't fit in the currently-small texture, sadly
struct rand_guys
{
    char *name;
    char anim;
    int frames;
	int	views;
	float	scale;
    Animation_Texture_Set *set;
    Animation_Mode *mode;

} RandGuys[] =
{
    { "sprites/batman/sprites/sarg", 'a', 4, 8, 0.8f },
    { "sprites/duel/sprites/tung", 'a', 4, 8, 0.8f },
    { "sprites/duel/sprites/sung", 'a', 4, 8, 0.8f },
    { "sprites/duel/sprites/sang", 'a', 4, 8, 0.8f },
    { "sprites/fistful/sprites/spos", 'a', 4, 8, 1.2f },
    { "sprites/gold64/sprites/boss", 'a', 4, 8, 1.0f },
    { "sprites/gold64/sprites/head", 'a', 4, 8, 1.2f },
//    { "sprites/gold64/sprites/spos", 'a', 4, 1.2f },

//	{ "sprites/duel/sprites/guy1", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/guy2", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/guy3", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/guy4", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/gal1", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/gal2", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/gal3", 'a', 1, 1, 0.6f },
//	{ "sprites/duel/sprites/gal4", 'a', 1, 1, 0.6f },
};
int const RandGuysCount = (sizeof(RandGuys)/sizeof(RandGuys[0]));

#endif

Animation_Texture_Set *gib_texture_set;
Animation_Mode *gib_mode;

enum GuyModeAI
{
	AI_NORMAL,
	AI_FLEE,
	AI_HUNTER,
	AI_FUGITIVE_PLAYER,
	AI_FUGITIVE_DEAD,
	AI_FUGITIVE_NORMAL,
	AI_FUGITIVE_FREEZE,
	AI_FUGITIVE_FLEE,

   kaiNone,
   kaiFlee,
   kaiToss1,  // while playing death anim
   kaiToss2,  // after playing death anim
   kaiKill,
   kaiFollowFirst,
   kaiFollowLast=kaiFollowFirst+31,
};

vec3f follow_target[kaiFollowLast+1];

///////////////////////
// context
#define INITIAL_GUYS 30000

font *duelFont=NULL;
GLuint ballDisplayList=0;

#define MAX_PICKED 1024
Guy *pickUpList[MAX_PICKED];
int  pickUpIdx=0;

///////////////////////
// scoring
#define SCORE_FILE "//walt/public/igj0/irate_score.txt"
//#define SCORE_FILE "irate_score.txt"
float cur_game_time=0.0f;
int   gBestKills=0;
int   gBallThrows=0;
int   gScoreMode=0;

// game controls
int   gBallLaunching=0;
float gPowerBarMax=0, gPowerBarCur=0, gPowerBarLaunch=0;

///////////////////////
// The Ball Setup

// the balls
#define MAX_BALLS  4
Guy   gBalls[MAX_BALLS];
float gBallsTime[MAX_BALLS];
int   gBallKills[MAX_BALLS];
float gBallSpd[MAX_BALLS];  // hmmm, a struct, perhaps, someday
#define BALL_SPEED    200  // or more?
#define BALL_H        14
#define BALL_V        30
// @todo - for real, have it slow, on zero, explode, end at edge of map
#define BALL_LIFETIME 60.0  // perhaps, really, slow it down, kill on stop?

#define BallActive(gIdx) (gBalls[gIdx].vel.x||gBalls[gIdx].vel.y||gBalls[gIdx].vel.z)

#define BALL_TEXTURE_WIDTH 512
#define BALL_TEXTURE_HEIGHT 256
Sound_Channel *gBallChan[MAX_BALLS];
GLuint gBallTextureHandles[MAX_BALLS];
uint32 *gBallTextures[MAX_BALLS];
bool gBallTextureDirty[MAX_BALLS];
matrix_3x3 gBallOrientation[MAX_BALLS];

sprite_texture *GibTexture = 0;

int gBallPickedup=0;
int gBallPause=0;

// Ball system management
// BallPasteImage pastes an alpha image into the ball texture map wrapping at edges
#define BALLXY(x,y) ((x)%BALL_TEXTURE_WIDTH + ((y)%BALL_TEXTURE_HEIGHT)*BALL_TEXTURE_WIDTH)
void BallPasteImage( int32 SrcX0, int32 SrcY0, int32 SrcX1, int32 SrcY1, int32 SrcStride, uint32 *SrcData,
                     int32 DstX0, int32 DstY0, uint32 *DstData )
{
    for(int sy = SrcY0,dy = DstY0;sy < SrcY1;++sy,++dy)
    {
        for(int sx = SrcX0,dx = DstX0;sx < SrcX1;++sx,++dx)
        {
            uint32 c = SrcData[sx+sy*SrcStride];
            if(c & 0xFF000000)
            {
                DstData[BALLXY(dx,dy)] = c;
            }
        }
    }
}
#undef BALLXY
inline int32 CModSucks( int32 a, int32 b )
{
    assert(b > 0);
    if(a > 0) {
        return a % b;
    } else {
        return b - (-a) % b;
    }
}
void
WetterGib(int32 BallIndex,float xsplat, float ysplat) 
{
    // splat coordinates are center point relative in world space
    gBallTextureDirty[BallIndex] = true;
    
    int width = BALL_TEXTURE_WIDTH;
    int height = BALL_TEXTURE_HEIGHT;
    uint32 *data = gBallTextures[BallIndex];

    sprite_image &Image =
        GibTexture->Animations[0].Frames[
            rand()%GibTexture->Animations[0].FrameCount].Directions[0];

    // find the point on the ball that points to the splat
    vector_3 p = Transpose(gBallOrientation[BallIndex]) *
        vector_3(xsplat/BALL_H,ysplat/BALL_H,-1).Normalize();

    // map to gluSphere lat/long texcoords

    // s = +Y -> 0, -X -> 0.25, -Y -> 0.5, +X -> 0.75
    // rotate +90 degrees (so -pi,pi is +Y ccw), normalize, then make 0,1
    float s = atan2(p(0),-p(1))/(2*RealPi) + 0.5;  

    // t = +Z,-Z -> 1,0 (through a sin)
    float t = asin(p(2))/RealPi + 0.5;

    int x = int(s * width)%width, y = int(t * height)%height;

    // blt the center of the image to the center of the texture point
    int32 SrcWidth = Image.XMax - Image.XMin;
    int32 SrcHeight = Image.YMax - Image.YMin;
    BallPasteImage(Image.XMin, Image.YMin,
                   Image.XMax, Image.YMax,
                   GibTexture->TextureDimension,(uint32 *)GibTexture->Pixels,
                   CModSucks(x-SrcWidth/2,BALL_TEXTURE_WIDTH),
                   CModSucks(y-SrcHeight/2,BALL_TEXTURE_HEIGHT),data);
}

void
BindBallTexture(int32 BallIndex)
{    
    glBindTexture(GL_TEXTURE_2D, gBallTextureHandles[BallIndex]);
    if(gBallTextureDirty[BallIndex])
    {
        int width = BALL_TEXTURE_WIDTH;
        int height = BALL_TEXTURE_HEIGHT;
        uint32 *data = gBallTextures[BallIndex];
        glTexImage2D(GL_TEXTURE_2D, 0,
                     4, width, height, 0,
                     GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);

        gBallTextureDirty[BallIndex] = false;
    }
}

int GetAvailableBall(void)
{
   for (int i=0; i<MAX_BALLS; i++)
      if (!(BallActive(i)))
         return i;
   return -1;
}

int FindOldestBall(void)
{
   int oldest=-1;
   for (int i=0; i<MAX_BALLS; i++)
      if (BallActive(i))
         if ((oldest==-1)||(gBallsTime[i]<gBallsTime[oldest]))
            oldest=i;
   return oldest;
}

void BallDeactivate(int gIdx)
{
   gBalls[gIdx].vel.x=gBalls[gIdx].vel.y=gBalls[gIdx].vel.z=0;
   if (gBallKills[gIdx]>gBestKills)
      gBestKills=gBallKills[gIdx];
   gBallKills[gIdx]=0;
}

// calls for the app
int BallCreate(void)
{
   int ballId=GetAvailableBall();
   if (ballId==-1)
   {
      ballId=FindOldestBall();
      if (gBallChan[ballId])
      {
         gBallChan[ballId]=NULL;
      }
      BallDeactivate(ballId);
   }
   if (ballId!=-1)
   {
#ifdef USE_POWER_BAR
      gBallSpd[ballId]=(BALL_SPEED/3)+(BALL_SPEED*gPowerBarLaunch/4);
#else
      gBallSpd[ballId]=(BALL_SPEED*3/5);
#endif
      Guy *pBall=&gBalls[ballId];
      real32 cam_pos[4];
      GetCameraLocation(*GlobalCamera,cam_pos);
      pBall->loc=*((vec3f*)&cam_pos);
      pBall->physics_mode=PM_GRAVITY;
      vec3f_scale(&pBall->vel,(vec3f*)CameraBackward(),-gBallSpd[ballId]);
      gBallsTime[ballId]=BALL_LIFETIME;
      pBall->ai_timer=INT_MAX_TIMER;
      gBallThrows++;
      // clear out the texture
      memset(gBallTextures[ballId],0,BALL_TEXTURE_WIDTH*BALL_TEXTURE_HEIGHT*sizeof(uint32));
      gBallTextureDirty[ballId] = true;
   }
   return ballId==-1?-1:0;
}

///////////////
// Callbacks and Kills
// global for use in callback - really should pass data, but screw it for now
#define MAX_KILL 1024
int gKillList[MAX_KILL];
int gKillCnt=0;

// should probably change to a parm, but oh well
Guy *pCurBall;

//Animation_Mode *death_mode = NULL, *post_death_mode = NULL;

void DudeKillCBack(Guy *g, float d2, void *parm)
{
   // dont reset the state if it is already in progress
   if (g->ai_mode == kaiToss1 || g->ai_mode == kaiToss2)
      return;

   // compute a new velocity that doesnt suck...
   // well, ok, ask Chris to do it...
   g->ai_mode = kaiToss1;

   g->vel.z=64+(64*frandom());
   g->vel.x=-pCurBall->vel.x;   // should map onto normal of velocity, bias by that
   g->vel.y=-pCurBall->vel.y;
   g->vel.x+=6*floatrandneg();
   g->vel.y+=6*floatrandneg();
   g->vel.x/=3;
   g->vel.y/=3;
   g->physics_mode=PM_GRAVITY;
   TYPE(g) = 1;

   aimGuyAlong(g);

//   g->ai_timer = animation_dude_set_mode_oneshot(g, death_mode);

   // really, should add a more real BallIdx, eh?
   gBallKills[pCurBall-gBalls]++;
   // @todo need to pass the actual splat coordinates here
   float r = BALL_H * 0.8;
   WetterGib(pCurBall - gBalls,floatrandneg()*r,floatrandneg()*r);
}

#define FLEE_BASE 6

// flee - probably should run less, really
void DudeFleeCBack(Guy *g, float d2, void *parm)
{  
   if (g->ai_mode == kaiToss1 || g->ai_mode == kaiToss2)
      return;

// set the guys velocity away from the flee point, proportional to d2
   vec3f delta;
   vec3f_sub(&delta,&g->loc,&pCurBall->loc);
   delta.z=0;  // for now, in 2d no-flooring land
   vec3f_normeq(&delta);
   float scale = (d2<(BALL_H*BALL_H))?16*FLEE_BASE:(d2<(BALL_H*BALL_H*4))?6*FLEE_BASE:4*FLEE_BASE;  // awwwwww yea
   vec3f_scale(&g->vel,&delta,scale);
   g->ai_timer = 1500;
   setSpeed(g, scale);
   aimGuyAlong(g);
   g->ai_mode = kaiFlee;
   TYPE(g)=2;
}

#define FOLLOW_SPEED 6.0
#define FOLLOW_RADIUS 24

void doDeferredKill(Guy *g)
{
   g->ai_mode = kaiKill;
   if (gKillCnt<MAX_KILL)
      gKillList[gKillCnt++]=g-gGuys;  // find the id
}


void	duel_fugitive_flee(Guy* g)
{
	if (duel.m_hunted_guy->ai_mode != AI_FUGITIVE_PLAYER) {
		duel.m_hunted_footstep_period = 0.3f;
		set_direction_and_speed(g, vec3f((rand() % 100) - 50, (rand() % 100) - 50, 0), FOLLOW_SPEED * 8);
		g->ai_timer = 1500;
		g->ai_mode = AI_FUGITIVE_FLEE;
	}
}


void duel_guy_ai_timer(Guy *g)
{
   g->ai_timer=0xffff;

   if (g->ai_mode == AI_NORMAL) {
	   g->ai_mode = AI_NORMAL;
   }
   if (g->ai_mode == AI_FLEE) {
	   // Go back to normal mode.
	   g->ai_mode = AI_NORMAL;
	   set_speed(g, FOLLOW_SPEED);
   }


   // 
   // fugitive states.
   //
   if (g == duel.m_hunted_guy) {
	   g->ai_timer = 1000;	// don't let the fugitive timer get wacky.

	   atof("100");	// debugger hook.

	   if (g->ai_mode == AI_FUGITIVE_DEAD) {
		   duel.m_hunted_footstep_period = 0.0f;
		   duel.m_hunted_footstep_timer = 0.0f;
		   g->ai_timer = 1000;
	   }
	   else if (g->ai_mode == AI_FUGITIVE_NORMAL || g->ai_mode == AI_FUGITIVE_FLEE) {
		   g->ai_mode = AI_FUGITIVE_NORMAL;
		   duel.m_hunted_footstep_period = 0.6f;
		   set_speed(g, FOLLOW_SPEED);
		   g->ai_timer = 1000;
	   }
	   else if (g->ai_mode == AI_FUGITIVE_FREEZE) {
		   // Flee randomly!
		   duel_fugitive_flee(g);
	   }
   }


#if 0
   if (g->ai_mode == kaiToss1)
   {
//      animation_dude_set_mode(g, post_death_mode);
      g->ai_timer=3000;  // millisecond
      g->ai_mode = kaiToss2;
      return;  // tossed guys are deferred death
   }

   if (g->ai_mode == kaiToss2)
   {
      doDeferredKill(g);
      return;  // tossed guys are deferred death
   }

   if (g->ai_mode == kaiFlee)
   {
      randneg_vec3f(&g->vel);
      g->vel.z=0;
      vec3f_scaleeq(&g->vel,FOLLOW_SPEED);
      setSpeed(g, FOLLOW_SPEED);
      g->ai_mode = kaiNone;
      if (GenerateUnitRandomReal() < 0.80) {
         float best_d = 200*200,d;
         for (int i=kaiFollowFirst; i <= kaiFollowLast; ++i) {
            d = sqr(g->loc.x - follow_target[i].x) + sqr(g->loc.y - follow_target[i].y);
            if (d < best_d) {
               g->ai_mode = i;
               best_d = d;
            }
         }
      }
      if (g->ai_mode == kaiNone) {
         aimGuyAlong(g);
      }
   }

   if (g->ai_mode >= kaiFollowFirst && g->ai_mode <= kaiFollowLast) {
      TYPE(g) = 2;
      g->ai_timer = guyAimNear(g, follow_target[g->ai_mode].x, follow_target[g->ai_mode].y, FOLLOW_RADIUS, FOLLOW_SPEED);
      return;
   }
#endif // 0
}


// a guy with gravity hit the ground. return a bounce percentage
float DuelGuyHitTerrain(Guy *g)
{
#if 0
   if (g->vel.z < -60.0) return 0.8;
   if (g->ai_mode==kaiToss1 || g->ai_mode==kaiToss2)  // he's tossed, kick his ass
      doDeferredKill(g);
#endif // 0
   return 0;
}

int cBackKillSort(const void *p1, const void *p2)
{
   int i1=*(int *)p1;
   int i2=*(int *)p2;
   return i1>i2?1:i1==i2?0:-1;
}

void FinishKills(void)
{
   // first, qsort the killist
   // base, num, size, compare
   if (gKillCnt)
   {
      qsort(gKillList,gKillCnt,sizeof(gKillList[0]),cBackKillSort);
      for (; gKillCnt>0; gKillCnt--)
         deleteGuy(gKillList[gKillCnt-1]);
   }
}

bool IsBallRolling( Guy *pCurBall, float *GroundZ )
{
    bool rolling = false;
    float ball_terr_z=ground(pCurBall->loc.x,pCurBall->loc.y);
    if (pCurBall->loc.z<=ball_terr_z+1.0) // -BALL_V/2)
    {
        if(GroundZ)
        {
            *GroundZ = ball_terr_z;
        }
        rolling = true;
    }
    return rolling;
}

///////////////
// Tick

void runGuyHack(Guy *g, int n, int dt)
{
   float cached_dt_float = INT_TIME_TO_FLOAT_TIME(dt);
   float cached_dt_float_2 = cached_dt_float/2;

   while (--n >= 0) {
      float grav = gravity[g->physics_mode];
      float z0 = g->vel.z;
      float dx, dy;
      float orig_h = ground_slope(g->loc.x, g->loc.y, &dx, &dy);

      // @todo: do this right - hack normal based modification in here
      if (g->physics_mode == PM_FLOORED || g->loc.z <= orig_h+0.02)
      {  // adjust velocity here - not sure why it is negative, why i need 0.02 above
         // hmmm, perhaps normalize this
         // float dist = dx*dx + dy*dy;
         g->vel.x -= dx * 100 * cached_dt_float;  // scale?
         g->vel.y -= dy * 100 * cached_dt_float;
      }

      g->vel.z -= grav * cached_dt_float;
      g->loc.x += g->vel.x * cached_dt_float;
      g->loc.y += g->vel.y * cached_dt_float;
      g->loc.z += (g->vel.z+z0) * cached_dt_float_2;
      float h = ground(g->loc.x, g->loc.y);
      if (g->physics_mode == PM_FLOORED)
         g->loc.z = h;
      else if (g->loc.z < h) {
         // for the ball, we dont want the same appGuyHit
         if (g->vel.z<0)
         {
            float r = (g->vel.z < -60.0)?0.42:0;
            g->vel.z = -g->vel.z*r;
         }
         g->loc.z = h;
      }

      // @todo: teach to bounce off top of buildings
      for (int i=0; i < num_buildings; ++i) {
         if (g->loc.x + BALL_H < building[i].x0) continue;
         if (g->loc.x - BALL_H > building[i].x1) continue;
         if (g->loc.y + BALL_H < building[i].y0) continue;
         if (g->loc.y - BALL_H > building[i].y1) continue;
         if (g->loc.z - BALL_H > building[i].h)  continue;

         // where we started the frame
         g->loc.x -= g->vel.x * cached_dt_float;
         g->loc.y -= g->vel.y * cached_dt_float;

         // @WARNING this is really _not_ the right approach
         // i overcomplicated it heavily
         // but it does seem to work, so im gonna fix it later

         // now, how far to each side
         float dxl = building[i].x0 - (g->loc.x + BALL_H);  // left
         float dxr = (g->loc.x - BALL_H) - building[i].x1;  // right
         float dyt = building[i].y0 - (g->loc.y + BALL_H);
         float dyb = (g->loc.y - BALL_H) - building[i].y1;

         int norm_idx=0;
         if (dxl>0) 
            if (dyt>0)
               norm_idx=(dyt>dxl)?0:3;
            else
               norm_idx=(dyb>dxl)?2:3;
         else
            if (dyt>0)
               norm_idx=(dyt>dxr)?0:1;
            else
               norm_idx=(dyb>dxr)?2:1;

         if (norm_idx&1)
            g->vel.x = -g->vel.x;
         else
            g->vel.y = -g->vel.y;

         g->vel.x *= 0.95;
         g->vel.y *= 0.95;

         break;
      }
   }
}


void	duel_game_start()
{
	duel.m_game_state = duel::GAME_START;
	duel.m_timer = 1.0f;
	duel.m_game_time = 180;
	duel.m_ping_timer = 0;
	duel_new_game();

	duel.m_first_person = false;
	setup_camera();
}


void	state_input_tick(float dt)
// Update game state.
{
	duel.m_timer -= dt;
	duel.m_muzzle_flash_timer -= dt;

	// compute and cache distance between players.
	{
		vec3f	delta = duel.m_hunter_pos - duel.m_hunted_guy->loc;
		duel.m_hunted_distance = sqrtf(delta * delta);
	}

	switch (duel.m_game_state) {
	default:
		assert(0);	// invalid game state
		break;

	case duel::ATTRACT:
		// no timeout.  But cycle the intro screen.
		if (duel.m_timer <= 0) {
			duel.m_timer = 20.f;
		}

		// press enter or mouse L to start game.
		if (NewKeyDown(VK_RETURN) || NewKeyDown(VK_LBUTTON)) {
			// only server can start the game.
			if (g_LocalPlayer == 0) {
				int	seed = time(NULL);
				srand(seed);
				AddClientInput(INPUT_GAME_START, seed, rand(), 0);
			}
		}

		break;

	case duel::GAME_START:
		if (duel.m_timer < 0 || NewKeyDown(VK_RETURN) || NewKeyDown(VK_LBUTTON))
		{
			duel.m_game_state = duel::PLAYING;
			duel.m_game_time = 180;

			duel.m_first_person = true;
			setup_camera();
		}
		NewKeyDown(VK_LBUTTON);	// eat mouse button presses.
		break;

	case duel::PLAYING:
		duel.m_game_time -= dt;
		if (duel.m_game_time < 0) {
			// time is up!
			duel.m_game_time = 0;
			duel.m_game_state = duel::HUNTER_LOSES;
			duel.m_timer = 3;
		}
		NewKeyDown(VK_RETURN);	// eat keypress.
		break;

	case duel::HUNTER_WINS:
		if (duel.m_timer < 0 || NewKeyDown(VK_RETURN) || NewKeyDown(VK_LBUTTON)) {
			duel.m_game_state = duel::ATTRACT;
			duel.m_first_person = false;
			setup_camera();
		}
		break;

	case duel::HUNTER_LOSES:
		if (duel.m_timer < 0 || NewKeyDown(VK_RETURN) || NewKeyDown(VK_LBUTTON)) {
			duel.m_game_state = duel::ATTRACT;
			duel.m_first_person = false;
			setup_camera();
		}
		break;
	}

#if 0
	//
	// footstep sound
	//
	if (duel.m_hunted_footstep_timer > 0) {
		duel.m_hunted_footstep_timer -= dt;
		if (duel.m_hunted_footstep_timer < 0) {
			// sound the fugitive footstep.

			// compute ping volume & pan.
			vec3f	delta = duel.m_hunted_guy->loc - viewpoint_position;
			float	distance = sqrtf(delta * delta);

			float	verb_volume;
			if (distance < 20) {
				verb_volume = 1.0f;
			} else {
				verb_volume = (1020.f - distance) / 1000.f;
				if (verb_volume < 0.02) verb_volume = 0.02;
				if (verb_volume > 1) verb_volume = 1;
			}
			verb_volume = 0.55f + verb_volume * 0.45f;

			float	dry_volume;
			if (distance < 20) {
				dry_volume = 1.0f;
			} else {
				dry_volume = 20.0f / distance;
				dry_volume = powf(dry_volume, 0.5f);
			}
//			volume = powf(volume, 2.0f);
//			volume = 0.25f + volume * 0.75f;

			float	pan;
			vec3f	delta_dir = delta;
			vec3f_norm(&delta_dir, &delta_dir);
			vec3f	cross;
			vec3f_cross(&cross, &delta_dir, &view_direction);
			pan = cross.z;

			// round out the response.
			if (pan < 0) {
				pan = -powf(-pan, 2.0f);
			} else {
				pan = powf(pan, 2.0f);
			}

			// Pick a footstep sound at random.
			if (0 /* ENABLE / DISABLE_FOOTSTEP */) {
				int	step = (rand() >> 5) % 4;

				char	buf[1000];

				sprintf(buf, "step%d", step + 1);
				sound_play(buf, dry_volume, pan);

				sprintf(buf, "step%dverb", step + 1);
				sound_play(buf, verb_volume, pan * 0.3);
			}

			duel.m_hunted_footstep_timer = duel.m_hunted_footstep_period;
		}
	}
	else if (duel.m_hunted_footstep_period > 0) {
		duel.m_hunted_footstep_timer = duel.m_hunted_footstep_period;
	}
#endif // 0
}


void hunted_tick(float dt)
{
}


bool	ray_vs_sphere(/*const*/ vec3f& pos, /*const*/ vec3f& ray, /*const*/ vec3f& center, float radius)
// Check for a hit between the spec'd ray and the spec'd sphere.  Only
// returns hits in the positive direction of the ray (no hits behind
// the starting pos).
{
	vec3f	disp;
	vec3f_sub(&disp, &center, &pos);
	
	float	hypotenuse = sqrtf(vec3f_dot(&disp, &disp));
	
	float	base = vec3f_dot(&disp, &ray);
	if (base < 0) {
		// sphere is behind ray.
		return false;
	}

	float	dist_squared = hypotenuse * hypotenuse - base * base;

	if (dist_squared < radius * radius) {
		return true;
	} else {
		return false;
	}
}


void	hunter_shoot_bullet()
// The hunter is shooting his bullet.  Check for collision vs. hunted
// guy.
{
	sound_play("gun_1", 1.0f);

	// show the muzzle flash.
	duel.m_muzzle_flash_timer = 0.2f;
			
	// Shoot a ray from the hunter, and see if it hits the hunted.
	float	radius = 3.0f;
	if (duel.m_hunted_guy->ai_mode == AI_FUGITIVE_PLAYER) {
		// use a larger radius for person vs. person mode.
		radius = 5.0f;
	}

	assert(duel.m_hunted_guy);
	if (ray_vs_sphere(duel.m_hunter_pos, duel.hunter_direction(), duel.m_hunted_guy->loc, 3.0f)) {
		// hit.
		duel.m_game_state = duel::HUNTER_WINS;
		duel.m_timer = 10;
//		set animation state of guy to dying.

		Guy*	g = duel.m_hunted_guy;
		g->ai_mode = AI_FUGITIVE_DEAD;
		duel.m_hunted_footstep_period = 0.0f;
		duel.m_hunted_footstep_timer = 0.0f;
		set_speed(g, 0);

//		victory sound.
	}
	else {
		// no hit.
		duel.m_game_state = duel::HUNTER_LOSES;
		duel.m_timer = 10;
	}
}


const float FLEE_DISTANCE = 30.f;
const float PROX_DISTANCE = 50.f;


void	hunter_prox_callback(Guy* g, float dsquared, void* p)
// Callback on all guys near the hunter.
{
	// Cause guys to flee.

	if (g->ai_mode == AI_FUGITIVE_NORMAL) {
		if (dsquared > FLEE_DISTANCE * FLEE_DISTANCE) {
			// Freeze the fugitive for a moment.
			g->ai_mode = AI_FUGITIVE_FREEZE;
			g->ai_timer = 600 + (rand() % 800);
			duel.m_hunted_footstep_period = 0.f;
			set_speed(g, 0);
		}
		else {
			// run away!
			duel_fugitive_flee(duel.m_hunted_guy);
		}
	}
	else if ((g->ai_mode == AI_NORMAL /*|| g->ai_mode == AI_FLEE*/)
			 && dsquared < FLEE_DISTANCE * FLEE_DISTANCE)
	{
		g->ai_mode = AI_FLEE;
		g->ai_timer = 1000 + (rand() % 1000);
		
		// run away!
		vec3f	delta = g->loc - duel.m_hunter_pos;
		float	dot = delta * duel.hunter_direction();
		if (dot > 0) {
			delta -= duel.hunter_direction() * dot;

			if (delta * delta > 0.001) {
				vec3f_norm(&delta, &delta);
				delta += duel.hunter_direction() * ((rand() % 100) / 200.f);
			}
		}
		set_direction_and_speed(g, delta, FOLLOW_SPEED * 2.0);
	}
}



void	hunter_ping()
{
	// compute distance to hunted.
	duel.m_ping_timer = duel.m_hunted_distance / 1000.f;

	if (duel.m_player_is_hunter) {
		// local ping.
		sound_play("beep1", 0.4f);
	}

	// spook the fugitive.
	if (duel.m_hunted_guy->ai_mode != AI_FUGITIVE_DEAD) {
		duel_fugitive_flee(duel.m_hunted_guy);
	}
}



void	hunted_input_tick(float dt)
{
    if (duel.m_player_is_hunter == false && duel.m_first_person) {
        GetMouseInput();

        //
        // Movement.
        //

		vec3f	original_pos = duel.m_hunted_guy->loc;
		vec3f	pos = original_pos;
		vec3f	dir(cosf(duel.m_hunted_angle), sinf(duel.m_hunted_angle), 0);

		if (duel.m_hunted_guy->ai_mode == AI_FUGITIVE_PLAYER)
		{
			if (IsKeyDown('W')) {
				// forward
				pos += dir * dt * 25.0f;
			}
			if (IsKeyDown('S')) {
				// back
				pos.x -= dir.x * dt * 15.0f;
				pos.y -= dir.y * dt * 15.0f;
			}
			if (IsKeyDown('A')) {
				// left
				pos.x += cosf(duel.m_hunted_angle + Pi32 / 2) * dt * 15.0f;
				pos.y += sinf(duel.m_hunted_angle + Pi32 / 2) * dt * 15.0f;
			}
			if (IsKeyDown('D')) {
				// right
				pos.x += cosf(duel.m_hunted_angle - Pi32 / 2) * dt * 15.0f;
				pos.y += sinf(duel.m_hunted_angle - Pi32 / 2) * dt * 15.0f;
			}

			// Don't penetrate buildings.
			if (inBuilding(pos.x, pos.y)) {
				assert(! inBuilding(original_pos.x, original_pos.y));
				// Cancel any move that enters a building.
				pos = original_pos;
			}

//			duel.m_hunted_guy->ai_mode = AI_FUGITIVE_PLAYER;
			appGuyMove(duel.m_hunted_guy, pos);
			set_direction_and_speed(duel.m_hunted_guy, vec3f(cosf(duel.m_hunted_angle), sinf(duel.m_hunted_angle), 0), 0.01);
		}

#if 0
        // vertical motion.
        duel.m_hunter_vel_z -= 64.f * dt;       // gravity
        duel.m_hunter_pos.z += duel.m_hunter_vel_z * dt;

        bool    on_ground = false;
        float   ground_z = ground(duel.m_hunter_pos.x, duel.m_hunter_pos.y);
        if (duel.m_hunter_pos.z <= ground_z) {
            // hit the ground.
            duel.m_hunter_pos.z = ground_z;
            duel.m_hunter_vel_z = 0;
            on_ground = true;
        }

        // shoot.
        if (duel.m_game_state == duel::PLAYING
			&& IsKeyDown(VK_LBUTTON))
		{
			AddClientInput(INPUT_SHOOT, 0, 0, 0);
        }

		// emit ping.
		if (NewKeyDown(VK_RBUTTON) && duel.m_ping_timer == 0) {
			AddClientInput(INPUT_PING, 0, 0, 0);
		}

		// ping echo.
		if (duel.m_ping_timer > 0) {
			duel.m_ping_timer -= dt;
			if (duel.m_ping_timer <= 0) {
				duel.m_ping_timer = 0;

				// echo has arrived.

				// compute ping volume & pan.
				vec3f	delta = duel.m_hunted_guy->loc - viewpoint_position;
				float	distance = sqrtf(delta * delta);
				
				float	verb_volume;
				if (distance < 20) {
					verb_volume = 1.0f;
				} else {
					verb_volume = (1020.f - distance) / 1000.f;
					if (verb_volume < 0.02) verb_volume = 0.02;
					if (verb_volume > 1) verb_volume = 1;
				}
				float	dry_volume = verb_volume;
				verb_volume = 0.60f + verb_volume * 0.40f;

				float	pan;
				vec3f	delta_dir = delta;
				vec3f_norm(&delta_dir, &delta_dir);
				vec3f	cross;
				vec3f_cross(&cross, &delta_dir, &view_direction);
				pan = cross.z;

				// round out the response.
				if (pan < 0) {
					pan = -powf(-pan, 2.0f);
				} else {
					pan = powf(pan, 2.0f);
				}
				
				sound_play("beep1", dry_volume, pan);
				sound_play("beep1verb", verb_volume * 0.7, pan * 0.3);
			}
		}
#endif // 0

        // mouselook.
        duel.m_hunted_angle += duel.m_mouse_speed_x * 60.0f * dt;
    }
    else {
        // We'll get our hunted info MyProcessNetworkMessage.
    }
}


void	hunter_input_tick(float dt)
{
    if (duel.m_player_is_hunter && duel.m_first_person) {
        GetMouseInput();

        //
        // Movement.
        //

		vec3f	original_pos = duel.m_hunter_pos;

        if (IsKeyDown('W')) {
            // forward
            duel.m_hunter_pos += duel.hunter_direction() * dt * 25.0f;
        }
        if (IsKeyDown('S')) {
            // back
            duel.m_hunter_pos.x -= cosf(duel.m_hunter_angle) * dt * 15.0f;
            duel.m_hunter_pos.y -= sinf(duel.m_hunter_angle) * dt * 15.0f;
        }
        if (IsKeyDown('A')) {
            // left
            duel.m_hunter_pos.x += cosf(duel.m_hunter_angle + Pi32 / 2) * dt * 15.0f;
            duel.m_hunter_pos.y += sinf(duel.m_hunter_angle + Pi32 / 2) * dt * 15.0f;
        }
        if (IsKeyDown('D')) {
            // right
            duel.m_hunter_pos.x += cosf(duel.m_hunter_angle - Pi32 / 2) * dt * 15.0f;
            duel.m_hunter_pos.y += sinf(duel.m_hunter_angle - Pi32 / 2) * dt * 15.0f;
        }

		// Don't penetrate buildings.
		if (inBuilding(duel.m_hunter_pos.x, duel.m_hunter_pos.y)) {
			assert(! inBuilding(original_pos.x, original_pos.y));
			// Cancel any move that enters a building.
			duel.m_hunter_pos = original_pos;
		}

		{
			// Keep track of overall hunter motion, for pistol bob effect.
			vec3f	delta;
			vec3f_sub(&delta, &duel.m_hunter_pos, &original_pos);
			duel.m_hunter_odometer += sqrtf(vec3f_dot(&delta, &delta));
		}

        // vertical motion.
        duel.m_hunter_vel_z -= 64.f * dt;       // gravity
        duel.m_hunter_pos.z += duel.m_hunter_vel_z * dt;

        bool    on_ground = false;
        float   ground_z = ground(duel.m_hunter_pos.x, duel.m_hunter_pos.y);
        if (duel.m_hunter_pos.z <= ground_z) {
            // hit the ground.
            duel.m_hunter_pos.z = ground_z;
            duel.m_hunter_vel_z = 0;
            on_ground = true;
        }

        // shoot.
        if (duel.m_game_state == duel::PLAYING
			&& IsKeyDown(VK_LBUTTON))
		{
//			hunter_shoot_bullet();
			AddClientInput(INPUT_SHOOT, 0, 0, 0);
        }

		// emit ping.
		if (NewKeyDown(VK_RBUTTON) && duel.m_ping_timer == 0) {
//			hunter_ping();
			AddClientInput(INPUT_PING, 0, 0, 0);
		}

        // mouselook.
        duel.m_hunter_angle += duel.m_mouse_speed_x * 60.0f * dt;
    }
    else {
        // We'll get our hunter info MyProcessNetworkMessage.
    }
}


void hunter_tick(float dt)
{
	// Place the hunter guy.
	appGuyMove(duel.m_hunter_guy, duel.m_hunter_pos);
	set_direction_and_speed(duel.m_hunter_guy, vec3f(cosf(duel.m_hunter_angle), sinf(duel.m_hunter_angle), 0), 0.01);

	// Cause people to flee from the hunter.
	findGuysInSphere(duel.m_hunter_pos.x, duel.m_hunter_pos.y, duel.m_hunter_pos.z, PROX_DISTANCE, hunter_prox_callback, NULL);

#if 0    

   static int fleeframe=0;
   static vec3f last_cursor={0,0,0};
   cur_game_time+=dt;

   // more value_map testing
   static value_map *m = MapCreate(256,256,0);
   MapSetPointSample(m,GlobalCamera->Position[0],GlobalCamera->Position[1],255);
//   MapSetTerrainTextureChannels(0,m,0);

#ifdef POWER_BAR
   switch (gBallLaunching)
   {  // exponentiate this crap, non-linear wack, etc...
      case 0: 
         gPowerBarMax+=dt/2;
         if (gPowerBarMax>1)
            gPowerBarMax=1;
         break;
      case 1:
         gPowerBarCur+=dt/2; 
         if (gPowerBarCur>gPowerBarMax)
            gPowerBarCur=gPowerBarMax;  // really, should auto-kick to next, i think
         break;
      case 2:
         gPowerBarCur-=dt; 
         if (gPowerBarCur>0)   // dont drain sub-zero
            gPowerBarMax-=dt; 
         break;
   }
#endif

   // handle pickup
   if (gBallPickedup)
   {
      vec3f delta;
      vec3f_sub(&delta,&playerCursorPos,&last_cursor);
      for (int j=0; j<pickUpIdx; j++)
         vec3f_add(&pickUpList[j]->loc,&pickUpList[j]->loc,&delta);
   }
   last_cursor=playerCursorPos;

   int i;

   // update the follow targets...the leaders (prophets?) are the
   // first kaiFollowLast guys...make sure they look like it
   for (i=0; i <= kaiFollowLast; ++i) {
      follow_target[i] = gGuys[i].loc;
      TYPE(gGuys+i) = 1;
      gGuys[i].collide_flags &= ~128;  // leaders don't flock

      animation_dude_set_mode(&gGuys[i], prophet_mode);
   }

   ++fleeframe;
   for (i=0; i<MAX_BALLS; i++)
      if (!gBallPause&&BallActive(i))
      {
         int rolling=0;
         pCurBall=&gBalls[i];

         // first, are we too old?
         gBallsTime[i]-=dt;
         if (gBallsTime[i]<0)
         {
            BallDeactivate(i);
            continue;
         }

         if (fabs(pCurBall->vel.x)+fabs(pCurBall->vel.y)+fabs(pCurBall->vel.z)<0.1)
         {
            BallDeactivate(i);
            continue;
         }

         // @todo: reaim?
         int idt=FLOAT_TIME_TO_INT_TIME(dt);
         if (idt==0) idt=1;

         // snap location before
         bool pre_rolling = IsBallRolling(pCurBall,0);
         vec3f loc0 = pCurBall->loc; 

         runGuyHack(pCurBall,1,idt);  // dont do binning

         vec3f loc1 = pCurBall->loc; // snap location after

         if (pre_rolling)
            if (loc1.x<0||loc1.x>WORLD_X||loc1.y<0||loc1.y>WORLD_Y)
            {
               BallDeactivate(i);
               continue;
            }

         float ball_terr_z;
         if(IsBallRolling(pCurBall,&ball_terr_z))
         {
            pCurBall->loc.z=ball_terr_z; // -BALL_V/2;
            rolling=1;
            if(pre_rolling)
            {
                // okay, we rolled this entire frame...theoretically
                // we know Guys travel at constant x,y speed in the plane...ithink
                vector_3 v = vector_3((float*)&loc1) - vector_3((float*)&loc0);
                float vl = v.Length();
                vector_3 axis = CrossProduct(vector_3(0,0,1),v).Normalize();
                float angle = vl/BALL_H;
                if(Length2(axis) > 1e-3)
                {
                    gBallOrientation[i] = matrix_3x3(axis,angle) * gBallOrientation[i];
                }
            }
            else  // track an angular acceleration
            {     // somehow, someday

            }
         }

         pCurBall->ai_timer=0xffff;  // for now, as we hack all this
         
         // check the stuff - look for nearby guys
         if (rolling)
            findGuysInCircle(pCurBall->loc.x,pCurBall->loc.y,BALL_H,DudeKillCBack,NULL);

         vec3f flee_loc;
         vec3f_add_scale(&flee_loc,&pCurBall->loc,&pCurBall->vel,gBallSpd[i]*60);
         findGuysInSphere(pCurBall->loc.x,pCurBall->loc.y,pCurBall->loc.z,BALL_H*3,DudeFleeCBack,NULL);
      }

   FinishKills();

#endif // 0

}



void BallsCollide(void)
{
   
}

void BallDraw(void)
{  // draw the ball
    glMatrixMode(GL_MODELVIEW);
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_NORMALIZE);
    GLfloat const LightPosition[4] = {1.0f, -1.0f, 1.0f, 0.0f};
    glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);
    float aAmbient[4] = { 0.5f, 0.5f, 0.5f, 1.0f };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT,aAmbient);
    float aSpecular[4] = { 1.f, 1.f, 1.f, 1.0f };
    float Shininess = 16.0;
    glMaterialfv(GL_FRONT,GL_SPECULAR,aSpecular);
    glMaterialfv(GL_FRONT,GL_SHININESS,&Shininess);
#define GL_LIGHT_MODEL_COLOR_CONTROL_EXT 0x81F8
#define GL_SINGLE_COLOR_EXT 0x81F9
#define GL_SEPARATE_SPECULAR_COLOR_EXT 0x81FA
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL_EXT,GL_SEPARATE_SPECULAR_COLOR_EXT);
    glEnable(GL_LIGHT0);
    
    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
    glColor3f(1.0f, 1.0f, 1.0f);

    for (int i=0; i<MAX_BALLS; i++)
    if (BallActive(i))
    {
        BindBallTexture(i);
        glPushMatrix();
        float M[16];
        glTranslatef(0, 0, BALL_H);
        CreateOpenGLTransform(gBallOrientation[i],vector_3((float*)&gBalls[i].loc),M);
        glMultMatrixf(M);
        glScalef(BALL_H, BALL_H, BALL_H);
        glCallList(ballDisplayList);
        glPopMatrix();
      }
   
   glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL_EXT,GL_SINGLE_COLOR_EXT);
   glDisable(GL_LIGHTING);
   glDisable(GL_COLOR_MATERIAL);
   glDisable(GL_NORMALIZE);
}

int lastScore=-1;


char  s_name[20][40];
int   s_throws[20], s_best[20];
float s_time[20];
int   s_count[20];
int   s_valid=0;

void BallDisplayScores(void)
{
   char  tmp[255];
   int   i, place=-1;

   if (s_valid==0)
   {
      FILE *fp = fopen(SCORE_FILE,"r");
      if (fp)
      {
         // skip the two comment lines
         fgets(tmp,255,fp);
         fgets(tmp,255,fp);
         for (i=0; i<20; i++)
            if (fscanf(fp,"%f %d %s %d %d",&s_time[i],&s_count[i],s_name[i],&s_throws[i],&s_best[i])!=5)
            {
               s_time[i]=0; s_count[i]=0; strcpy(s_name[i],"RocksMyWorld"); s_throws[i]=0; s_best[i]=0;
            }
         fclose(fp);
         s_valid=1;
      }
   }

   AppSetupFontDraw((WindowWidth/2)-220,WindowHeight-35);
   if (s_valid)
   {
      for (i=0; i<20; i++)
      {
         if (i==10)
            glTranslatef(0.0f,-12.0f, 0.0f);
         sprintf(tmp,"%4.2f %6.6d %20.20s %3.3d %4.4d",s_time[i],s_count[i],s_name[i],s_throws[i],s_best[i]);
         if (lastScore==i)
            glColor3f(1,0,0);
         DrawString(*duelFont,tmp);
         if (lastScore==i)
            glColor3f(0.8f,0.8f,0.8f);
         glTranslatef(0.0f,-23.0f, 0.0f);
      }
   }
   else
   {
      sprintf(tmp,"No Score Data found in %s",SCORE_FILE);
      glColor3f(1,0,0);
      DrawString(*duelFont,tmp);
   }
}

void DrawPowerBar(void)
{
#ifdef POWER_BAR
   // hey, how about a power bar, eh?
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_TEXTURE_2D);
   glBegin(GL_QUADS);

   float bar_max=gPowerBarMax;
   float bar_cur=gPowerBarCur;
   float bar_width=(WindowWidth/20);
   float bar_height=(WindowHeight-90);
   float bar_x=WindowWidth-(bar_width*3/2);
   float bar_y=50;

   float bar_st=(bar_cur<0)?0:bar_cur;

   glColor3f(0.1f,0.1f,0.8f);
   glVertex3f(bar_x,bar_y+(bar_height*bar_st),0);
   glVertex3f(bar_x+bar_width,bar_y+(bar_height*bar_st),0);
   glVertex3f(bar_x+bar_width,bar_y+(bar_height*bar_max),0);
   glVertex3f(bar_x,bar_y+(bar_height*bar_max),0);

   // @todo: red bar near 0
   if (bar_cur>0)
   {
      glColor3f(0.1f,0.8f,0.1f);
      glVertex3f(bar_x,bar_y,0);
      glVertex3f(bar_x+bar_width,bar_y,0);
      glVertex3f(bar_x+bar_width,bar_y+(bar_height*bar_cur),0);
      glVertex3f(bar_x,bar_y+(bar_height*bar_cur),0);
   }
   else if (bar_cur<0)
   {
      glColor3f(0.8f,0.1f,0.1f);
      glVertex3f(bar_x,bar_y+(bar_height*bar_cur),0);
      glVertex3f(bar_x+bar_width,bar_y+(bar_height*bar_cur),0);
      glVertex3f(bar_x+bar_width,bar_y,0);
      glVertex3f(bar_x,bar_y,0);
   }

   glEnd();

   glEnable(GL_TEXTURE_2D);
   glEnable(GL_DEPTH_TEST);
#endif
}

void draw_debug_string(char *s, int sx, int sy) {
      AppSetupFontDraw(sx, sy);
      DrawString(*duelFont, s);
      glTranslatef(-sx, -sy, 0);
}


void duel_text()
// Draw text prompts/messages/etc.
{
   char buf[80];

   if (1 /*duel.m_game_state != duel::ATTRACT */) {
      // Show game time.
      int       minutes = int(duel.m_game_time / 60.0);
      int       seconds = int(duel.m_game_time - minutes * 60);
      sprintf(buf, "%2d:%02d", minutes, seconds);

	  glDisable(GL_BLEND);
	  duel.m_time_label.draw(WindowWidth/2 - 85, WindowHeight - 20, 1);

      AppSetupFontDraw(WindowWidth/2 - 50, WindowHeight - 35);
	  glColor3f(194/255.0f, 170/255.0f, 38/255.0f);
      DrawString(*duelFont, buf);
   }
   if (duel.m_game_state != duel::ATTRACT)
   {
	   if (duel.m_player_is_hunter) {
		   // Label the fugitive
		   glDisable(GL_BLEND);
		   duel.m_fugitive_label.draw(50, WindowHeight - 130, 1);
	   }
	   else {
		   // Label the hunter.
	   }
   }

   float	left_x = WindowWidth/2 - 150;
   float	top_y = WindowHeight/2 + 150;
   switch (duel.m_game_state) {
   case duel::ATTRACT:
//	   AppSetupFontDraw(left_x, top_y);
//	   DrawString(*duelFont, "The Dueling Machine\npress ENTER to play");

	   break;

   case duel::GAME_START:
#if 0
	   AppSetupFontDraw(left_x, top_y);
	   DrawString(*duelFont,
				  "Find the fugitive before time runs out!\n"
				  "You have one bullet only!\n\n"
				  "mouse move  - steer\n"
				  "mouse left  - shoot\n"
				  "mouse right - sonar\n"
				  "   w\n"
				  "a  *  d  - move\n"
				  "   s\n"
		   );
#endif // 0
	   break;

   case duel::PLAYING:
	   break;

   case duel::HUNTER_WINS:
	   if (duel.m_timer < 9.0) {
		   glDisable(GL_BLEND);
		   if (duel.m_player_is_hunter) {
			   duel.m_success_label.draw(WindowWidth/2, WindowHeight/2, 1);
		   } else {
			   duel.m_failure_label.draw(WindowWidth/2, WindowHeight/2, 1);
		   }
	   }
	   break;

   case duel::HUNTER_LOSES:
	   if (duel.m_timer < 9.0) {
		   glDisable(GL_BLEND);
		   if (duel.m_player_is_hunter == false) {
			   duel.m_success_label.draw(WindowWidth/2, WindowHeight/2, 1);
		   } else {
			   duel.m_failure_label.draw(WindowWidth/2, WindowHeight/2, 1);
		   }
	   }
	   break;
   }


#if 0
   if (gScoreMode)
      BallDisplayScores();
   else
   {
      AppSetupFontDraw(WindowWidth-450,WindowHeight-28);
      s_valid=0;
      
      float dGdF=0;
      {
#define SIZE 16
         static int dCall[SIZE];
         static int oldGuy=-1;
         int dFrame=0, total=0;
         if (oldGuy==-1)
            oldGuy=gGuyCount;
         int deltaG=gGuyCount-oldGuy;
         total-=dCall[dFrame];
         total+=deltaG;
         dCall[dFrame]=deltaG;
         dFrame=(dFrame+1)%SIZE;
         dGdF=(float)total/(float)SIZE;
      }

      // undo the font translation
      glTranslatef(-(WindowWidth-450),-(WindowHeight-28), 0);  

      int sx = WindowWidth - 450;
      int sy = WindowHeight - 80;
      int h = 30;

      Sound_Manager *manager = g_sound_manager;
/*
      List debug_strings;
      if (manager) manager->get_debug_strings(&debug_strings);
      char *s;
      Foreach(&debug_strings, s) {
          draw_debug_string(s, sx, sy);
          free(s);
          sy -= h;
      } Endeach;
*/

      DrawPowerBar();
   }
#endif // 0
}

// this is a new game
void duel_guys_init(int guyCnt)
{
	static bool	inited = false;
	if (inited == false) {
		inited = true;
		for(int ii = 0;ii < RandGuysCount;++ii) {
			rand_guys &rg = RandGuys[ii];
			rg.mode = animation_mode_create();
			rg.set = animation_texture_set_create_doom_style(rg.name, rg.anim, rg.frames, rg.views, /*loop*/ true, rg.scale, rg.scale);
			animation_mode_set_texture(rg.mode, rg.set);
		}

		// value_map testing
		MapSetTerrainTexture("data/dueling_machine.psd");
		value_map *HF = MapLoad("data/dueling_machine.psd",3); assert(HF);
		value_map *Density = MapLoad("data/dueling_machine.psd",4);  assert(Density);
		value_map *Positions = MapLoad("data/dueling_machine.psd",5); assert(Positions);
		value_map *Heights = MapLoad("data/dueling_machine.psd",6); assert(Heights);
		MapSetRange(HF,0,128);
		MapSetHeightField(HF);
//    MapSetTerrainTextureChannels(Density,0,0);
		MapSetRange(Heights,100,200);
		MapAllocateBuildings(Positions, Positions/* Heights */);
//    MapAllocateGuyDensity(Density,pGuy,0,MAX_GUYS,true,true);
		MapFree(HF);
		MapFree(Density);
		MapFree(Positions);
		MapFree(Heights);
	}

	setNumGuys(0);

	// template guy.
    Guy guy_template;
    guysClearGuy(&guy_template);
    Guy *pGuy = &guy_template;
    TYPE(pGuy) = 0;
    pGuy->physics_mode = PM_FLOORED;
    pGuy->ai_timer = floatrand() * 1024;
    pGuy->ai_mode = AI_NORMAL;

	assert(RandGuysCount > 1);
	{
		// Pick an animation variant for the hunted guy.
		duel.m_hunted_variant = rand() % RandGuysCount;

		// Make a texture from a representative image of the hunted guy.
		char        hunted_filename[1000];
		strcpy(hunted_filename, RandGuys[duel.m_hunted_variant].name);
		if (RandGuys[duel.m_hunted_variant].views == 1) {
			strcat(hunted_filename, "a0.bmp");
		} else {
			strcat(hunted_filename, "a1.bmp");
		}
		duel.m_hunted_sprite.load(hunted_filename);
	}

	{
		do {
			// Pick an animation variant for the hunter guy.
			duel.m_hunter_variant = rand() % RandGuysCount;
		} while (duel.m_hunter_variant == duel.m_hunted_variant);

		// Make a texture from a representative image of the hunted guy.
		char        hunter_filename[1000];
		strcpy(hunter_filename, RandGuys[duel.m_hunter_variant].name);
		if (RandGuys[duel.m_hunter_variant].views == 1) {
			strcat(hunter_filename, "a0.bmp");
		} else {
			strcat(hunter_filename, "a1.bmp");
		}
		duel.m_hunter_sprite.load(hunter_filename);
	}


    // stuff their velocities randomly
    for(int i = 0; i < MAX_GUYS; ++i) {
        Guy *pGuy = gGuys+i;

		*pGuy = guy_template;

        if (i == 0) {
            // This is the hunted guy.
            duel.m_hunted_guy = pGuy;
			duel.m_hunted_footstep_timer = 0;
			duel.m_hunted_footstep_period = 0.6f;

            animation_dude_set_mode(pGuy, RandGuys[duel.m_hunted_variant].mode);
			if (g_LocalPlayer > 0 || IsPlayerConnected(1))
			{
				// player-controlled fugitive
				pGuy->ai_mode = AI_FUGITIVE_PLAYER;
				pGuy->be_collided_with_flags = 0;
				set_direction_and_speed(pGuy, vec3f(0, 0, 0), 0.01);
			}
			else {
				// normal AI fugitive
				pGuy->ai_mode = AI_FUGITIVE_NORMAL;
				pGuy->be_collided_with_flags = 129;
				float vx = floatrandneg() * 4.0;
				float vy = floatrandneg() * 4.0;
				set_direction_and_speed(pGuy, vec3f(vx, vy, 0), FOLLOW_SPEED);
			}
			pGuy->collide_flags = 0;
        }
		else if (i == 1) {
			// This is the hunter guy.
			duel.m_hunter_guy = pGuy;

			animation_dude_set_mode(pGuy, RandGuys[duel.m_hunter_variant].mode);
			pGuy->ai_mode = AI_HUNTER;
			pGuy->collide_flags = 0;
			pGuy->be_collided_with_flags = 0;

			set_direction_and_speed(pGuy, vec3f(0, 0, 0), 0.01);
		}
		else {
            // Not a hunted guy.  Pick a different variant.
            int variant;
            do {
                variant = rand() % RandGuysCount;
            } while (variant == duel.m_hunted_variant
					 || variant == duel.m_hunter_variant);

            animation_dude_set_mode(pGuy, RandGuys[variant].mode);

			pGuy->collide_flags = 128;   // bit 7 = flocking
			pGuy->be_collided_with_flags = 129;
//			pGuy->collide_flags = 0;
//			pGuy->be_collided_with_flags = 0;
			pGuy->debug_color = (i & 0x0FF);

			float vx = floatrandneg() * 4.0;  // @todo stupid consts?!?!
			float vy = floatrandneg() * 4.0;
			set_direction_and_speed(pGuy, vec3f(vx, vy, 0), FOLLOW_SPEED);
        }

        // Randomly place, but not in buildings.
        do {
            pGuy->loc.x = rand() % WORLD_X;
            pGuy->loc.y = rand() % WORLD_Y;
        } while (inBuilding(pGuy->loc.x, pGuy->loc.y));
        pGuy->loc.z = ground(pGuy->loc.x, pGuy->loc.y);

        pGuy->animation_direction = 0;
        pGuy->num_flock = 0;
    }
    
	// Place the hunter.
	vec3f	delta;
	do {
		duel.m_hunter_pos.x = rand() % WORLD_X;
		duel.m_hunter_pos.y = rand() % WORLD_Y;
		duel.m_hunter_pos.z = ground(duel.m_hunter_pos.x, duel.m_hunter_pos.y);

		delta = duel.m_hunter_pos - duel.m_hunted_guy->loc;
	} while (inBuilding(duel.m_hunter_pos.x, duel.m_hunter_pos.y)
			 || (delta * delta < 500 * 500));

    guysFinishInit(guyCnt);
}


void duel_new_game()
{
	duel_guys_init(INITIAL_GUYS);
}


void PickupCBack(Guy *g, float d2, void *parm)
{
   if (pickUpIdx<MAX_PICKED)
   {
      TYPE(g)=3;
      g->physics_mode=PM_FLY_THROUGH_BUILDINGS;
      g->loc.z+=24.0;
      g->vel.x=g->vel.y=g->vel.z=0;
      g->collide_flags = 0;
      pickUpList[pickUpIdx++]=g;
   }
}

void BallHandSelect(void)
{
   if (gBallPickedup)
   {
      for (int i=0; i<pickUpIdx; i++)
      {
         Guy *g=pickUpList[i];
         TYPE(g)=0;
         g->physics_mode=PM_GRAVITY;
         g->collide_flags = 128;
         aimGuyAlong(g);
      }
      pickUpIdx=0;
   }
   else
      findGuysInSphere(playerCursorPos.x,playerCursorPos.y,playerCursorPos.z,20,PickupCBack,NULL);
   gBallPickedup=!gBallPickedup;
}

// set things up - load sound samples...
void BallInit(void)
{
   gameGuyHitTerrain = DuelGuyHitTerrain;
   gameGuyAITimer = duel_guy_ai_timer;
   guysInit();
   duel_guys_init(INITIAL_GUYS);

   duelFont = LoadFont("data/arial20.font");

   prophet_mode = animation_mode_create();
   prophet_texture_set = animation_texture_set_create_doom_style("sprites/hexen/sprites/icey", 'A', 4, 8, true, 3, 5);
   animation_mode_set_texture(prophet_mode, prophet_texture_set);

//   death_mode = animation_create("sprites/doom95/sprites/sarg", 'I', 6, 1, false); // non-looping
//   animation_mode_set_speed(death_mode, 3.0);
//   post_death_mode = animation_create("sprites/doom95/sprites/sarg", 'N', 1, 1); // non-looping

   // @todo this gib stuff needs to make its way into the sprite_texture
   GibTexture = CreateSpriteTexture(256, 1);
   AddAnimationToSpriteTexture(*GibTexture, "sprites/doom95/sprites/sarg", 'I', 6, 1, 1.0, 1.0, true);
   
   ballDisplayList = glGenLists(1);

   {for(int32 BallIndex = 0;
        BallIndex < MAX_BALLS;
        ++BallIndex)
   {
       int width = BALL_TEXTURE_WIDTH;
       int height = BALL_TEXTURE_HEIGHT;
       void *data = gBallTextures[BallIndex] = (uint32 *)malloc(4 * width * height);
       memset(data, 0x10, 4 * width * height);
       glGenTextures(1, &gBallTextureHandles[BallIndex]);
       glBindTexture(GL_TEXTURE_2D, gBallTextureHandles[BallIndex]);
       assert(gBallTextureHandles);

       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_CLAMP);
       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_CLAMP);
       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

       // @todo: hey, we should subimage correctly update this
       glTexImage2D(GL_TEXTURE_2D, 0,
                    4, width, height, 0,
                    GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
   }}
   
   glNewList(ballDisplayList, GL_COMPILE);
   GLUquadricObj *BallQuadric = gluNewQuadric();
   gluQuadricTexture(BallQuadric, GL_TRUE);
   gluQuadricNormals(BallQuadric, GLU_SMOOTH);
   gluQuadricDrawStyle(BallQuadric, GLU_FILL);
   assert(glGetError() == GL_NO_ERROR);
   gluSphere(BallQuadric, 1.0f, 12, 12);   
   gluDeleteQuadric(BallQuadric);
   glEndList();
}

void BallLaunch(void)
{
#ifdef POWER_BAR
   switch (gBallLaunching)
   {
   case 0: // go from waiting around to first click (ramp up)
      gBallLaunching++;
      break;
   case 1: // start back down, freeze launch power
      gPowerBarLaunch=gPowerBarCur;
      gBallLaunching++;
      break;
   case 2:
      // do the launch - send a power and an accuracy
      BallCreate();
      gBallLaunching=0;
      gPowerBarMax-=gPowerBarLaunch;
      gPowerBarCur=0;
      break;
   }
#else
   BallCreate();
#endif
}
void infect_guys();


//////////
// misc dumb stuff i want somewhere else

void rand_vec3f(vec3f *v)
{
   v->x=floatrand();
   v->y=floatrand();
   v->z=floatrand();
}

void randneg_vec3f(vec3f *v)
{
   v->x=floatrandneg();
   v->y=floatrandneg();
   v->z=floatrandneg();
}

void InfectCallback(Guy *guy, float d2, void *parm) {
    TYPE(guy) = 3;
    guy->debug_color = 1;
    guy->collide_flags |= 1;
}

void infect_guys() {
    float x = playerCursorPos.x;
    float y = playerCursorPos.y;
    float z = playerCursorPos.z;
    findGuysInSphere(x, y, z, 30, InfectCallback, NULL);
}


void    draw_sprite(int texid, int x0, int y0, int x1, int y1)
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0, (real32)WindowWidth, 0, (real32)WindowHeight, -1.0f, 1.0f);

    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();

    glColor3f(1, 1, 1);
    glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, texid);
    glBegin(GL_QUADS);
    {
        glTexCoord2f(0, 0);
        glVertex2f(x0, y0);

        glTexCoord2f(1, 0);
        glVertex2f(x1, y0);

        glTexCoord2f(1, 1);
        glVertex2f(x1, y1);

        glTexCoord2f(0, 1);
        glVertex2f(x0, y1);
    }
    glEnd();

    glMatrixMode(GL_TEXTURE);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
}


void duel_render()
// Render callback, from app.
{
	if (duel.m_game_state == duel::HUNTER_WINS
		|| duel.m_game_state == duel::HUNTER_LOSES
		|| duel.m_game_state == duel::ATTRACT
		|| duel.m_first_person == false)
	{
		{
			// Show an icon above the hunted guy.
			vec3f	pos = duel.m_hunted_guy->loc;
			pos.z += 6.5;

			glDisable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
			glColor3f(1, 1, 0);

			glBegin(GL_TRIANGLES);

			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x - 10, pos.y, pos.z + 10);
			glVertex3f(pos.x + 10, pos.y, pos.z + 10);
			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x + 10, pos.y, pos.z + 10);
			glVertex3f(pos.x - 10, pos.y, pos.z + 10);

			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x, pos.y - 10, pos.z + 10);
			glVertex3f(pos.x, pos.y + 10, pos.z + 10);
			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x, pos.y + 10, pos.z + 10);
			glVertex3f(pos.x, pos.y - 10, pos.z + 10);

			glEnd();
		}

		{
			// Show an icon above the hunter guy.
			vec3f	pos = duel.m_hunter_guy->loc;
			pos.z += 6.5;

			glDisable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
			glColor3f(0.2, 0.2, 1);

			glBegin(GL_TRIANGLES);

			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x - 10, pos.y, pos.z + 10);
			glVertex3f(pos.x + 10, pos.y, pos.z + 10);
			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x + 10, pos.y, pos.z + 10);
			glVertex3f(pos.x - 10, pos.y, pos.z + 10);

			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x, pos.y - 10, pos.z + 10);
			glVertex3f(pos.x, pos.y + 10, pos.z + 10);
			glVertex3f(pos.x, pos.y, pos.z);
			glVertex3f(pos.x, pos.y + 10, pos.z + 10);
			glVertex3f(pos.x, pos.y - 10, pos.z + 10);

			glEnd();
		}
	}

	if (duel.m_hunted_guy->ai_mode == AI_FUGITIVE_DEAD)
	{
		// Draw a red 'X' over the dead guy.
		vec3f	disp;
		vec3f_sub(&disp, &viewpoint_position, &duel.m_hunted_guy->loc);
		disp.z = 0;
		if (vec3f_mag(&disp) > 0.001) {
			// Normalize the flattened guy-to-cam vector.
			vec3f_norm(&disp, &disp);
		} else {
			// fallback.
			disp = view_direction;
//			disp.x = 1;
//			disp.y = 0;
		}

		vec3f	pos = duel.m_hunted_guy->loc + disp;

		vec3f	up(0, 0, 1);
		vec3f	side;
		vec3f_cross(&side, &disp, &up);

		// Draw a cross.

		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
		glColor3f(0.8, 0.1, 0.1);

		glBegin(GL_QUADS);

		glVertex3fv(pos + side * 3);
		glVertex3fv(pos + side * 1.5);
		glVertex3fv((pos - side * 3) + up * 6);
		glVertex3fv((pos - side * 1.5) + up * 6);

		glVertex3fv(pos - side * 1.5);
		glVertex3fv(pos - side * 3);
		glVertex3fv((pos + side * 1.5) + up * 6);
		glVertex3fv((pos + side * 3) + up * 6);

		glEnd();
	}

	if (duel.m_game_state != duel::ATTRACT)
	{
		if (duel.m_player_is_hunter) {
			// Show hunted guy.
			duel.m_hunted_sprite.draw(50, WindowHeight - 50, 100.f / duel.m_hunted_sprite.m_height);
			
			if (duel.m_first_person) {
				// Show pistol
				float	px = WindowWidth / 2 + sinf(duel.m_hunter_odometer * 0.10) * 4;
				float	py = 50 + sinf(duel.m_hunter_odometer * 0.2) * 8;

				// show muzzle flash.
				if (duel.m_muzzle_flash_timer > 0) {
					duel.m_muzzle_flash.draw(px, py + 60, 2);
				}

				// show pistol.
				duel.m_hunter_pistol.draw(px, py, 2);
			}
		}
		else {
			// Show hunter guy.
			duel.m_hunter_sprite.draw(50, WindowHeight - 50, 100.f / duel.m_hunter_sprite.m_height);
		}
	}

	if (duel.m_game_state == duel::ATTRACT 
		&& duel.m_timer > 10)
	{
		duel.m_intro.draw(WindowWidth / 2, WindowHeight / 2, 1);
	}
	
	duel_text();
}


void DuelTick(float time, float dt)
{
	hunter_tick(dt);
	hunted_tick(dt);

	// ping echo.
	if (duel.m_ping_timer > 0) {
		duel.m_ping_timer -= dt;
		if (duel.m_ping_timer <= 0) {
			duel.m_ping_timer = 0;

			// echo has arrived.

			// compute ping volume & pan.
			vec3f	delta;

			if (duel.m_player_is_hunter) {
				delta = duel.m_hunted_guy->loc - viewpoint_position;
			} else {
				delta = duel.m_hunter_pos - viewpoint_position;
			}

			float	distance = sqrtf(delta * delta);
				
			float	verb_volume;
			if (distance < 20) {
				verb_volume = 1.0f;
			} else {
				verb_volume = (1020.f - distance) / 1000.f;
				if (verb_volume < 0.02) verb_volume = 0.02;
				if (verb_volume > 1) verb_volume = 1;
			}
			float	dry_volume = verb_volume;
			verb_volume = 0.60f + verb_volume * 0.40f;

			float	pan;
			vec3f	delta_dir = delta;
			vec3f_norm(&delta_dir, &delta_dir);
			vec3f	cross;
			vec3f_cross(&cross, &delta_dir, &view_direction);
			pan = cross.z;

			// round out the response.
			if (pan < 0) {
				pan = -powf(-pan, 2.0f);
			} else {
				pan = powf(pan, 2.0f);
			}
				
			sound_play("beep1", dry_volume, pan);
			sound_play("beep1verb", verb_volume * 0.7, pan * 0.3);
		}
	}



    if (duel.m_first_person) {
        // Set the camera position/orientation.
		vec3f	pos;
		float	angle;
		if (duel.m_player_is_hunter) {
			pos = duel.m_hunter_pos;
			angle = duel.m_hunter_angle;
		} else {
			pos = duel.m_hunted_guy->loc;
			angle = duel.m_hunted_angle;
		}

        GlobalCamera->Position[0] = pos.x;
        GlobalCamera->Position[1] = pos.y;
        GlobalCamera->Position[2] = pos.z + 6.0f;
        GlobalCamera->Position[3] = 1;

        GlobalCamera->EAR[0] = 0;        // elevation (pitch)
        GlobalCamera->EAR[1] = angle - Pi32/2.f;   // azimuth (yaw)
        GlobalCamera->EAR[2] = 0;        // roll

        GlobalCamera->Offset[0] = 0;
        GlobalCamera->Offset[1] = 0;
        GlobalCamera->Offset[2] = 0;

        BuildCameraMatrices(*GlobalCamera);
    }
}


void	process_guy_state(float x, float y, uint32 state);
static server_next_guy = 0;



void	send_guy_state()
// We're the server; send out a bunch of guy state packets.
{
	static int	next_guy = 0;

	if (gGuyCount <= 2) return;

	if (next_guy < 2 || next_guy >= gGuyCount) {
		next_guy = 2;	// guys 0 and 1 are hunter/hunted.
	}

	int	seed = rand();
	srand(seed);
	int	next_rand = rand();
	AddClientInput(INPUT_GUY_START, next_guy, seed, next_rand);

	int	count = 500;	// guys to replicate per-frame.
	for (int i = 0; i < count; i++, next_guy++) {
		if (next_guy >= INITIAL_GUYS) {
			break;
		}
		Guy*	g = &gGuys[next_guy];
		
		uint8	buf[4];
		buf[0] = g->ai_mode;
		float angle = atan2f(-g->vel.y, -g->vel.x);
		buf[1] = int(255.0 * (angle + Pi32) / (2 * Pi32));
		buf[2] = g->ai_timer & 0x0FF;
		buf[3] = (g->ai_timer >> 8) & 0x0FF;
		AddClientInput(INPUT_GUY_STATE, *(int*) &(g->loc.x), *(int*) &(g->loc.y), *(int*) (buf));

		server_next_guy = next_guy;
		process_guy_state(*(float*) &(g->loc.x), *(float*) &(g->loc.y), *(int*) (buf));	// for debugging!
	}
}


void	set_server_next_guy(int x)
{
	server_next_guy = x;
}


void	process_guy_state(float x, float y, uint32 state)
// Process a state packet from the server.
{
	if (server_next_guy < 2 || server_next_guy >= gGuyCount) {
		return;
	}

	Guy*	g = gGuys + server_next_guy++;

	float	z = ground(x, y);
	appGuyMove(g, vec3f(x, y, z));

	uint8*	buf = (uint8*) &state;

	g->ai_mode = buf[0];
	float	angle = (buf[1] / 256.0f) * (Pi32 * 2);
	float	speed = FOLLOW_SPEED;
	if (g->ai_mode == AI_FLEE) {
		speed = FOLLOW_SPEED * 2.0f;
	}
	set_direction_and_speed(g, vec3f(cosf(angle), sinf(angle), 0), speed);

	g->ai_timer = (buf[2]) | (buf[3] << 8);
}


void	Player_ProcessInput(float time, float dt)
{
	if (g_LocalPlayer == 0) {
		// Replicate the guy state across the network.
		send_guy_state();
	}

	state_input_tick(dt);
	hunter_input_tick(dt);
	hunted_input_tick(dt);
}


void	setup_camera()
// Make the camera match the duel.m_first_person state.
{
	if (duel.m_first_person) {
		AppCameraInit(FIRST_PERSON_MODE);

		while (ShowCursor(FALSE) >= 0) ;
	}
	else {
		extern Guy*	FollowCamGuy;
		if (duel.m_player_is_hunter && duel.m_game_state == duel::PLAYING) {
			FollowCamGuy = duel.m_hunter_guy;
		} else {
			FollowCamGuy = duel.m_hunted_guy;
		}

		AppCameraInit(FOLLOW_ORBIT_MODE);

		while (ShowCursor(TRUE) < 0) ;
	}
}


int DuelChar(int charcode)
{
	// toggle camera.
	if (charcode == 'f') {
		duel.m_first_person = !duel.m_first_person;
		setup_camera();
	}

    return 0;
//   return BallChar(charcode);
}

void BioTerrorCallback(Guy *collider, Guy *collidee) {
    collidee->collide_flags |= 1;
    collider->debug_color = 1;
}


// g_server_time comes from the server each frame
double g_server_time = - 0.1;
double g_last_server_time = - 0.2;

float ServerDeltaTime()
{
	double delta = (g_server_time - g_last_server_time);
	if ( delta > 0.2 )
		return 0.2f;
	else if ( delta < (1.0/200.0) )
		return (1.f/200.f);
	else
		return float(delta);
}



void duel_init()
{
    EngineSoundInit();

	// Set up the network handling
	g_PostNetworkMessages = MyPostNetworkMessages;
	g_ProcessNetworkMessages = MyProcessNetworkMessage;

    // Get a texture for the hunter's gun.
	duel.m_hunter_pistol.load("sprites/duel/sprites/pisga0.bmp");

	// Texture for muzzle flash.
	duel.m_muzzle_flash.load("sprites/aodtcspr/sprites/bfe1a0.bmp");

	// Intro screen, labels, etc.
	duel.m_intro.load("data/duel_intro1.bmp");
	duel.m_fugitive_label.load("data/duel_fuglabel.bmp");
	duel.m_failure_label.load("data/duel_failure.bmp");
	duel.m_success_label.load("data/duel_success.bmp");
	duel.m_time_label.load("data/duel_time.bmp");

    // Clear the heightfield.
//    memset(hf_old, 0, sizeof(hf_old));
        
	guysInit();
//	setNumGuys(INITIAL_GUYS);
	duel_guys_init(INITIAL_GUYS);

	duelFont = LoadFont("data/arial20.font");

	gameGuyHitTerrain = DuelGuyHitTerrain;
	gameGuyAITimer = duel_guy_ai_timer;

//   BallInit();

   // guys flock twice per second, looking around them at most 8.0 world units,
   // using bit #7 of the collision flags
   // if you want multiple flocking groups, specify the flocking behaviors
   // in the collision flags (e.g. two groups, one sets #6, one sets #7, and
   // pass in #6 + #7 here, i.e. 0x80+0x40)
//   initFlocking(2.0, 8.0, 0x80);

   // bio terror tests happen once every three seconds,
   // look around them at most 2.0 world units, and use
   // bit #0 of the collision flags for bioterror collisions
   // (all guys are collidees, only infected guys are colliders
   // in bit #0)
//   createCollisionManager(0.33, 2, 0x01, BioTerrorCallback);
   // bio terror effect is only visible by turning on DEBUG_COLOR
   // in guys.cpp or by using a debug build.

   setup_camera();
}











/*----------------------------------------------------------------------------

  the game functions

  There's an array of Players

  each client moves his own Player
  and transmits pos & rot to the server

  then server sends out all pos & rot & all button input

  and all clients process that

  server also sends the current "time"

  I am g_LocalPlayer, from net.h

*/

void MyPostNetworkMessages( )
{
//  	// Compute the player position....
//  	vec3f pos = g_players[g_LocalPlayer].guy.loc;
//  	AddClientInput( INPUT_PLAYER_MOVE, pos );
//  	vec3f face = g_players[g_LocalPlayer].facing;
//  	AddClientInput( INPUT_PLAYER_FACE, face );
//  	vec3f vel = g_players[g_LocalPlayer].vel;
//  	AddClientInput( INPUT_PLAYER_VEL, vel );

	if (duel.m_player_is_hunter) {
		// hunter info
		AddClientInput(INPUT_PLAYER_MOVE, vec3f(duel.m_hunter_pos.x, duel.m_hunter_pos.y, 0));
		AddClientInput(INPUT_PLAYER_FACE, vec3f(duel.m_hunter_angle, 0.01, 0));
	}
	else {
		// hunted info
		vec3f	pos = duel.m_hunted_guy->loc;
		pos.z = 1;	// not the hunter.
		AddClientInput(INPUT_PLAYER_MOVE, pos);
		AddClientInput(INPUT_PLAYER_FACE, vec3f(duel.m_hunted_angle, 0.01, 0));
	}
}


// handle net message.
void MyProcessNetworkMessage( int player, Input_t input, int x, int y, int z )
{
	vec3f vec;
	vec.x = *(float*)&x;
	vec.y = *(float*)&y;
	vec.z = *(float*)&z;

	switch(input)
	{
	case INPUT_SERVER_TIME:
		{
			
		extern double g_last_server_time;
		g_last_server_time = g_server_time;

		int buf[3] = { x,y,z };
		g_server_time = *((double *)buf);

		break;
		}
	case INPUT_SERVER_RESET:
		break;

	case INPUT_GAME_START:
		{
			srand(x);
			int	next_rand = rand();
			assert(next_rand == y);
			duel_game_start();
		}
		break;

	case INPUT_PING:
		hunter_ping();
		break;

	case INPUT_SHOOT:
		hunter_shoot_bullet();
		break;

	case INPUT_PLAYER_MOVE:
		{
//			g_players[player].is_active = true;

			// z coord encodes the server's role.  0 == hunter, nonzero == hunted
			if (player == 0) {
				// set our role.
				if (g_LocalPlayer > 0) {
					if (z == 0) {
						duel.m_player_is_hunter = false;
					} else {
						duel.m_player_is_hunter = true;
					}
				}
			}

			if (player != g_LocalPlayer) {
				// replicate remote state.
				vec.z = ground(vec.x, vec.y);
				if (duel.m_player_is_hunter) {
//					duel.m_hunted_guy->ai_mode = AI_FUGITIVE_PLAYER;	// xxx
					appGuyMove(duel.m_hunted_guy, vec);
				}
				else {
					duel.m_hunter_pos = vec;
				}
			}
		}
		break;
	case INPUT_PLAYER_FACE:
		// actually, just angle and speed.
		{
			float	angle = vec.x;
			float	speed = vec.y;

			if (player != g_LocalPlayer) {
				// replicate remote state.
				if (duel.m_player_is_hunter) {
					set_direction_and_speed(duel.m_hunted_guy, vec3f(cosf(angle), sinf(angle), 0), speed);
				}
				else {
					duel.m_hunter_angle = angle;
					set_direction_and_speed(duel.m_hunter_guy, vec3f(cosf(angle), sinf(angle), 0), speed);
				}
			}
//			g_players[player].facing = vec;
		}
		break;

	case INPUT_GUY_START:
		if (g_LocalPlayer > 0) {
			// we're a client.  pull in server stuff.
			srand(y);
			int	next_rand = rand();
			assert(next_rand == z);

			set_server_next_guy(x);
		}
		break;

	case INPUT_GUY_STATE:
		if (g_LocalPlayer > 0) {
			// we're a client.  pull in server stuff.
			process_guy_state(vec.x, vec.y, z);
		}
		break;

#if 0
	case INPUT_PLAYER_VEL:
		{
			g_players[player].vel = vec;
		}
		break;
	case INPUT_BUTTON_DOWN:
		if (x == 0)
		{
			Player_Shoot(g_players[player]);
		}
		else if (x ==  1)
		{
			Player_Scoop(g_players[player]);
		}
		break;

	case INPUT_BUTTON_UP:
		break;

	case INPUT_PLAYER_MOVE:
		{
			g_players[player].is_active = true;
			g_players[player].guy.loc = vec;
		}
		break;
	case INPUT_PLAYER_FACE:
		{
			g_players[player].facing = vec;
		}
		break;
	case INPUT_PLAYER_VEL:
		{
			g_players[player].vel = vec;
		}
		break;

	case INPUT_CHAR:
		{
			switch( x )
			{
			case ' ':
				Player_DropGuys(g_players[player]);
				break;
			case '\\':
				// server can reset with "\\"
				// only accept it if the sender was 0
				if ( player == 0 )
				{
					ResetGame();
				}
				break;
			default:
				break;
			}
		}
		break;
#endif // 0
	}
}
