#if !defined(FONT_H)
/* ========================================================================
   $RCSfile: font.h,v $
   $Date: 2001/12/17 06:46:43 $
   $Revision: 1.2 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#define MaxCharacterCount 128

/*
  OK.  So to load a font, you just use:

  font *Font = LoadFont("something.font");

  If the display mode changes, you can reload the textures like this:

  ReloadFont(*Font);

  When you're done, free it all with

  FreeFont(Font);
  
  To actually draw shit, use

  glEnable(GL_TEXTURE_2D);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glTranslatef(DrawAtX, DrawAtY, 0.0f);
  DrawString(*Font, "This is my string!");

  or to get the dimensions of the string, use

  int Width, Height;
  GetDimensions(*Font, "This is my string!", Width, Height);  

  Note that you will need to do everything as an ortho pass
  (unless the text is supposed to be in 3D), so you probably want
  to have done this:
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glOrtho(0, (real32)WindowWidth, 0, (real32)WindowHeight, -1.0f, 1.0f);

  somewhere before the font action.

  To debug the font textures, you can draw them like this:

  RenderFontTexture(*Font, 0);

  where the 0 can be anything up to GetFontTextureCount.
  
 */

struct font_character
{
    int TextureIndex;
    int Width, Height;
    float XMin, XMax, YMin, YMax;
    
    int Kerning[MaxCharacterCount];
};

struct font_header
{
    int Height;
    int BaseLine;

    int TextureWidth;
    int TextureHeight;
    int TextureCount;

    int CharacterCount;
    font_character Characters[MaxCharacterCount];
};

struct font;
font *LoadFont(char const *FontFileName);
bool ReloadFont(font &Font);
void FreeFont(font *Font);

int GetFontTextureCount(font &Font);
void RenderFontTexture(font &Font, int Index);

void DrawString(font &Font, char const *String);
void GetDimensions(font &Font, char const *String, int &Width, int &Height);

#define FONT_H
#endif
