/*----------------------------------------------------------------------------

  game.cpp - a simple skeleton to modify into your Indie Game Jam game

*/

#include <assert.h>

#include "globals.h"
#include "system.h"
#include "guys.h"
#include "animation_public.h"
#include "value_map.h"

/*----------------------------------------------------------------------------

  some variables to control stuff

*/

Animation_Texture_Set *setA, *setB;
Animation_Mode *modeA, *modeB;
int const TotalGuysA = 50000;
int const TotalGuysB = 50000;
int const TotalGuys = TotalGuysA + TotalGuysB;

/*----------------------------------------------------------------------------

  the game functions

*/

void GameInit(void)
{
    guysInit();

    // load the terrain texture
    MapSetTerrainTexture("data/game.psd");

    // load the height field
    value_map *HF = MapLoad("data/game.psd",5); assert(HF);
    MapSetHeightField(HF);
    
    // create our guys

    // team A is on the left side of the map

    // set up texture stuff
    modeA = animation_mode_create();
    setA = animation_texture_set_create_doom_style("sprites/doom95/sprites/play",'a',4,8);
    animation_mode_set_texture(modeA,setA);
    // create a template for them
    Guy teamA_template;
    guysClearGuy(&teamA_template);
    setSpeed(&teamA_template,10);
    setVelocity(&teamA_template,1,0,0);
    animation_dude_set_mode(&teamA_template,modeA);
    // load density map and distribute the guys
    value_map *DensityA = MapLoad("data/game.psd",3);  assert(DensityA);
    MapAllocateGuyDensity(DensityA,&teamA_template,0,TotalGuysA,true);
    MapFree(DensityA);

    // team B is on the right side of the map
    modeB = animation_mode_create();
    setB = animation_texture_set_create_doom_style("sprites/fistful/sprites/cpos",'a',4,8);
    animation_mode_set_texture(modeB,setB);
    Guy teamB_template;
    guysClearGuy(&teamB_template);
    setSpeed(&teamB_template,12);
    setVelocity(&teamB_template,-1,0,0);
    animation_dude_set_mode(&teamB_template,modeB);
    value_map *DensityB = MapLoad("data/game.psd",4);  assert(DensityB);
    MapAllocateGuyDensity(DensityB,&teamB_template,TotalGuysA,TotalGuysB,true);
    MapFree(DensityB);

    

    guysFinishInit(TotalGuys);
}

void GameRender(void)
{
}

void GameTick(float time, float dt)
{
}

int GameChar(int charcode)
{
    return 0;
}

/*----------------------------------------------------------------------------

  system hooks

*/

#if 0  // change this to 1 to enable (make sure you change angry.cpp as well)
// define system hooks and hook ourselves in
system_hooks SystemHooks = {
    GameInit,
    GameRender,
    GameTick,
    GameChar,
};
#endif

