#ifndef INC_GUYS_H
#define INC_GUYS_H

struct Guy
{
    // engine variables
    // there are APIs for setting these below
    // don't touch them directly!
    vec3f  loc;                         // 12
    vec3f  vel;                         // 24
    uint16 bin;                         // 26
    uint8  animation_mode_index;        // 27
    uint8  animation_direction;         // 28
    uint8  animation_timing_offset;     // 29
    uint8  collide_flags;               // 30
    uint8  be_collided_with_flags;      // 31
    uint8  ai_mode;                     // 32
    uint16 ai_timer;                    // 34
    uint16 num_flock;                   // 36
    uint16 bin_offset;                  // 38
    uint8  speed;                       // 39
    uint8  physics_mode;                // 40
    Guy  **ref_guy;                     // 44

    // game variables - 32bit aligned
    // party on these all you want
    // suggested use examples:
    // #define TYPE(g)  (*(uint8*)((g)->game_data+0))  // 0   uint8
    // #define ANGER(g) (*(float*)((g)->game_data+4))  // 4-7 float
    // etc., make sure non uint8's are aligned
    // or change this header to add new fields, but make sure it adds to 64
    uint8  game_data[19];               // 63

    // in debug mode (#define _DEBUG), this will
    // choose a color from the debug_colors enum
    // that will modulate the guy
    // for non-debug, it's best to use a different sprite
    uint8  debug_color;                 // 64
};

enum debug_colors {
   DEBUG_COLOR_NORMAL,
   DEBUG_COLOR_WHITE = DEBUG_COLOR_NORMAL,
   DEBUG_COLOR_GREY,
   DEBUG_COLOR_RED,
   DEBUG_COLOR_BLUE,
   DEBUG_COLOR_GREEN,
   DEBUG_COLOR_CYAN,
   DEBUG_COLOR_YELLOW,
   DEBUG_COLOR_MAGENTA,
};

extern int  gGuyCount;  // # of guys
extern Guy *gGuys;      // all the guys

#define GUYBIN(g)  (g->bin)

// basic guy stuff
extern void guysDraw(float t);
extern void guysInit(void);
extern void guysKey(int ch);
extern void guysFinishInit(int guyCnt);
extern void guysFindCollisions(float dt, float radius, int flags);
extern void guysClearGuy( Guy *guy );  // for templates, etc.

// @todo document invariants...ref currently only changed on deleteGuy
extern void guysRegisterRef( Guy *guy, Guy **ref_guy );
extern void guysUnregisterRef( Guy **ref_guy );
// you really shouldn't be keeping unregistered Guy pointers, use the ref_guy
// facility above, but if you have one, this will tell you if it's
// valid (but it might not point to the same guy)
inline bool IsGuyValid( Guy *guy )
{
    if(!guy || (uint32(guy - gGuys) >= uint32(gGuyCount))) {
        return false;
    }
    return true;
}

// more guy stuff
extern void deleteGuy(int g);
extern void rebinGuy(int g);
extern void binGuy(int g);

// safe ways to set guy state...
extern void setSpeed(Guy *g, float speed);
extern void setVelocity(Guy *g, float vx, float vy, float vz);
extern void	setDirectionAndSpeed(Guy *g, float vx, float vy, float speed);
extern void setNumGuys(int num);

// run guys by hand
extern void runGuys(Guy *g, int n, int t);

// draw guy by hand
extern void drawCenteredWithScale(Guy *g, int n, float hscale, float vscale);

// query for nearby guys, the super dumb-ass way (tm) - cback ret 1 to delete
extern void findGuysInSquare(float x0, float y0, float x1, float y1, void (*f)(Guy *, void *p), void *p);
extern void findGuysInCircle(float cx, float cy, float r, void (*f)(Guy *, float dsquared, void *p), void *p);
extern void findGuysInSphere(float cx, float cy, float cz, float r, void (*f)(Guy *, float dsquared, void *p), void *p);

// try and go and move along this thing...
extern void aimGuyAlong(Guy *g);
extern int guyAimFor(Guy *g, float x, float y, float speed);
extern int guyAimNear(Guy *g, float x, float y, float r, float speed);

// callbacks for collision stuff, stuff these to get these events
extern float (*gameGuyHitTerrain)(Guy *g);
extern void (*gameGuyAITimer)(Guy *g);

// flocking and guy-guy collisions
struct Collision_Manager {
    float fraction_of_bins_to_process_per_second;
    int next_bin_index_to_process;
    float collision_radius;
    int flags;
    void (*hit_func)(Guy *collider, Guy *collidee);
    void (*end_func)(void);
    bool active;
};

extern Collision_Manager *createCollisionManager(float freq, float radius, int flags, void (*func)(Guy *a, Guy *b));
extern void deleteCollisionManager(Collision_Manager *c);
extern void runCollisionManager(Collision_Manager *c, float dt);
extern void findCollisions(float dt);
extern void initFlocking(float freq, float radius, int flags);


#define MAX_GUYS  (1024*100)

// time is fun, lets go shopping
#define INT_TIME_PER_FLOAT_TIME     1000   // integer timescale is milliseconds
#define INT_TIME_TO_FLOAT_TIME(x)   ((float) (x) / INT_TIME_PER_FLOAT_TIME)
#define FLOAT_TIME_TO_INT_TIME(x)   ((int) ((float) (x) * INT_TIME_PER_FLOAT_TIME + 0.5))

#define INT_MAX_TIMER  ((1 << 16) - 2)
#define FLOAT_MAX_TIMER  INT_TIME_TO_FLOAT_TIME(INT_MAX_TIMER)

// TODO: Hey guys, so the GUY_WIDTH and GUY_HEIGHT things are
// how big the things are drawn, but then in animation.cpp there's
// the code that loads the bitmaps and tells them what width and
// height THEY should be, and it's pretty crucial to make sure
// the aspect ratio on both is the same, so someone should probably
// decide how they want to do that.
#define GUY_WIDTH   6
#define GUY_HEIGHT  6
#define GUY_SPEED   8

#define GUY_MAX_SPEED  96

#define ONE_OVER_GUY_SPEED   (1.0 / GUY_SPEED)

#define GUY_TIMER   (256*4)

#define DEFAULT_GRAVITY 48.0  // was once 16, have some units

enum
{
   PM_FLOORED,
   PM_FLY_THROUGH_BUILDINGS,
   PM_FLY_AVOID_BUILDINGS,
   PM_GRAVITY,
};

extern float gravity[];  // 0 = floored, 1,2 = flying, 3+ = gravity modes above

extern float appGuyHitTerrain(Guy *g); // percentage of vertical velocity to keep

extern void appGuyMove(Guy* g, const vec3f& new_loc);	// takes care of binning.

// really in app, should be moved to utitlies
// these are all things that should be setup like that...
extern void  AppSetupFontDraw(float x, float y);
// HEY - how do I know how wide a string is while using the font?

// where is the 3d cursor
extern vec3f playerCursorPos;

extern void appGuyCollisionResponse(Guy *collder, Guy *collidee);

#endif // INC_GUYS_H
