#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <math.h>
#include <assert.h>

#include "globals.h"
#include "buildings.h"
#include "hf.h"
#include "camera.h"

#define MULTITEXTURE 1

float height_field[HF_Y][HF_X];
int ground_tex, detail_tex;
int ground_tex_w, ground_tex_h;
uint8 *ground_tex_data;    // this always has to point to a BGRA uint8 array w*h*4 that has the bits of ground_tex in it

void texMake32(int w, int h, uint8 *data, int texture_id, uint WrapMode);
int texMake32(int w, int h, uint8 *data, uint WrapMode);

void hfInit(void)
{
   int i,j;
   for (i=0; i < HF_X; ++i)
     for (j=0; j < HF_Y; ++j)
        height_field[j][i] = 0.f;
}

float ground_outside_z = 0.f;
float ground(float x, float y)
{
   int gx, gy;
   float h00, h01, h10, h11, h0, h1, h;
   x /= HF_SCALE;
   y /= HF_SCALE;

   gx = floor(x);
   gy = floor(y);
   x -= gx;
   y -= gy;

   // @todo this should handle gx = HF_X-1 and do the linear special case thing
   if (gx < 0 || gy < 0 || gx >= HF_X-1 || gy >= HF_Y-1) return ground_outside_z;

   h00 = height_field[gy  ][gx  ];
   h01 = height_field[gy  ][gx+1];
   h10 = height_field[gy+1][gx  ];
   h11 = height_field[gy+1][gx+1];

   if (x > y) {
      h10 = h00 + h11 - h01;
   } else {
      h01 = h00 + h11 - h10;
   }

   h0 = h00 + (h01-h00)*x;
   h1 = h10 + (h11-h10)*x;

   h = h0 + (h1-h0)*y;

   return h;
}

float ground_slope(float x, float y, float *dx, float *dy)
{
   assert(dx && dy);
   int gx, gy;
   float h00, h01, h10, h11, h0, h1, h;
   x /= HF_SCALE;
   y /= HF_SCALE;

   gx = floor(x);
   gy = floor(y);
   x -= gx;
   y -= gy;

   if (gx < 0 || gy < 0 || gx >= HF_X-1 || gy >= HF_Y-1) {
       *dx = 0;
       *dy = 0;
       return ground_outside_z;
   }

   h00 = height_field[gy  ][gx  ];
   h01 = height_field[gy  ][gx+1];
   h10 = height_field[gy+1][gx  ];
   h11 = height_field[gy+1][gx+1];

   if (x > y) {
      h10 = h00 + h11 - h01;
   } else {
      h01 = h00 + h11 - h10;
   }

   *dx = (h01 - h00) / HF_SCALE;
   *dy = (h10 - h00) / HF_SCALE;

   h0 = h00 + (h01-h00)*x;
   h1 = h10 + (h11-h10)*x;

   h = h0 + (h1-h0)*y;

   return h;
}

// @todo
int texMake24(int w, int h, uint8 *data, uint WrapMode)
{
   uint id;
   glGenTextures(1, &id);
   glBindTexture(GL_TEXTURE_2D, id);

   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     WrapMode);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     WrapMode);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);

   gluBuild2DMipmaps(GL_TEXTURE_2D, 3, w, h, GL_RGB, GL_UNSIGNED_BYTE, data);
   return id;
}

void texMake32(int w, int h, uint8 *data, int id, uint WrapMode)
{
   glBindTexture(GL_TEXTURE_2D, id);

   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     WrapMode);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     WrapMode);

   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

#if 0
   gluBuild2DMipmapsThatDontSuckCompleteAss(GL_TEXTURE_2D, 4, w, h, GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
#else
   gluBuild2DMipmaps(GL_TEXTURE_2D, 4, w, h, GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
#endif
}

int texMake32(int w, int h, uint8 *data, uint WrapMode) {
   uint id;
   glGenTextures(1, &id);
   glBindTexture(GL_TEXTURE_2D, id);

   texMake32(w, h, data, id, WrapMode);

   return id;
}

int const checkerSize = 8;
const int det_x=128, det_y=128;
void texInit(void)
{
    ground_tex_data = new uint8[checkerSize*checkerSize*4];  assert(ground_tex_data);
    ground_tex_w = ground_tex_h = checkerSize;
    uint8 (*checker)[checkerSize][3] = (uint8 (*)[checkerSize][3])ground_tex_data;
    for (int i=0; i < checkerSize; ++i) {
        for (int j=0; j < checkerSize; ++j) {
            checker[j][i][0] = 0x0; // (i & 1   ) << 7;
            checker[j][i][1] = 0x0; // (j & 1) << 7;     // ((i+3) << 4) ^ ((j+6)<<4);
            checker[j][i][2] = 0x0; // ((i+6) << 4) ^ ((j+3)    <<4);
        }
    }
    ground_tex = texMake24(checkerSize,checkerSize,checker[0][0],GL_CLAMP_TO_EDGE);
    uint8 *detail = new uint8 [det_x * det_y * 3];
    for (int j=0; j < det_y; ++j)
       for (int i=0; i < det_x; ++i) {
          int z = (j*det_x + i) * 3;
          int t = random(215,255);
          detail[z+0] = t * (1-0.1*frandom());
          detail[z+1] = t * (1-0.1*frandom());
          detail[z+2] = t * (1-0.1*frandom());
       } 
    detail_tex = texMake24(det_x, det_y, detail,GL_REPEAT);
    delete [] detail;
}

extern void glMultiTexCoord2f(int n, GLfloat s, GLfloat t);
extern void glActiveTexture(int n);

#define DETAIL_SCALE  9.25

static void drawHFregular(bool skirt)
{
   float du = 1.0 / (double)(HF_X - 1);
   float dv = 1.0 / (double)(HF_Y - 1);
   int i,j;
   
#define VERTNMXYZ(i,j,x,y,z) \
    glTexCoord2f((i) * du, (j) * dv); \
    glVertex3f((x), (y), (z));
#define VERTNM(i,j) VERTNMXYZ(i,j,(i)*HF_SCALE,(j)*HF_SCALE,height_field[(j)][(i)])
#if MULTITEXTURE
#define VERT(i,j) glMultiTexCoord2f(1, (i)*DETAIL_SCALE,(j)*DETAIL_SCALE);  VERTNM(i,j)
#define VERTXYZ(i,j,x,y,z) glMultiTexCoord2f(1, (i)*DETAIL_SCALE,(j)*DETAIL_SCALE);  VERTNMXYZ(i,j,x,y,z)
#else
#define VERT(i,j) VERTNM(i,j)
#define VERTXYZ(i,j,x,y,z) VERTNM(i,j,x,y,z)
#endif   


   // draw the height field itself
   for (j=0; j < HF_Y-1; ++j) {
      glBegin(GL_TRIANGLE_STRIP);
      VERT(0,j+1);
      for (int i=0; i < HF_X-1; ++i) {
          VERT(i,j);
          VERT(i+1,j+1);
      }
      VERT(i,j);
      glEnd();
   }
   if (!skirt) return;

   // now draw the border walls
   float infinity = 100.0*max(WORLD_X,WORLD_Y);
   float borderz = ground_outside_z;
   
   // disable texturing and make it black, clamp_to_edge looks like crap
#if 1
#if MULTITEXTURE
   glActiveTexture(1);
   glDisable(GL_TEXTURE_2D);
   glActiveTexture(0);
#endif
   glDisable(GL_TEXTURE_2D);
   glColor3f(0,0,0);
#endif
   
   // bottom row
   i = j = 0;
   glBegin(GL_TRIANGLE_STRIP);
   VERT(i,j);
   for (i=0; i < HF_X-1; ++i) {
      VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
      VERT(i+1,j);
   }
   VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
   glEnd();
   // middle rows
   for (j=0; j < HF_Y-1; ++j) {
      glBegin(GL_QUADS);
      i = 0;
      VERTXYZ(i,j+1,i*HF_SCALE,(j+1)*HF_SCALE,borderz);
      VERTXYZ(i,j+0,i*HF_SCALE,(j+0)*HF_SCALE,borderz);
      VERT(i,j+0);
      VERT(i,j+1);
      i = HF_X-1;
      VERT(i,j+1);
      VERT(i,j);
      VERTXYZ(i,j+0,i*HF_SCALE,(j+0)*HF_SCALE,borderz);
      VERTXYZ(i,j+1,i*HF_SCALE,(j+1)*HF_SCALE,borderz);
      glEnd();
   }
   // top row
   glBegin(GL_TRIANGLE_STRIP);
   i = 0; j = HF_Y-1;
   VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
   for (i=0; i < HF_X-1; ++i) {
       VERT(i,j);
       VERTXYZ(i+1,j,(i+1)*HF_SCALE,j*HF_SCALE,borderz);
   }
   VERT(i,j);
   glEnd();

   // now do border floor
   // bottom row
   i = j = 0;
   glBegin(GL_TRIANGLE_STRIP);
   VERTXYZ(i,j,-infinity,j*HF_SCALE,borderz);
   VERTXYZ(i,j,-infinity,-infinity,borderz);
   VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
   for (i=0; i < HF_X-1; ++i) {
      VERTXYZ(i,j,i*HF_SCALE,-infinity,borderz);
      VERTXYZ(i+1,j,(i+1)*HF_SCALE,j*HF_SCALE,borderz);
   }
   VERTXYZ(i,j,i*HF_SCALE,-infinity,borderz);
   VERTXYZ(i,j,infinity,j*HF_SCALE,borderz);
   VERTXYZ(i,j,infinity,-infinity,borderz);
   glEnd();
   // middle rows
   for (j=0; j < HF_Y-1; ++j) {
      glBegin(GL_QUADS);
      i = 0;
      VERTXYZ(i,j+1,-infinity,(j+1)*HF_SCALE,borderz);
      VERTXYZ(i,j+0,-infinity,(j+0)*HF_SCALE,borderz);
      VERTXYZ(i,j+0,i*HF_SCALE,(j+0)*HF_SCALE,borderz);
      VERTXYZ(i,j+1,i*HF_SCALE,(j+1)*HF_SCALE,borderz);
      i = HF_X-1;
      VERTXYZ(i,j+1,i*HF_SCALE,(j+1)*HF_SCALE,borderz);
      VERTXYZ(i,j+0,i*HF_SCALE,(j+0)*HF_SCALE,borderz);
      VERTXYZ(i,j+0,infinity,(j+0)*HF_SCALE,borderz);
      VERTXYZ(i,j+1,infinity,(j+1)*HF_SCALE,borderz);
      glEnd();
   }
   // top row
   glBegin(GL_TRIANGLE_STRIP);
   i = 0; j = HF_Y-1;
   VERTXYZ(i,j,-infinity,infinity,borderz);
   VERTXYZ(i,j,-infinity,j*HF_SCALE,borderz);
   VERTXYZ(i,j,i*HF_SCALE,infinity,borderz);
   for (i=0; i < HF_X-1; ++i) {
       VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
       VERTXYZ(i+1,j,(i+1)*HF_SCALE,infinity,borderz);
   }
   VERTXYZ(i,j,i*HF_SCALE,j*HF_SCALE,borderz);
   VERTXYZ(i,j,infinity,infinity,borderz);
   VERTXYZ(i,j,infinity,j*HF_SCALE,borderz);
   glEnd();
}

void drawHF(void)
{
#if MULTITEXTURE
   glActiveTexture(1);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, detail_tex);
   glActiveTexture(0);
#endif   
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, ground_tex);

   glDisable(GL_BLEND);
   glDisable(GL_LIGHTING);
   glColor3f(1.0, 1.0, 1.0);

   assert(HF_X > 1);
   assert(HF_Y > 1);
   drawHFregular(true);
#if MULTITEXTURE
   glActiveTexture(1);
   glDisable(GL_TEXTURE_2D);
   glActiveTexture(0);
#endif
   glDisable(GL_TEXTURE_2D);
}

#define BUF_SIZE  64
bool pickHF(vec3f *res, int x, int y)
{
   y = WindowHeight - y;

   GLuint buffer[BUF_SIZE];
   glSelectBuffer(BUF_SIZE, buffer);
   glRenderMode(GL_SELECT);
   glInitNames();

   GLint viewport[4];
   glGetIntegerv(GL_VIEWPORT, viewport);

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluPickMatrix(x, y, 2, 2, viewport);
   glMultMatrixf((float *)(*GlobalCamera).Projection4x4);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadMatrixf((float *)(*GlobalCamera).View4x4);

   drawHFregular(false);

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

   int n = glRenderMode(GL_RENDER);
   if (n <= 0) return false;
   float depth = buffer[1] / (float) ((uint32) -1);

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glMultMatrixf((float *)(*GlobalCamera).Projection4x4);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadMatrixf((float *)(*GlobalCamera).View4x4);

   {
      GLdouble model[16];
      GLdouble proj[16];
      GLdouble out_x, out_y, out_z;
      glGetDoublev(GL_MODELVIEW_MATRIX, model);
      glGetDoublev(GL_PROJECTION_MATRIX, proj);
      gluUnProject(x, y, depth, model, proj, viewport, &out_x, &out_y, &out_z);
      res->x = out_x;
      res->y = out_y;
      res->z = out_z;
   }
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

   return true;
}

