#if !defined(MEMORY_ARENA_H)
/* ========================================================================
   $RCSfile: memory_arena.h,v $
   $Date: 2002/03/14 01:08:43 $
   $Revision: 1.2 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#if !defined(GLOBALS_H)
#include "globals.h"
#endif

/*
  The memory_arena is just a braindead thing you can use to automatically
  group a bunch of allocation together, and they chain themselves, so you
  can free them as a batch.  Do NewMemoryArena() to make one, then call
  MemoryArenaPush() to get space on it.  When you're done, do a
  FreeMemoryArena().  There are also some macros for doing the pushes, so
  that they're "type safe" (or, at least, more convenient).
 */
struct memory_arena;

memory_arena *NewMemoryArena(void);
void ClearArena(memory_arena *Arena);
void FreeMemoryArena(memory_arena *Arena);

void *MemoryArenaPush(memory_arena &Arena, int32 Size);

#define ArenaPush(Arena, type) (type *)MemoryArenaPush(Arena, sizeof(type))
#define ArenaPushSize(Arena, Size) MemoryArenaPush(Arena, Size)
#define ArenaPushArray(Arena, Count, type) (type *)MemoryArenaPush(Arena, (Count) * sizeof(type))

#define MEMORY_ARENA_H
#endif
