// (c) 2000 by Zachary Booth Simpson
// This code may be freely used, modified, and distributed without any license
// as long as the original author is noted in the source file and all
// changes are made clear disclaimer using a "MODIFIED VERSION" disclaimer.
// There is NO warranty and the original author may not be held liable for any damages.
// http://www.totempole.net

#include "windows.h"
#include "zocket.h"
#include "msgzocket.h"
#include "hashtable.h"
#include "stdio.h"
#include "string.h"
#include "malloc.h"
#ifdef WIN32
	#include "conio.h"
#else
	#include "sys/time.h"
	#include "sys/types.h"
	#include "termios.h"
	#include "unistd.h"
#endif
#include "net.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "vec3f.h"
//#include "myplayer.h"

#define MAX_SERVER_WAIT_FOR_CLIENT	(2.0)
#define MAX_CLIENT_WAIT_TO_LOGIN	(5.0)

int g_LocalPlayer = 0;
//bool in_input_gather = false;

PostNetworkMessages_t g_PostNetworkMessages = NULL;
ProcessNetworkMessage_t g_ProcessNetworkMessages = NULL;

void	dprintf(const char* fmt, ...)
{
	va_list			ArgPtr;
	char			TempStr[1024];

	va_start(ArgPtr, fmt);
	_vsnprintf(TempStr, 1024, fmt, ArgPtr);
	va_end(ArgPtr);

	OutputDebugString(TempStr);
}

void checkZocket(Zocket * z)
{
	assert(z);
	ZocketError err = z->getZocketError();
	if ( err == zeNO_ERROR )
	{
		assert(z->isOpen());
		assert(z->getSockFD() != INVALID_SOCKET);
	}
	else
	{
		dprintf("zock error: %d\n",err);
	}
}

// Client Lists, used by server
//-----------------------------

#define MAX_CLIENTS	(10)

struct Client 
{
	int gotInput;
	int player;
	SimpleMsgZocket *zocket;
	int gotFirstInput;

	bool loggedIn() { return zocket != NULL; }

	Client() : gotInput(0), player(0), zocket(NULL), gotFirstInput(0) { }
} clients[MAX_CLIENTS];

int server = 0;
	// Am I a server

bool gotNewClients = false;

// Server Zocket, used by client
//-----------------------------

SimpleMsgZocket *serverZocket = NULL;

SimpleMsgZocket *listeningZocket = NULL;
	// This is the listening zocket used
	// only by the server.

int client = 0;
	// Am I a client

int clientIsLoggedIn = 0;
	// client is logged in

int clientIsConnected = 0;

int inputFlags = 0;

// Messages
//======================================================

struct MsgLogin : public SimpleMsg 
{
	int player;
	MsgLogin() { type = msgLogin; len = sizeof(*this); }
};

struct MsgClientInput : public SimpleMsg 
{
	int m_Count;
	MsgClientInput() { type = msgClientInput; len = sizeof(*this); }
};

struct MsgAllPlayerInput : public SimpleMsg 
{
	int m_TotalCount;
	unsigned short m_Count[MAX_PLAYERS];
	unsigned short m_FirstIndex[MAX_PLAYERS];

	MsgAllPlayerInput() { type = msgAllPlayerInput; len = sizeof(*this); }
};

struct MsgInput_t
{
	Input_t m_Input;

	int m_X;
	int m_Y;
	int m_Z;

	vec3f & vec() { return *((vec3f *)&m_X); }
};


enum { MAX_INPUT = 1024 };

struct MsgBuffer_t : public MsgClientInput
{
	MsgInput_t m_MsgInput[MAX_INPUT];
};

struct MsgAllPlayer_t : public MsgAllPlayerInput
{
	MsgInput_t m_MsgInput[ MAX_PLAYERS * MAX_INPUT ];
};

static MsgBuffer_t s_MsgBuffer;
static MsgAllPlayer_t s_AllMsgBuffer;
static MsgAllPlayer_t s_ClientAllMsgBuffer;
static bool s_AllInputReceived;

void AddClientInput( Input_t input, int x, int y, int z )
{
	assert( s_MsgBuffer.m_Count < MAX_INPUT );
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].m_Input = input;
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].m_X = x;
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].m_Y = y;
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].m_Z = z;
	++s_MsgBuffer.m_Count;
}

void AddClientInput( Input_t input, const vec3f & v )
{
	assert( s_MsgBuffer.m_Count < MAX_INPUT );
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].m_Input = input;
	s_MsgBuffer.m_MsgInput[s_MsgBuffer.m_Count].vec() = v;
	++s_MsgBuffer.m_Count;
}

void AddServerTime()
{
	// server time must be the first message each frame !!
	assert( s_AllMsgBuffer.m_TotalCount == 0 );
	MsgInput_t & input = s_AllMsgBuffer.m_MsgInput[s_AllMsgBuffer.m_TotalCount];
	s_AllMsgBuffer.m_TotalCount++;

	input.m_Input = INPUT_SERVER_TIME;
	int *pData = &input.m_X;
	*((double *)pData) = GetTime();
}

void AddServerReset()
{
	AddClientInput(INPUT_SERVER_RESET,rand(),0,0);
}

void TransmitClientInput()
{
	if (client)
	{
		AddClientInput(INPUT_CLIENT_EOM,0,0,0);

		s_MsgBuffer.len = sizeof(MsgClientInput) + s_MsgBuffer.m_Count * sizeof(MsgInput_t);
		serverZocket->write( &s_MsgBuffer );
		s_MsgBuffer.m_Count = 0;
	}
}

bool IsPlayerConnected( int player )
{
	return clients[player].loggedIn();
}


// Message handlers
//======================================================

void handleTerminate( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	//assert( zocket != serverZocket );
	if ( zocket == serverZocket )
	{
		dprintf("terminating server!!\n");
		zocket->kill();
		serverZocket = NULL;
	}
	else if ( zocket == listeningZocket )
	{
		dprintf("terminating listener!!\n");
		zocket->kill();
		listeningZocket = NULL;
	}
	else
	{
		Client* pClient = (Client*)zocket->appData;

		dprintf("handleTerminate : %d\n",pClient->player);

		zocket->kill();

		if( pClient ) 
		{
			// I use the app data to store a pointer to
			// my global zocket handles.  This makes the
			// code so much easier to deal with, without
			// having to do a level of indirection.
			pClient->zocket = NULL;
			pClient->gotFirstInput = false;
		}
		zocket->appData = 0;
	}
}

void handleAcceptedConnection( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	// this zocket just got made
	dprintf("handleAcceptedConnection!!\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	{
		int is_open = zocket->isOpen();
		dprintf("login zocket open : %d\n",is_open);
		int is_con = zocket->isConnected();
		dprintf("login zocket connected : %d\n",is_con);
	}

	if( server ) 
	{
		zocket->setActivityTimeout( 3 * 60 * 60 );
		zocket->setLoginTimeout( 15 * 60 );
		zocket->setRecvTimeout( 15 * 60 );
		
			{
				int is_open = zocket->isOpen();
				dprintf("login zocket open : %d\n",is_open);
				int is_con = zocket->isConnected();
				dprintf("login zocket connected : %d\n",is_con);
			}
	}
}

void handleConnected( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	dprintf("handleConnected!!\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());
	
	if (zocket == serverZocket)
		clientIsConnected = 1;
}

void handleLoginConfirm( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	dprintf("handleLoginConfirm!!\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	clientIsLoggedIn = 1;
}

void handleLogin( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	dprintf("handleLogin!!\n");
	
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	// A login associates a name with a zocket so
	// that the server can send to people by name
	// if it chooses to.
	MsgLogin *m = (MsgLogin *)msg;
	if( g_LocalPlayer == 0 ) // I'm the server
	{
		int c = m->player;
		if( clients[c].loggedIn() ) 
		{
			dprintf("handleLogin : %d already loggin in!!\n",c);
		}
		else 
		{
			// brand new client
			int iClient = 0;
			for(; ;)
			{
				if ( clients[iClient].zocket == NULL )
					break;
				iClient++;
				if ( iClient == MAX_CLIENTS )
					return; // can't log him in !
			}

			dprintf("handleLogin : %d new client %d!\n",c,iClient);

			clients[iClient].player = m->player;
			clients[iClient].zocket = zocket;
			zocket->appData = (int)&clients[iClient];
				// Set the app data to point to the
				// cached handle.  It will be reset by
				// handleTerminate

			{
				int is_open = zocket->isOpen();
				dprintf("login zocket open : %d\n",is_open);
				int is_con = zocket->isConnected();
				dprintf("login zocket connected : %d\n",is_con);
			}
			
			zocket->setActivityTimeout( 3 * 60 * 60 );
			zocket->setLoginTimeout( 15 * 60 );
			zocket->setRecvTimeout( 15 * 60 );

			{
				int is_open = zocket->isOpen();
				dprintf("login zocket open : %d\n",is_open);
				int is_con = zocket->isConnected();
				dprintf("login zocket connected : %d\n",is_con);
			}

			SimpleMsg msg;
			msg.type = msgLoginConfirm;
			msg.len = sizeof(msg);
			zocket->write( &msg );
			//zocket->setLoginTimeout( 0 );

			Sleep(10);

			// track getting new clients so the server can reset
			if ( iClient != 0 )
			{
				gotNewClients = true;
			}
		}
	}
}

void handleClientInput( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	// can get input at any time !
	//assert(in_input_gather);
	//if ( ! in_input_gather )
	//	dprintf("got handleClientInput not in handleClientInput\n");

	Client* pClient = (Client*)zocket->appData;
	pClient->gotInput = 1;
	pClient->gotFirstInput = true;

	// Add the message into the aggregate
	MsgClientInput *m = (MsgClientInput *)msg;

	s_AllMsgBuffer.m_Count[pClient->player] = m->m_Count;
	s_AllMsgBuffer.m_FirstIndex[pClient->player] = s_AllMsgBuffer.m_TotalCount;
	memcpy( &s_AllMsgBuffer.m_MsgInput[s_AllMsgBuffer.m_TotalCount],
		(m+1), m->m_Count * sizeof(MsgInput_t) );
	s_AllMsgBuffer.m_TotalCount += m->m_Count;

	// last message should be INPUT_CLIENT_EOM
	MsgInput_t * inputs = (MsgInput_t *)(m+1);
	MsgInput_t & last = inputs[m->m_Count-1];
	assert( last.m_Input == INPUT_CLIENT_EOM );
	if ( last.m_Input != INPUT_CLIENT_EOM )
		dprintf("got client input without EOM!\n");
}

void handleAllPlayerInput( SimpleMsgZocket *zocket, SimpleMsg *msg ) 
{
	Client* pClient = (Client*)zocket->appData;

	// Add the message into the aggregate
	MsgAllPlayerInput *m = (MsgAllPlayerInput *)msg;

	memcpy( &s_ClientAllMsgBuffer, m, msg->len );
	s_AllInputReceived = true;
}


// State
//======================================================

void startupServer() 
{
	dprintf("StartupServer\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	listeningZocket = new SimpleMsgZocket( "tcp(lb)://*:5050" );
	checkZocket(listeningZocket);
	listeningZocket->appData = (int)&listeningZocket;
	server = 1;

	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());
}

void shutdownServer() 
{
	if( server ) 
	{
		if( listeningZocket ) 
		{
			listeningZocket->kill();
			listeningZocket = NULL;
		}
		for( int i=1; i<MAX_CLIENTS; i++ ) 
		{
			if( clients[i].loggedIn() ) 
			{
				clients[i].zocket->kill();
			}
		}
		memset( clients, 0, sizeof(Client) * MAX_CLIENTS );
		server = 0;
	}
}

void startupClient() 
{
	client = 1;
	clientIsLoggedIn = 0;
	clientIsConnected = 0;
}

void shutdownClient() 
{
	if( client ) 
	{
		if( serverZocket ) 
		{
			serverZocket->kill();
		}
		clientIsLoggedIn = 0;
		client = 0;
	}
}


int GetNumClientsConnected()
{
	int n = 0;
	for ( int i = 0; i < MAX_CLIENTS; ++i)
	{
		if (clients[i].loggedIn())
			n++;
	}
	return n;
}



void NetInit( char const* pCommandLine ) 
{
	dprintf("Starting NetInit\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	dTimeInit();

	SimpleMsgZocket::registerHandler( -1, handleTerminate );
	SimpleMsgZocket::registerHandler( -2, handleAcceptedConnection );
	SimpleMsgZocket::registerHandler( -3, handleConnected );
	SimpleMsgZocket::registerHandler( msgLogin, handleLogin );
	SimpleMsgZocket::registerHandler( msgLoginConfirm, handleLoginConfirm );
	SimpleMsgZocket::registerHandler( msgClientInput, handleClientInput );
	SimpleMsgZocket::registerHandler( msgAllPlayerInput, handleAllPlayerInput );

	// Determine player...
	g_LocalPlayer = 0;
	char* pPlayer = strstr( pCommandLine, "-player" );
	if ( pPlayer )
	{
		char* pEnd;
		g_LocalPlayer = strtol( pPlayer + strlen("-player"), &pEnd, MAX_CLIENTS );
	}
	else
	{
		g_LocalPlayer = 0;
	}

	if( g_LocalPlayer != 0 ) 
	{
		dprintf("Starting a client...\n");

		startupClient();

	    char remoteName[80];
		char* pIP = strstr( pCommandLine, "-connect" );
		pIP += strlen("-connect") + 1;
		char* pSpace = strchr( pIP, ' ' );
		int len;
		if ( !pSpace )
			len = strlen( pIP ) + 1;
		else
			len = (int)pSpace - (int)pIP;
		memcpy( remoteName, pIP, len );
		remoteName[len] = 0;

		char feh[1024];
		sprintf(feh,"tcp://%s:5050", remoteName );

		serverZocket = new SimpleMsgZocket( feh, NULL, "b" );
		checkZocket(serverZocket);
		serverZocket->appData = (int)&serverZocket;

		Sleep( 1000 );

		dprintf("Trying to connect to server...\n");

		double time_started = GetTime();
		while (!clientIsConnected)
		{
			if ( (GetTime() - time_started) > MAX_CLIENT_WAIT_TO_LOGIN )
			{
				exit(10);
			}

			Sleep(10);
			SimpleMsgZocket::pollAll();
		}
		double actual_time = GetTime() - time_started;
		dprintf("time waited for !clientIsConnected = %f\n",actual_time);

		Sleep( 1000 );

		// Log in to the server
		MsgLogin login;
		login.player = g_LocalPlayer;
		serverZocket->write( &login,1 );
	}
	else
	{
		dprintf("Making a server\n");
		dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

		startupServer();

		startupClient();

		serverZocket = new SimpleMsgZocket( "tcp://127.0.0.1:5050", NULL, "b" );
		checkZocket(serverZocket);
		serverZocket->appData = (int)&serverZocket;
		
		dprintf("Connecting to self...\n");
		dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

		while (!clientIsConnected)
		{
			Sleep(10);
			SimpleMsgZocket::pollAll();
		}

		dprintf("Sending login to server...\n");
		dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

		// Log in to the server
		MsgLogin login;
		login.player = g_LocalPlayer;
		serverZocket->write( &login );
	}

	dprintf("Trying to log in...\n");
	dprintf("zockets in list : %d\n",SimpleMsgZocket::countAll());

	double time_started = GetTime();
	while (!clientIsLoggedIn)
	{
		if ( (GetTime() - time_started) > MAX_CLIENT_WAIT_TO_LOGIN )
		{
			exit(10);
		}

		Sleep(10);
		SimpleMsgZocket::pollAll();
	}
	double actual_time = GetTime() - time_started;
	dprintf("time waited for !clientIsLoggedIn = %f\n",actual_time);
	
	// send server time here the first time
	if ( server )
	{
		assert( s_AllMsgBuffer.m_TotalCount == 0 );
		AddServerTime();
	}

	dprintf("Done with NetInit\n");
}


void ConstructCompositeMessagePacket()
{
	int i;

	//in_input_gather = true;

	bool done = false;
	double time_started = GetTime();
	while( !done )
	{
		Sleep(10);
		SimpleMsgZocket::pollAll();
		
		bool first_input_wait = false;
		done = true;
		for (int i = 0; i < MAX_CLIENTS; ++i)
		{
			if (clients[i].loggedIn() && !clients[i].gotInput)
			{
				done = false;

				if ( ! clients[i].gotFirstInput )
					first_input_wait = true;
			}
		}

		double time_waited = GetTime() - time_started;
		double time_to_wait = MAX_SERVER_WAIT_FOR_CLIENT;
		if ( first_input_wait )
			time_to_wait *= 3.f;
		
		if ( ! done && time_waited > time_to_wait )
		{
			double actual_time = GetTime() - time_started;
			dprintf("killing client due to long wait!\n");
			dprintf("time waited for input from client = %f\n",actual_time);

			// disable any client that didn't give input :

			for (int i = 0; i < MAX_CLIENTS; ++i)
			{
				if (clients[i].loggedIn() && !clients[i].gotInput)
				{
					// kills itself?
//					clients[i].zocket->kill();
					clients[i].zocket = NULL;
				}
			}

			break;
		}
		if ( done && time_waited > 1.0 )
		{
			dprintf("time_waited = %f\n",time_waited);
		}
	}

	//in_input_gather = false;

	// Now re-echo all input out....
	s_AllMsgBuffer.len = sizeof(MsgAllPlayerInput) + s_AllMsgBuffer.m_TotalCount * sizeof(MsgInput_t);
	for ( i = 0; i < MAX_CLIENTS; ++i)
	{
		if (clients[i].loggedIn())
		{
			clients[i].zocket->write( &s_AllMsgBuffer );
		}
	}

	// Blat out the aggregate buffer
	s_AllMsgBuffer.m_TotalCount = 0;
	s_AllMsgBuffer.m_Count[0] = 0;
	s_AllMsgBuffer.m_Count[1] = 0;
	s_AllMsgBuffer.m_Count[2] = 0;
	s_AllMsgBuffer.m_Count[3] = 0;

	for ( i = 0; i < MAX_CLIENTS; ++i)
	{
		if (clients[i].loggedIn())
		{
			clients[i].gotInput = 0;
		}
	}
}

// FIXME: Hack to get player position
void ComputeLocalPlayerPosition(vec3f& pos);

void ProcessNetworkMessages()
{
	if (!g_ProcessNetworkMessages)
		return;

	// process server time :
	{
	MsgInput_t& msg = s_ClientAllMsgBuffer.m_MsgInput[0];
	assert( msg.m_Input == INPUT_SERVER_TIME );
	g_ProcessNetworkMessages( -1, msg.m_Input, msg.m_X, msg.m_Y, msg.m_Z );
	}

	for (int i = 0; i < MAX_PLAYERS; ++i )
	{
		int firstIndex = s_ClientAllMsgBuffer.m_FirstIndex[i];
		for (int j = 0; j < s_ClientAllMsgBuffer.m_Count[i]; ++j)
		{
			MsgInput_t& msg = s_ClientAllMsgBuffer.m_MsgInput[firstIndex + j];
			g_ProcessNetworkMessages( i, msg.m_Input, msg.m_X, msg.m_Y, msg.m_Z );
		}
	}
}

void NetFrame()
{
	if (client && g_PostNetworkMessages)
		g_PostNetworkMessages();

	TransmitClientInput();

	dTimeLatchFrame();

	s_AllInputReceived = false;

	// Wait till I hear from everyone
	if (server)
	{
		ConstructCompositeMessagePacket();
		
		// must do these AFTER ConstructCompositeMessagePacket
		// this is the only time s_AllMsgBuffer.m_TotalCount == 0

		AddServerTime();

		if ( gotNewClients )
		{
			AddServerReset();
			gotNewClients = false;
		}
	}

	if (client)
	{
		// wait until I get the aggregate input
		double time_started = GetTime();
		while (!s_AllInputReceived)
		{
			if ( (GetTime() - time_started) > MAX_SERVER_WAIT_FOR_CLIENT )
			{
				double actual_time = GetTime() - time_started;
				dprintf("killing client due to long wait!\n");
				dprintf("time waited for input from server = %f\n",actual_time);
				exit(10);
			}

			Sleep(10);
			SimpleMsgZocket::pollAll();
		}

		// Iterate over all messages + process them
		ProcessNetworkMessages();
	}

}


void NetShutdown()
{
	if( client ) 
	{
		shutdownClient();
	}

	if( server ) 
	{
		shutdownServer();
	}
}





