#include "sound_internal.h"
#include "tagmap.h"
#include "wav_load.h"

//#include "vorbis_decoder.h"
char *ASSET_DIRECTORY = "data\\";

Sound_Catalog::Sound_Catalog() {
    sounds_string_to_global = new Oneway_Tagmap(DEFAULT_TAGMAP_SIZE, false);
    num_sounds = 0;

    int i;
    for (i = 0; i < MAX_SOUND_EFFECTS; i++) {
        sounds[i] = NULL;
    }
}

Sound_Catalog::~Sound_Catalog() {
    delete sounds_string_to_global;

    int i;
    for (i = 0; i < num_sounds; i++) {
        delete sounds[i];
    }
}

Sound_Data *Sound_Catalog::get_sound(int index) {
    if (index < 0) return NULL;
    if (index >= num_sounds) return NULL;
    return sounds[index];
}

int Sound_Catalog::get_sound_index(char *name) {
    Sound_Data *data = get_sound(name);
    if (data == NULL) return -1;
    return data->get_catalog_index();
}

Sound_Data *Sound_Catalog::get_sound(char *name) {
    Sound_Data *sound = NULL;

    int index;
    bool success = sounds_string_to_global->lookup(name, &index);
    if (success) {
        assert(index >= 0);
        assert(index < MAX_SOUND_EFFECTS);
        assert(index < num_sounds);

        return sounds[index];
    }

    sound = load_linear_sample_array(name);
    return sound;
}

bool load_pcm44(FILE *f, int *size_ret, byte **data_ret) {
    fseek(f, 0, SEEK_END);
    int start_point = 0;
    int end_point = ftell(f);

    int bytes = end_point - start_point;

    fseek(f, 0, SEEK_SET);
    byte *data = new byte[bytes];

    fread(data, bytes, 1, f);

    *size_ret = bytes;
    *data_ret = data;

    return true;
}

#ifdef NOT
void save_pcm(Loader *loader, char *name, char *suffix,
              int size, byte *data) {
    Concatenator c;
    c.add("data/sounds/");
    c.add(name);
    c.add(".");
    c.add(suffix);
    char *filename = c.get_result();

    FILE *f = loader->fopen(filename, "wb");
    free(filename);

    if (f == NULL) return;

    fwrite(data, size, 1, f);
    fclose(f);
}

#endif NOT

bool ogg_load(char *name,
              FILE *f, int *size_ret, byte **data_ret,
              int *nchannels_ret) {

    fseek(f, 0, SEEK_END);
    long end_pos = ftell(f);
    fseek(f, 0, SEEK_SET);
    int total_bytes = end_pos;

    char *data = new char[total_bytes];
    int success = fread(data, total_bytes, 1, f);
    if (success < 1) return false;

    *size_ret = total_bytes;
    *data_ret = (byte *)data;

    *nchannels_ret = 2; // XXX

    return true;
}

Sound_Data *Sound_Catalog::load_linear_sample_array(char *name) {

    // Increment index even if we don't find it, so that sounds
    // match up on both ends even if one guy is missing files.
    // This kind of thing won't be necessary in the future if
    // we have some kind of init-time tagmap sync.

    int index = num_sounds;
    num_sounds++;

    char *filename;
    FILE *f;

    Concatenator c;
    c.add(ASSET_DIRECTORY);
    c.add(name);
    c.add(".wav");
    filename = c.get_result();

    bool success = false;

    int size;
    unsigned char *data;
    int nchannels;

    Sound_Data::Type type;

    f = fopen(filename, "rb");
    if (f) {
        Waveformatex format;
        success = wav_load(f, &format, &size, &data);
        if (success) nchannels = format.nChannels;
        fclose(f);
        type = Sound_Data::LINEAR_SAMPLE_ARRAY;
    } else {
        c.reset();
        c.add(ASSET_DIRECTORY);
        c.add(name);
        c.add(".ogg");
        filename = c.get_result();
        f = fopen(filename, "rb");
        free(filename);

        if (f) {
            success = ogg_load(name, f, &size, &data, &nchannels);
            fclose(f);
            type = Sound_Data::VORBIS_COMPRESSED;
        }
    }

    if (!success) return NULL;

    // The size parameter is in bytes.

    Sound_Data *sound = new Sound_Data(this, index);
    sounds[index] = sound;

    sound->type = type;
    sound->name = strdup(name);
    sound->nchannels = nchannels;

    if (type == Sound_Data::LINEAR_SAMPLE_ARRAY) {
        int nsamples = size / 2;

        sound->nsamples = nsamples;
        sound->samples = (short *)data;
    } else {
        sound->data = (void *)data;
        sound->length_in_bytes = size;
    }

    sounds_string_to_global->add(sound->name, index);
    return sound;
}


Sound_Data::Sound_Data(Sound_Catalog *catalog, int index) {
    my_catalog = catalog;
    my_index_in_catalog = index;

    name = NULL;
    samples = NULL;
    
    type = UNINITIALIZED;
    nsamples = 0;
    nchannels = 1;
}

Sound_Data::~Sound_Data() {
    if (samples) free(samples);
}

int Sound_Data::get_catalog_index() {
    return my_index_in_catalog;
}
