#include "sound_public.h"
#include "sound_internal.h"

extern Sound_Manager *g_sound_manager;

Sound_Manager *sound_init() {
    Sound_Manager *manager = new Sound_Manager();
    manager->init();

    return manager;
}

void sound_play(char *name, float volume, float pan) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    manager->play_sound(name, volume, pan);
}

void sound_preload(char *name) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    manager->catalog->get_sound(name);
}


void sound_shutdown(Sound_Manager *manager) {
    if (!manager) return;
    manager->shutdown();
}



Sound_Channel *sound_channel_create() {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return NULL;

    return manager->create_persistent_channel();
}

void sound_channel_release(Sound_Channel *channel) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    if (channel == NULL) return;
    manager->release_persistent_channel(channel);
}

bool sound_channel_append(Sound_Channel *channel, char *effect_name) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return false;

    if (channel == NULL) return false;

    Sound_Data *sound = manager->catalog->get_sound(effect_name);
    if (sound == NULL) return false;

    manager->acquire_lock();
    manager->append(channel, sound);
    manager->release_lock();

    return true;
}

void sound_channel_play(Sound_Channel *channel) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    if (channel == NULL) return;

    manager->acquire_lock();
    channel->play();
    manager->release_lock();
}

void sound_channel_set_volume(Sound_Channel *channel, float volume) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    if (channel == NULL) return;

    manager->acquire_lock();
    channel->base_volume = volume;
    manager->release_lock();
}

void sound_channel_set_position(Sound_Channel *channel, float x, float y, float z) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    if (channel == NULL) return;

    manager->acquire_lock();
    channel->position_x = x;
    channel->position_y = y;
    channel->position_z = z;
    channel->spatialized = true;
    manager->release_lock();
}

void sound_channel_set_distance_volume_control(Sound_Channel *channel, float distance_min, float distance_max) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    if (channel == NULL) return;

    manager->acquire_lock();
    channel->distance_min = distance_min;
    channel->distance_max = distance_max;
    manager->release_lock();
}

int sound_channel_get_samples_remaining(Sound_Channel *channel) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return 0;

    if (channel == NULL) return 0;

    manager->acquire_lock();
    int result = channel->count_samples();
    manager->release_lock();

    return result;
}

void sound_listener_set_position(float x, float y, float z) {
    Sound_Manager *manager = g_sound_manager;
    if (!manager) return;

    manager->acquire_lock();
    manager->player->listener_x = x;
    manager->player->listener_y = y;
    manager->player->listener_z = z;
    manager->release_lock();
}
