#include "sound_internal.h"

#include <windows.h>
#include <assert.h>

#include "thread.h"


class Thread::Thread_Private {
public:
    HANDLE thread_;
    DWORD thread_id_;
};


// ---------------------------------------------------


Thread::Thread(unsigned long (*tproc)(void *arg)) {
    priv = new Thread::Thread_Private;
    thread_proc = tproc;
    thread_proc_argument = NULL;
}

unsigned long __stdcall thread_middleman(void *arg) {
    Thread *t = (Thread *)arg;
    unsigned long result = (*t->thread_proc)(t->thread_proc_argument);
    return result;
}

bool Thread::start(void* arg) {
	assert(priv);

	thread_proc_argument = arg;

	priv->thread_ = CreateThread(NULL, 
								 0, 
								 thread_middleman,
								 (void *)this,
								 0, 
								 &priv->thread_id_);
	assert(priv->thread_);
  
	return true;
}


bool Thread::stop(void) {
    return false;
}

void Thread::wait_for_exit(void) {
    while (1) {
        DWORD exit_code;
        BOOL success = GetExitCodeThread(priv->thread_, &exit_code);
		if (!success) return;  // Thread does not exist?

		DWORD dont_want = STILL_ACTIVE;
		if (exit_code != dont_want) return;

		os_sleep(50);
    }
}


void Thread::join(void) {
    assert(0);
}






class Mutex::Mutex_Private {
public:
    CRITICAL_SECTION csection;
    bool initialized;
};


Mutex::Mutex() {
    priv = new Mutex::Mutex_Private;
    priv->initialized = false;
}


Mutex::~Mutex() {
    if (priv->initialized) {
        unlock();
		DeleteCriticalSection(&priv->csection);
    }
}


void Mutex::init(bool immediate_lock) {
    InitializeCriticalSection(&priv->csection);
    priv->initialized = true; 
    if (immediate_lock) lock();
}


void Mutex::lock() {
    EnterCriticalSection(&priv->csection);
}


void Mutex::unlock() {
    LeaveCriticalSection(&priv->csection);
}

class Condition::Condition_Private {
public:
    HANDLE event_;
};


Condition::Condition() {
    priv = new Condition::Condition_Private;
    assert(priv);
}


void Condition::init(void) {
    priv->event_ = CreateEvent(NULL, FALSE, FALSE, NULL);
    assert(priv->event_);
}


void Condition::signal(void) {
    BOOL ok = PulseEvent(priv->event_);
    assert(ok);
}


void Condition::wait_for(void) {
    DWORD res = WaitForSingleObject(priv->event_, INFINITE);
    assert(WAIT_FAILED != res);
}

