#include "sound_internal.h"
#include "vorbis_decoder.h"

#include <windows.h>
#include <mmsystem.h>

#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

#pragma warning(disable: 4244)

const int PREFETCH_BLOCK_SIZE = 16384;
const int SLACK_TIME = 250;

#define get_ogg_file() ((OggVorbis_File *)void_pointer_to_ogg_file)

Vorbis_Decoder::Vorbis_Decoder() {
    channels = 0;
    void_pointer_to_ogg_file = NULL;
}

Vorbis_Decoder::~Vorbis_Decoder() {
    OggVorbis_File *ogg_file = get_ogg_file();
    if (ogg_file) delete ogg_file;
}

void Vorbis_Decoder::set_seek_position(int position) {
    // I haven't implemented this yet (don't know if I will need to.)
    assert(0);
}

size_t Vorbis_Decoder::consume_bytes(void *dest, 
                                     size_t size, size_t num_items) {
    assert(my_source != NULL);

    int bytes_to_consume = size * num_items;
    int bytes_available = my_source->length_in_bytes - consumer_index;
    int items_to_consume = num_items;

    if (bytes_to_consume > bytes_available) {
        int items = bytes_available / size;
        bytes_to_consume = items * size;
        items_to_consume = items;
    }

    char *csrc = ((char *)my_source->data) + consumer_index;
    memcpy(dest, csrc, bytes_to_consume);

    consumer_index += bytes_to_consume;
    return items_to_consume;
}

static size_t read_callback(void *ptr, size_t size, size_t nmemb, 
                            void *data_source) {

    Vorbis_Decoder *decoder = (Vorbis_Decoder *)data_source;
    return decoder->consume_bytes(ptr, size, nmemb);
}

static int close_callback(void *data_source) {
    return 0;
}

static int seek_callback(void *data_source, ogg_int64_t offset, int whence) {
    return -1;
}

static long tell_callback(void *data_source) {
    return -1;
}

void Vorbis_Decoder::set_decode_source(Sound_Data *data) {
    assert(data->type == Sound_Data::VORBIS_COMPRESSED);

    my_source = data;
    consumer_index = 0;
}

int Vorbis_Decoder::get_decoded_file_size() {
    OggVorbis_File *ogg_file = (OggVorbis_File *)void_pointer_to_ogg_file;
    assert(ogg_file != NULL);

    int samples = ov_pcm_total(ogg_file, 0);
    return samples * bytes_per_sample;
}

void Vorbis_Decoder::begin_streaming() {
    ov_callbacks callbacks;
    callbacks.read_func = read_callback;
    callbacks.seek_func = seek_callback;
    callbacks.close_func = close_callback;
    callbacks.tell_func = tell_callback;

    OggVorbis_File *ogg_file;
    ogg_file = new OggVorbis_File;
    void_pointer_to_ogg_file = ogg_file;

    int result = ov_open_callbacks(this, ogg_file, NULL, -1,
                                   callbacks);
    assert(result != -1);

    int i;
    for(i = 0;i < ov_streams(ogg_file); i++) {
        vorbis_info *vi = ov_info(ogg_file, i);

        channels = vi->channels;
        rate = vi->rate;
        bytes_per_sample = 2;

        /*
          printf("\tlogical bitstream section %d information:\n",i+1);
          printf("\t\t%ldHz %d channels bitrate %ldkbps serial number=%ld\n",
          vi->rate,vi->channels,ov_bitrate(&ogg_file,i)/1000,
          ov_serialnumber(&ogg_file,i));
          printf("\t\tcompressed length: %ld bytes ",ov_raw_total(&ogg_file,i));
          printf(" play time: %lds\n",(long)ov_time_total(&ogg_file,i));
        */

        break; // Use only first stream.
    }
}

int Vorbis_Decoder::read(char *buf, int bytes) {
    OggVorbis_File *ogg_file = get_ogg_file();

    assert(ogg_file);

    int section;
    long result = ov_read(ogg_file, buf, bytes, 
                          0, bytes_per_sample, 1, &section);
    return (int)result;
}
