#include "sound_internal.h"
#include "file.h"
#include "wav_load.h"

#define WFMTEX_SIZE (2+2+4+4+2+2)

bool load_waveformatex(FILE *f, Waveformatex *w){
    bool ok = true;

    w->wFormatTag       = fread_2b(f,&ok); 
    w->nChannels        = fread_2b(f,&ok); 
    w->nSamplesPerSec   = fread_4b(f,&ok); 
    w->nAvgBytesPerSec  = fread_4b(f,&ok); 
    w->nBlockAlign      = fread_2b(f,&ok); 
    w->wBitsPerSample   = fread_2b(f,&ok); 
    w->cbSize           = 0;

    return ok;
}

bool match_format_type(FILE *f, char *type, bool *ok){
    int i=0;
    char bt;
  
    while(i<4){
        bt = (char) fread_1b(f,ok);
        if(!*ok){
            return false;
        }
        if(bt == type[i]) i++;
        else{
            i=0;
            if(bt == type[i]) i++;
        }
    }
    return true;
}

// seek_riff_chunk()
// 
// this looks for a specfic riff chunks.  It will skip over any 
// other riff chunks that do not match <type>.
// it will consume the ID tag, and the size identification.
// it will not read the data.
//
// inputs:
//   f:    file to deal with
//   type: must be 5 characters long [4 RIFF ids, and one NULL char]
//   size: address of the size of the chunk
// outputs:
//   false: on failure.
//   true : if it did find the chunk etc. etc.
//
// NOTE:  if the type is < 4 letters long, it's gotta be buffered
//        with spaces:  "foo "  NOT "foo"...

bool seek_riff_chunk(FILE *f, char *type, long *size){
    bool ok,match;
    int  res;
    long hunk_size;

    ok = true;

    match = match_format_type(f,type,&ok);

    hunk_size = fread_4b(f,&ok);
    if(!ok) return false;

    if(!match){
        res = fseek(f,hunk_size,SEEK_CUR);
        if(res==-1) return false;
        return seek_riff_chunk(f,type,size);
    }
  
    if(size!=NULL) *size = hunk_size;
    return true;
}
// Same as the old wav_load, except it takes an open FILE* instead of
// trying to use a Loader to open the file itself.

bool wav_load(FILE *fileptr,
              Waveformatex *w,
              int *size,
              unsigned char **raw_data) {

    assert(fileptr);
    assert(w);
    assert(size);
    assert(raw_data);

    FILE *wav_file = fileptr;
    bool ok,match;
    long hunk_size;
    char *mem = NULL, *pos;
    int res;
  
    match = ok = true;

    seek_riff_chunk(wav_file,"RIFF",NULL);
    match_format_type(wav_file,"WAVE",&ok);
    seek_riff_chunk(wav_file,"fmt ",&hunk_size);

    if(hunk_size!=WFMTEX_SIZE) goto bad_exit;

    load_waveformatex(wav_file,w);
    ok = seek_riff_chunk(wav_file,"data",&hunk_size);

    if(!ok) goto bad_exit;

    mem = (char *) malloc(hunk_size);
    if(mem == NULL) return false;

    pos = mem;

    if(size!=NULL) *size = hunk_size;
    if(raw_data!=NULL) *raw_data = (unsigned char *)pos;
  
    while(hunk_size>0){
        res = (int) fread(pos,1,hunk_size,wav_file);
        if(res == 0) break;
        pos   += res;
        hunk_size -= res;
    }

    if(hunk_size!=0) goto bad_exit;

//  good_exit:

    return true;

  bad_exit:

    printf("wav_load.cpp, wav_load_helper badness\n");
    if(mem) free(mem);
    return false;
  
}

