/* ========================================================================
   $RCSfile: statistics.cpp,v $
   $Date: 2002/03/13 00:20:14 $
   $Revision: 1.2 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#include <assert.h>
#include <windows.h>

#if !defined(STATISTICS_H)
#include "statistics.h"
#endif

static counter_statistic IgnoredInitialParent;
counter_statistic *timer_block::CurrentCounter = &IgnoredInitialParent;
static int32 CounterCount = 0;

static int32 const MaximumCounterCount = 512;
static counter_statistic Counters[MaximumCounterCount];

system_clock
GetSystemSeconds(void)
{
    system_clock Clock;
    QueryPerformanceCounter(&Clock.Data);
    
    return(Clock);
}

real32
GetSecondsElapsed(system_clock const &StartClock,
                  system_clock const &EndClock)
{
    static real64 TimerSecondsConversion = 0;
    if(TimerSecondsConversion == 0)
    {
        LARGE_INTEGER Frequency;
        QueryPerformanceFrequency(&Frequency);
        TimerSecondsConversion = 1.0 / (real64)Frequency.QuadPart;
    }

    LARGE_INTEGER const *Start = &StartClock.Data;
    LARGE_INTEGER const *End = &EndClock.Data;

    if(Start->QuadPart < End->QuadPart)
    {
        real64 const Difference = (real64)(End->QuadPart - Start->QuadPart);
        real32 const SecondsElapsed = (real32)(Difference * TimerSecondsConversion);
        return(SecondsElapsed);
    }
    else
    {
        // TODO: Proper code for handling wrapping
        return(0);
    }
}

void
SleepForSeconds(real32 Seconds)
{
    Sleep((int32)(Seconds * 1000.0f));
}

int32
GetCounterCount(void)
{
    return(CounterCount);
}

counter_statistic &
GetCounter(int32 CounterIndex)
{
    assert(CounterIndex < CounterCount);
    assert(Counters);

    return(Counters[CounterIndex]);
}

int64
GetCounterTicksPerSecond(void)
{
    static int64 TicksPerSecond = 0;
    if(TicksPerSecond == 0)
    {
        int64 StartCycles;
        ReadTimeStampCounter(StartCycles);
        system_clock const StartClock = GetSystemSeconds();

        SleepForSeconds(1.0f);

        int64 EndCycles;
        ReadTimeStampCounter(EndCycles);
        system_clock const EndClock = GetSystemSeconds();

        real64 const CycleDelta = (real64)(EndCycles - StartCycles);
        real64 const SecondsDelta =
            (real64)(GetSecondsElapsed(StartClock, EndClock));
        TicksPerSecond = (int64)((CycleDelta / SecondsDelta) + 0.5f);
    }

    return(TicksPerSecond);
}

counter_statistic *
AddCounter(char const *Name)
{
    // Add the counter
    assert(Counters && (CounterCount < MaximumCounterCount));

    counter_statistic &NewCounter = Counters[CounterCount++];
    NewCounter.Name = Name;
    
    return(&NewCounter);
}

void
ResetCounters(void)
{   
    if(Counters)
    {
        {for(int32 CounterIndex = 0;
             CounterIndex < CounterCount;
             ++CounterIndex)
        {
            counter_statistic &CurrentCounter = Counters[CounterIndex];
            Reset(CurrentCounter);
        }}
    }
}

void
ResetCounterPeaks(void)
{   
    if(Counters)
    {
        {for(int32 CounterIndex = 0;
             CounterIndex < CounterCount;
             ++CounterIndex)
        {
            counter_statistic &CurrentCounter = Counters[CounterIndex];
            ResetPeaks(CurrentCounter);
        }}
    }
}

void
GetCounterResults(int32 CounterIndex, counter_results &Results)
{
    counter_statistic &Counter = GetCounter(CounterIndex);

    Results.Name = Counter.Name;
    
    Results.Count = Counter.Count;
    Results.TotalSeconds = ((real64)Counter.TotalTime /
                            (real64)GetCounterTicksPerSecond());
    Results.SecondsWithoutChildren = ((real64)(Counter.TotalTime -
                                               Counter.ChildTime) /
                                      (real64)GetCounterTicksPerSecond());

    Results.PeakCount = Counter.PeakCount;
    Results.PeakTotalSeconds = ((real64)Counter.PeakTotalTime /
                                (real64)GetCounterTicksPerSecond());
    Results.PeakSecondsWithoutChildren = ((real64)Counter.PeakTimeWithoutChildren /
                                          (real64)GetCounterTicksPerSecond());
}
