#if !defined(STATISTICS_H)
/* ========================================================================
   $RCSfile: statistics.h,v $
   $Date: 2002/03/13 00:20:14 $
   $Revision: 1.2 $
   $Creator: Casey Muratori $
   $Notice: $
   ======================================================================== */

#if !defined(GLOBALS_H)
#include "globals.h"
#endif

struct system_clock
{
    LARGE_INTEGER Data;
};
extern system_clock GetSystemSeconds(void);
extern real32 GetSecondsElapsed(system_clock const &StartClock, system_clock const &EndClock);

// X86 timer
inline void ReadTimeStampCounter(__int64 &Value)
{
    __int64 Temp;
    __asm          
    {              
        rdtsc       
        mov dword ptr [Temp],eax
        mov dword ptr [Temp+4],edx
    }
    Value = Temp;
}

#define COUNT_BLOCK(Name) static counter_statistic *Counter__ = AddCounter(Name); timer_block TimerBlock__(Counter__)

struct counter_statistic
{
    char const *Name;
    
    int32 Count;
    int64 TotalTime;
    int64 ChildTime;

    int32 PeakCount;
    int64 PeakTotalTime;
    int64 PeakTimeWithoutChildren;
};

int32 GetCounterCount(void);
counter_statistic &GetCounter(int32 CounterIndex);
int64 GetCounterTicksPerSecond(void);

counter_statistic *AddCounter(char const *Name);
void ResetCounters(void);
void ResetCounterPeaks(void);

struct counter_results
{
    char const *Name;
    
    int32 Count;
    real64 TotalSeconds;
    real64 SecondsWithoutChildren;
    
    int32 PeakCount;
    real64 PeakTotalSeconds;
    real64 PeakSecondsWithoutChildren;
};
void GetCounterResults(int32 CounterIndex, counter_results &Results);

inline void 
AddSample(counter_statistic &Statistic, int64 const dTimer)
{
    ++Statistic.Count;
    Statistic.TotalTime += dTimer;
}

inline void
RemoveChildSample(counter_statistic &Statistic, int64 const dTimer)
{
    Statistic.ChildTime += dTimer;
}

inline void
Reset(counter_statistic &Statistic)
{
    if(Statistic.PeakCount < Statistic.Count)
    {
        Statistic.PeakCount = Statistic.Count;
    }
    
    if(Statistic.PeakTotalTime < Statistic.TotalTime)
    {
        Statistic.PeakTotalTime = Statistic.TotalTime;
    }

    int64 TimeWithoutChildren = Statistic.TotalTime - Statistic.ChildTime;
    if(Statistic.PeakTimeWithoutChildren < TimeWithoutChildren)
    {
        Statistic.PeakTimeWithoutChildren = TimeWithoutChildren;
    }

    Statistic.Count = 0;
    Statistic.TotalTime = 0;
    Statistic.ChildTime = 0;
}

inline void
ResetPeaks(counter_statistic &Statistic)
{
    Statistic.PeakCount = 0;
    Statistic.PeakTotalTime = 0;
    Statistic.PeakTimeWithoutChildren = 0;
}

struct timer_block
{
    static counter_statistic *CurrentCounter;
    
    counter_statistic *ParentCounter;
    int64 StartClocks;
    
    inline timer_block(counter_statistic *Counter)
    {
        // Remember who our parent was
        ParentCounter = CurrentCounter;
        
        // Make ourselves the active counter
        CurrentCounter = Counter;
        
        // Get the starting clock time
        ReadTimeStampCounter(StartClocks);
    }
    
    inline ~timer_block(void)
    {
        // Get the finishing clock time and compute the delta
        int64 EndClocks;
        ReadTimeStampCounter(EndClocks);
        int64 const dCounter = EndClocks - StartClocks;

        // Add the elapsed time to our counter
        AddSample(*CurrentCounter, dCounter);

        // Remove the elapsed time from our parent counter and make
        // our parent the current counter again
        CurrentCounter = ParentCounter;
        RemoveChildSample(*CurrentCounter, dCounter);
    }
};

#define STATISTICS_H
#endif
