/*----------------------------------------------------------------------------

  system.cpp - windows user interface stuff for Indie Game Jam
  
*/

#include <windows.h>
#include <windowsx.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <dinput.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "system.h"
#include "guys.h"
#include "hf.h"
#include "camera.h"
#include "app_camera.h"
#include "sound/sound_public.h"
#include "math3d.h"  // yet another vector library!!!
#include "statistics.h"
#include "net.h"

// some globals for the window

int ScreenWidth = 640;
int ScreenHeight = 480;
int WindowWidth = 640;
int WindowHeight = 480;
char AppName[] = "Indie Game Jam 0 - The Dueling Machine";

HWND WindowHandle;
HDC WindowDC;
HGLRC WindowRC;
LPDIRECTINPUT2 dinput;
LPDIRECTINPUTDEVICE2 di_gamepad;
DIJOYSTATE2 di_gamepad_state;
bool di_gamepad_state_valid = false;
LPDIRECTINPUTDEVICE2 di_mouse;
DIMOUSESTATE di_mouse_state;
bool di_mouse_state_valid = false;

int gShowStats=0;

vec3f playerCursorPos;

static bool
KeyState(int KeyCode)
{
    int const DownBit = (1 << 15);
    return((GetKeyState(KeyCode) & DownBit) == DownBit);
}


void cameraSlewSim(float simtime);
void cameraOrbitSim(float simtime);
void cameraFollowSim(float simtime);
void cameraFirstPersonSim(float simtime);
void setupCamera(camera &Camera);
void openglInit(int width, int height);


// TODO: Hey guys, move this into the app when you start doing
// that stuff!
float CameraTarget[3];
float OrbitHeight = 0.f;
camera_mode GlobalCameraMode = DEFAULT_CAMERA_MODE;
bool DrawCameraCrossHairs = false;
float CrossHairsColor[3] = {1,1,1};
bool DrawOrbitHeight = false;
Guy *FollowCamGuy = 0;
static vec3f vel, ang_vel;
void FollowCamCallback(Guy *guy, float d2, void *parm) {
    guysUnregisterRef(&FollowCamGuy);
    guysRegisterRef(guy,&FollowCamGuy);
}
void AppCameraInit( camera_mode mode, bool InitPosOff)
{
    if(mode == DEFAULT_CAMERA_MODE) {
        char *env = getenv("IGJ_CAMERA_MODE");
        if(env) {
            mode = (camera_mode)atoi(env);
        }
    }

    // sean wins all arguments, but not anymore!
    if(mode == DEFAULT_CAMERA_MODE) {
        mode = MOUSE_ORBIT_MODE;
    }

    assert(mode > DEFAULT_CAMERA_MODE && mode < MAX_CAMERA_MODE);
    GlobalCameraMode = mode;

    if(InitPosOff) {
        InitializeDefaultCamera(*GlobalCamera);
        GlobalCamera->Orientation[0] = sin(0.25f*3.1415926f);
        GlobalCamera->Orientation[3] = cos(0.25f*3.1415926f);
    }
    vel.x = vel.y = vel.z = ang_vel.x = ang_vel.y = ang_vel.z = 0.;
    switch(mode) {
        case SEAN_SLEW_MODE:
            if(InitPosOff) {
                GlobalCamera->Offset[2] = 0.0f;
                GlobalCamera->Position[2] = 100.0;
            }
            DrawCameraCrossHairs = false;
            break;
        case FOLLOW_ORBIT_MODE: {
            float x = playerCursorPos.x;
            float y = playerCursorPos.y;
            float z = playerCursorPos.z;
            findGuysInSphere(x, y, z, 5, FollowCamCallback, NULL);
            if(!FollowCamGuy) {
                // widen our search
                findGuysInSphere(x, y, z, 30, FollowCamCallback, NULL);
                if(!FollowCamGuy) {
                    FollowCamCallback(gGuys + gGuyCount - 1,0,0);
                }
            }
            // fall through
        }
        case MOUSE_ORBIT_MODE:
            if(InitPosOff) {
                GlobalCamera->Offset[2] = 300.0f;
                GlobalCamera->Position[2] = 0.0;
                OrbitHeight = 0.f;
            }
//            DrawCameraCrossHairs = true;
			DrawCameraCrossHairs = false;
            break;
	    case FIRST_PERSON_MODE:
//			GlobalCamera->Offset[0
			DrawCameraCrossHairs = false;
			break;
        default:
            assert(!"what the fuck? buggy camera?");
            GlobalCamera->Offset[2] = 100.0f;
            GlobalCamera->Position[2] = 100.0;
            break;
    }
    if(InitPosOff) {
        GlobalCamera->Position[0] = 500.;
        GlobalCamera->Position[1] = 500.;
        GlobalCamera->NearClipPlane = 1.0f;
        GlobalCamera->FarClipPlane = 10000.0f;
        // elevation is postive around the x axis
        GlobalCamera->EAR[0] -= 3.14f / 6.0f;
    }
}

void renderBegin(void)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   
   assert(GlobalCamera);
   setupCamera(*GlobalCamera);
}

void renderEnd(void)
{
    glFinish();  // @todo taking this out makes the app 20% slower on checkerath?!?!?
    SwapBuffers(WindowDC);
}

static double prevTime;

#define LARGEST_ALLOWED_TIMESTEP   0.125

void AppSetupFontDraw(float x, float y)
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glOrtho(0, (real32)WindowWidth, 0, (real32)WindowHeight, -1.0f, 1.0f);

   glEnable(GL_TEXTURE_2D);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glEnable(GL_BLEND);
   glTranslatef( x, y, 0.0f );
   glColor3f(0.8f, 0.8f, 0.8f);
}

// this uses the current glColor because I didn't want to write the 4 permutations
void DrawCrossHairs( float o[3], float normal[3], float length )
{
    glPushMatrix();
    glTranslatef(o[0],o[1],o[2]);
    vector_3 n = normal;
    vector_3 z(0.,0.,1.);
    vector_3 v = CrossProduct(n,z);
    glRotatef(DegreesFrom(acos(DotProduct(n,z))),v(0),v(1),v(2));
    glBegin(GL_LINES);
    glVertex3f(-length,0,0); glVertex3f(length,0,0);
    glVertex3f(0,-length,0); glVertex3f(0,length,0);
    glVertex3f(0,0,-length); glVertex3f(0,0,length);
    glEnd();
    glPopMatrix();
}

#define RUBBER_BAND_ELEVATION  3

void DrawGroundLine(float x0, float y0, float x1, float y1)
{
   float len = sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0));
   int steps = floor(len/HF_SCALE*2+1.5)+1;
   glVertex3f(x0, y0, ground(x0,y0)+RUBBER_BAND_ELEVATION);

   // if steps == 3, then make 3 pieces,
   // which is 4 points, so loop over
   // *2* mid points, each spaced at 1/3
   for (int i=1; i < steps; ++i) {
      float x = x0 + (x1-x0)*i/steps;
      float y = y0 + (y1-y0)*i/steps;
      float z = ground(x,y) + RUBBER_BAND_ELEVATION;
      glVertex3f(x,y,z);
      glVertex3f(x,y,z);
   }
   glVertex3f(x1, y1, ground(x1,y1)+RUBBER_BAND_ELEVATION);
}

void DrawWorldRect(vec3f *a, vec3f *b, vec3f *horz)
{
   vec3f width = *horz;
   width.z = 0;
   vec3f_normeq(&width);
   vec3f height( -width.y, width.x, 0 );

   // now we need the two opposite points to a and b, such that
   //    c = a + width*s = b + height*t
   //    d = a - height*t = b - width*s
   //
   // i.e. d is at the intersection of a-height*t and b-width*s

   // a->x + s*width.x - b->x - t*height.x = 0
   // a->y + s*width.y - b->y - t*height.y = 0
   // width.x * s + -height.x * t = b->x - a->x
   // width.y * s + -height.y * t = b->y - a->y
   
   float s,t, denom;
   denom = width.x*-height.y - width.y*-height.x;

   s = ((b->x-a->x)*-height.y - (b->y-a->y)*-height.x) / denom;
   t = (-(b->x-a->x)*width.y  + (b->y-a->y)*width.x) / denom;

   vec3f c,d;

   c.x = a->x + s*width.x;
   c.y = a->y + s*width.y;

#if 0
   d.x = b->x - s*width.x;
   d.y = b->y - s*width.y;
#else
   d.x = a->x - t*height.x;
   d.y = a->y - t*height.y;
#endif
   glBegin(GL_LINES);
   DrawGroundLine(a->x, a->y, c.x, c.y);
   DrawGroundLine(a->x, a->y, d.x, d.y);
   DrawGroundLine(c.x, c.y, b->x, b->y);
   DrawGroundLine(d.x, d.y, b->x, b->y);
   glEnd();
}

static vec3f rect_a, rect_b;
void ComputeWorldRect(vec3f *a, vec3f *b, vec3f *horz)
{
   vec3f width = *horz;
   width.z = 0;
   vec3f_normeq(&width);
   vec3f height( -width.y, width.x, 0 );
 
   float s,t, denom;
   denom = width.x*-height.y - width.y*-height.x;

   s = ((b->x-a->x)*-height.y - (b->y-a->y)*-height.x) / denom;
   t = (-(b->x-a->x)*width.y  + (b->y-a->y)*width.x) / denom;

   rect_a.x = s*width.x;
   rect_a.y = s*width.y;

   rect_b.x = -t*height.x;
   rect_b.y = -t*height.y;
}

static real64 counter_vals[16];
void TickCounters(void)
{
   int cnt=GetCounterCount();
   for (int sc=0; sc<cnt; sc++)
   {
      counter_results cres;
      GetCounterResults(sc,cres);
      counter_vals[sc]+=cres.TotalSeconds;
   }
}

void PrintCounters(void)
{
   int cnt=GetCounterCount();
   FILE *fp=fopen("Igj0PerfData.out_no_conflict_yea","w");
   for (int sc=0; sc<cnt; sc++)
   {
      counter_results cres;
      GetCounterResults(sc,cres);
      fprintf(fp,"Counter %s : %lf\n",cres.Name,counter_vals[sc]);
   }
   fclose(fp);
}

extern void clearFindRegions(void);

static bool clickloc_valid = false;
static vec3f clickloc;

static bool rect_dragging;
static vec3f rect_start;
static vec3f rect_end;
static vec3f rect_horz;

extern float ServerDeltaTime();


void AppIdle( void )
{
	// only Player_ProcessInput gets local time
	{
		static double prevTime = GetTime() - (1.f/20);

	   double thisTime = GetTime();
	   double elapsed = thisTime - prevTime;
	   prevTime = thisTime;

	   if (elapsed > LARGEST_ALLOWED_TIMESTEP)
		  elapsed = LARGEST_ALLOWED_TIMESTEP;

	   extern void Player_ProcessInput(float, float);
	   Player_ProcessInput(thisTime, float(elapsed));
	}

   // Network update
   NetFrame();

   double thisTime = g_server_time;
   double elapsed = ServerDeltaTime();	// thisTime - prevTime;
   if (elapsed > LARGEST_ALLOWED_TIMESTEP)
      elapsed = LARGEST_ALLOWED_TIMESTEP;

   // poll the gamepad
   if(di_gamepad) {
       HRESULT hr = di_gamepad->Poll();
       hr = di_gamepad->Acquire();
       while(hr == DIERR_INPUTLOST) {
           hr = di_gamepad->Acquire();
       }
       if(!FAILED(hr)) {
           hr = di_gamepad->GetDeviceState(sizeof(di_gamepad_state),&di_gamepad_state);
           assert(!FAILED(hr));
       }
       if(!FAILED(hr)) {
           di_gamepad_state_valid = true;
       } else {
           di_gamepad_state_valid = false;
       }
   }
   // and the mouse
   if(di_mouse) {
       HRESULT hr = di_mouse->Poll();
       hr = di_mouse->Acquire();
       while(hr == DIERR_INPUTLOST) {
           hr = di_mouse->Acquire();
       }
       if(!FAILED(hr)) {
           hr = di_mouse->GetDeviceState(sizeof(di_mouse_state),&di_mouse_state);
           assert(!FAILED(hr));
       }
       if(!FAILED(hr)) {
           di_mouse_state_valid = true;
       } else {
           di_mouse_state_valid = false;
       }
   }
   
   clearFindRegions();
   if (SystemHooks.GameTick)
      (*SystemHooks.GameTick)(thisTime, elapsed);

   // run all collision detection systems
   findCollisions(elapsed);

   switch(GlobalCameraMode) {
       case SEAN_SLEW_MODE:
           cameraSlewSim(elapsed);
           break;
       case MOUSE_ORBIT_MODE:
           cameraOrbitSim(elapsed);
           break;
       case FOLLOW_ORBIT_MODE:
           cameraFollowSim(elapsed);
           break;
       case FIRST_PERSON_MODE:
		   cameraFirstPersonSim(elapsed);
		   break;
   }

   EngineUpdate();

   renderBegin();

#if 0
   // color coded axes, for your convenience
   glBegin(GL_LINES);
   glColor3f(1,0,0);   glVertex3f(0,0,0);   glVertex3f(100,0,0);
   glColor3f(0,1,0);   glVertex3f(0,0,0);   glVertex3f(0,100,0);
   glColor3f(0,0,1);   glVertex3f(0,0,0);   glVertex3f(0,0,100);
   glEnd();
#endif

   // camera cross hairs
   if(DrawCameraCrossHairs) {
      float x = GlobalCamera->Position[0], y = GlobalCamera->Position[1];
      float z = ground(x,y);
      vector_3 p(GlobalCamera->Position); p(2) = z;
      // p is the spot on the ground - why do we have only _3_ vector classes
      playerCursorPos.x=p(0);
      playerCursorPos.y=p(1);
      playerCursorPos.z=p(2);
      vector_3 b = CameraBackward();
      vector_3 h = CameraHorizontal();
      vector_3 o = GlobalCamera->Offset;
      glColor3fv(CrossHairsColor);
      float s = 600.0/min(o.Length(),200.0);
      glLineWidth(s);
      glDisable(GL_TEXTURE_2D);
      float d[3] = {0.,0.,1.};
      ground_slope(x,y,d+0,d+1);
      vector_3 n = vector_3(d).Normalize();
      float const l = 10.;
      DrawCrossHairs((p+0.5*n).GetElements(),n.GetElements(),l);
      glLineWidth(1.);
      glColor3f(0.5f,0.5f,0.5f);
      if(DrawOrbitHeight && (OrbitHeight > l)) {
          glBegin(GL_LINES);
          // @todo still overlaps by half the linewidth at top
          glVertex3fv((p+vector_3(0.,0.,l)).GetElements());
          glVertex3fv((p+vector_3(0.,0.,OrbitHeight)).GetElements());
          glEnd();
      }
   }

   if (rect_dragging) {
      glLineWidth(5.);
      glDisable(GL_LIGHTING);
      glDisable(GL_TEXTURE_2D);
      glColor3f(1.0f,1.0f,1.0f);
      DrawWorldRect(&rect_start, &rect_end, &rect_horz);
      glLineWidth(1.);
   }

   guysDraw(elapsed);

   if (SystemHooks.GameRender)
      (*SystemHooks.GameRender)();

//   DrawTimingBars();

   // update timing info
   TickCounters();
//   if (gShowStats)
//      ShowCounters();
   ResetCounters();

   renderEnd();

   prevTime = thisTime;
}

// AppWndProc - process all the events

// but first, some camera crap
// camera variables stuff...overloaded based on what camera is active
#define SLEW_ANG_STEP 2.5
#define SLEW_LOC_STEP 5.0
void cameraSlewSim(float simtime)
{
   MoveCameraRelative(*GlobalCamera, simtime*vel.x, simtime*vel.z, -simtime*vel.y);
   MouseOrbitCamera(*GlobalCamera, simtime*ang_vel.z, -simtime*ang_vel.x);
}
#define ORB_LOC_STEP 200.0
#define ORB_FLOOR_HEIGHT 5.0
// mouse movement stuff
float orb_dx = 0.f, orb_dy = 0.f, orb_dtx = 0.f, orb_dty = 0.f, orb_dzoom = 0.f, orb_dheight = 0.f;
bool orb_flooring_camera = false;
void Clip( float &a, float const min, float const max )
{
    if(a < min) {
        a = min;
    } else if (a > max) {
        a = max;
    }
}
void cameraMouseOrbitSim( float simtime, bool SetPosition )
{
    // do gamepad movement
    if(di_gamepad && di_gamepad_state_valid) {
        float s = 1.0;
        if(di_gamepad_state.rgbButtons[1] & 0x80) {  // B button is down
            s = 5.0;
        }
        if(di_gamepad_state.rgbButtons[0] & 0x80) {  // A button is down
            // @todo needs debouncing
            //BallChar('b');  // bowl  - @todo: real App pad callback
        }
        if(di_gamepad_state.rgbButtons[3] & 0x80) {  // D button is down
            if(di_gamepad_state.rgbButtons[4] & 0x80) {  // L button is down
                orb_dzoom += 3*s;
            }
            if(di_gamepad_state.rgbButtons[5] & 0x80) {  // R button is down
                orb_dzoom += -3*s;
            }
            float x = di_gamepad_state.lX/32768.0 - 1.0;
            float y = 1.0 - di_gamepad_state.lY/32768.0;
            orb_dx = 3*s*x;
            orb_dy = 3*s*y;
        } else {
            if(di_gamepad_state.rgbButtons[4] & 0x80) {  // L button is down
                orb_dx += 3*s;
            }
            if(di_gamepad_state.rgbButtons[5] & 0x80) {  // R button is down
                orb_dx += -3*s;
            }
            float x = di_gamepad_state.lX/32768.0 - 1.0;
            float y = 1.0 - di_gamepad_state.lY/32768.0;
            orb_dtx = s*0.125*x*ORB_LOC_STEP;
            orb_dty = s*0.125*y*ORB_LOC_STEP;
        }
    }
    // do mouse movement
    if(di_mouse && di_mouse_state_valid && !rect_dragging) {
        int dX = di_mouse_state.lX;
        int dY = -di_mouse_state.lY;
        if(KeyState(VK_LBUTTON) && !KeyState(VK_RBUTTON)) {
            if(KeyState(VK_CONTROL)) {
                orb_dtx = 3*dX;
                orb_dty = 3*dY;
            } else {
                orb_dx = dX;
                orb_dy = dY;
            }
        } else if(KeyState(VK_RBUTTON) && !KeyState(VK_LBUTTON)) {
            if(KeyState(VK_CONTROL)) {
                orb_dtx = dX;
                orb_dty = dY;
            } else {
                orb_dzoom = dY;
            }
        } else if(KeyState(VK_MBUTTON) || (KeyState(VK_LBUTTON) && KeyState(VK_RBUTTON))) {
            orb_dheight = dY;
        }
    }

    if(SetPosition) {
        MouseMoveCameraPlaneRelative(*GlobalCamera, simtime*vel.x + orb_dtx, simtime*vel.y + orb_dty);
        GlobalCamera->Position[2] = ground(GlobalCamera->Position[0],GlobalCamera->Position[1]);
    }

    OrbitHeight = max(0,OrbitHeight+orb_dheight);
    GlobalCamera->Position[2] += OrbitHeight;
    MouseZoomCamera(*GlobalCamera, orb_dzoom);
    MouseOrbitCamera(*GlobalCamera, orb_dx, orb_dy);
    
    // Stop gliding along the ground if the user tries to rotate up at all
    if(orb_dy > 0.)
    {
        orb_flooring_camera = false;
    }

    // don't let the camera go out of the positive half-space
    // @todo we clip to 45 deg below because we don't want the iterator below to be given 0
    Clip(GlobalCamera->EAR[0],-Pi32/2.0,Pi32/4.0);
    // reset the mouse movement stuff every sim
    orb_dzoom = orb_dx = orb_dy = orb_dtx = orb_dty = orb_dheight = 0.f;  

    // now floor the camera if it's underground
    bool Underground = true;
    int i = 0;
    // @todo this should start from a smarter place...or not iterate
    // @todo it currently costs measurable perf (7% on checkerath?)
    // @todo it also doesn't terminate if you fly the camera underground :)
    while(Underground && i++ < 100)
    {
        float c[3];
        BuildCameraMatrices(*GlobalCamera);
        GetCameraLocation(*GlobalCamera,c);
        float gz = ground(c[0],c[1]);

        Underground = (c[2] < (gz + ORB_FLOOR_HEIGHT));
        if(orb_flooring_camera || Underground)
        {
            float FloorHeightEpsilon = 0.1f;
            float x = sqrt(sqr(c[0] - GlobalCamera->Position[0]) + sqr(c[1] - GlobalCamera->Position[1]));
            float y = FloorHeightEpsilon + gz + ORB_FLOOR_HEIGHT - GlobalCamera->Position[2];
            GlobalCamera->EAR[0] = -atan2(y,x);
            orb_flooring_camera = true;
        }
    }
}

void cameraOrbitSim(float simtime)
{
    cameraMouseOrbitSim(simtime,true);
}

void cameraFollowSim(float simtime)
{
    if(!FollowCamGuy) {
        FollowCamCallback(gGuys + gGuyCount - 1,0,0);
    }
    assert(FollowCamGuy);
    GlobalCamera->Position[0] = FollowCamGuy->loc.x;
    GlobalCamera->Position[1] = FollowCamGuy->loc.y;
    GlobalCamera->Position[2] = FollowCamGuy->loc.z;
    cameraMouseOrbitSim(simtime,false);
}

void cameraFirstPersonSim(float simtime)
{
	// App should hack upon GlobalCamera directly
#if 0
    if(!FollowCamGuy) {
        FollowCamCallback(gGuys + gGuyCount - 1,0,0);
    }
    assert(FollowCamGuy);
    GlobalCamera->Position[0] = FollowCamGuy->loc.x;
    GlobalCamera->Position[1] = FollowCamGuy->loc.y;
    GlobalCamera->Position[2] = FollowCamGuy->loc.z;

	GlobalCamera->
//    cameraMouseOrbitSim(simtime,false);
#endif // 0
}

static void startdrag(int x, int y)
{
   // start dragging if it's a valid click
   rect_dragging = pickHF(&rect_start, x, y);
   rect_end = rect_start;
   rect_horz = *(vec3f*)CameraHorizontal();
}

static void movedrag(int x, int y)
{
   pickHF(&rect_end, x, y);
}

static void makeProphet(Guy *g, void *p)
{
   vec3f temp;
   vec3f_sub(&temp, &g->loc, &rect_start);
   float s,t;
   s = vec3f_dot(&temp, &rect_a);
   t = vec3f_dot(&temp, &rect_b);
   if (s >= 0 && s <= 1 && t >= 0 && t <= 1) {
      g->debug_color = (g->debug_color+1) & 7;
   }
}

static void enddrag(int x, int y)
{
   // leave end location where the last move put it,
   // to avoid mouse-jerk-on-release
   rect_dragging = FALSE;
   float temp;
   if (rect_start.x > rect_end.x) { temp = rect_start.x; rect_start.x = rect_end.x; rect_end.x = temp; }
   if (rect_start.y > rect_end.y) { temp = rect_start.y; rect_start.y = rect_end.y; rect_end.y = temp; }
   ComputeWorldRect(&rect_start, &rect_end, &rect_horz);
   // now rect_a, rect_b are rotated edges
   // find the bounding box of the new rectangle
   vec3f a,b;
   a = rect_start;
   b = rect_end;
   if (rect_a.x < rect_b.x) {
      if (rect_a.x < 0) a.x += rect_a.x;
      if (rect_b.x > 0) b.x += rect_b.x;
   } else {
      if (rect_b.x < 0) a.x += rect_b.x;
      if (rect_a.x > 0) b.x += rect_a.x;
   }
   if (rect_a.y < rect_b.y) {
      if (rect_a.y < 0) a.y += rect_a.y;
      if (rect_b.y > 0) b.y += rect_b.y;
   } else {
      if (rect_b.y < 0) a.y += rect_b.y;
      if (rect_a.y > 0) b.y += rect_a.y;
   }

   float d = vec3f_mag2(&rect_a);
   vec3f_scaleeq(&rect_a, 1.0/d);
   d = vec3f_mag2(&rect_b);
   vec3f_scaleeq(&rect_b, 1.0/d);

   findGuysInSquare(a.x, a.y, b.x, b.y, makeProphet, NULL);
}

LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        case WM_CHAR:
        {

            // you need to check a couple different places in this
            // WM_CHAR (and WM_KEYDOWN) case to know what keys are
            // available, and I figured it was best to keep them right
            // here.  yes, this is ass
            bool KeyUsedByCamera = false;   // don't even talk to me about it
            switch(GlobalCameraMode) {
                case SEAN_SLEW_MODE: {
                    KeyUsedByCamera = true;
                    switch(wParam) {
                        case 'Q': ang_vel.y += SLEW_ANG_STEP; break;
                        case 'E': ang_vel.y -= SLEW_ANG_STEP; break;

                        case 'a': ang_vel.z += SLEW_ANG_STEP; break;
                        case 'd': ang_vel.z -= SLEW_ANG_STEP; break;
                        case 'r': ang_vel.x += SLEW_ANG_STEP; break;
                        case 'v': ang_vel.x -= SLEW_ANG_STEP; break;

                        case 'w': vel.y += SLEW_LOC_STEP; break;
                        case 'x': vel.y -= SLEW_LOC_STEP; break;
                        case 'z': vel.x -= SLEW_LOC_STEP; break;
                        case 'c': vel.x += SLEW_LOC_STEP; break;
                        case 'q': vel.z += SLEW_LOC_STEP; break;
                        case 'e': vel.z -= SLEW_LOC_STEP; break;
                   
                        case '\r': AppCameraInit(DEFAULT_CAMERA_MODE);
                        case ' ': vel.x = vel.y = vel.z = 0;
                            ang_vel.x = ang_vel.y = ang_vel.z = 0;
                            break;

                        default:
                            KeyUsedByCamera = false;
                            break;
                    }
                    break;
                }
                case MOUSE_ORBIT_MODE: {
                    // look in WM_KEYDOWN/WM_KEYUP
                    break;
                }
            }
            if(KeyUsedByCamera) {
                break;
            }

            if (SystemHooks.GameChar)
               if ((*SystemHooks.GameChar)(wParam))
                  break;

            // the non-camera keyboard stuff
            switch(wParam) {
                // Sound Test
                case '1': {
                    sound_play("sound181", 1.0f);
                    break;
                }
                case '2': {
                    sound_play("sound19", 1.0f);
                    break;
                }
                case '3': {
                    sound_play("sound20", 1.0f);
                    break;
                }
                case '4': {
                    sound_play("sound21", 1.0f);
                    break;
                }
                case '5': {
                    sound_play("sound22", 1.0f);
                    break;
                }
                case '8': {
                    // I am going to leak this for now since I don't want to remember it
                    Sound_Channel *channel = sound_channel_create();
                    if (channel) {
                        sound_channel_append(channel, "bowling");
                        sound_channel_set_volume(channel, 1.0f);
                        sound_channel_play(channel);
                    }
                    break;
                }
                case '9': {
                    sound_play("bowling", 1.0f);
                    break;
                }
                case '0': {
                    EngineSoundInit();
                    break;
                }
                default: guysKey(wParam);
            }
            break;
        }
        case WM_KEYDOWN:
        {
            bool KeyUsedByCamera = false;   // don't even talk to me about it
            switch(GlobalCameraMode) {
                case MOUSE_ORBIT_MODE: {
                    KeyUsedByCamera = true;
                    switch(wParam) {  // could check the repeat bit, but whatever
                        case 'A': vel.x = -ORB_LOC_STEP; break;
                        case 'D': vel.x =  ORB_LOC_STEP; break;

                        case 'W': vel.y =  ORB_LOC_STEP; break;
                        case 'X': // fallthrough, because I love you enough to make both x and s work for you, dc
                        case 'S': vel.y = -ORB_LOC_STEP; break;

                        default:
                            KeyUsedByCamera = false;
                            break;
                    }
                    break;
                }
            }
            if(KeyUsedByCamera) {
                break;
            }

            // the non-camera keyboard stuff
            switch(wParam)
            {
                case VK_ESCAPE:
                    PostMessage(hwnd,WM_CLOSE,0,0L);
                    break;            

                case VK_F1:
                    AppCameraInit(SEAN_SLEW_MODE);
                    break;
                case VK_F2:
                    AppCameraInit(MOUSE_ORBIT_MODE);
                    break;
                case VK_F3:
                    AppCameraInit(FOLLOW_ORBIT_MODE);
                    break;
                case 'F': {
                    if(!KeyState(VK_CONTROL)) {
                        if(GlobalCameraMode == FOLLOW_ORBIT_MODE) {
                            AppCameraInit(MOUSE_ORBIT_MODE,false);
                        } else
                        if(GlobalCameraMode == MOUSE_ORBIT_MODE) {
                            AppCameraInit(FOLLOW_ORBIT_MODE,false);
                        }
                    }
                    break;
                }

                case 'S':
                    break;
            }
            break;
        }
        case WM_KEYUP:
        {
            bool KeyUsedByCamera = false;   // don't even talk to me about it
            switch(GlobalCameraMode) {
                case MOUSE_ORBIT_MODE: {
                    KeyUsedByCamera = true;
                    switch(wParam) {  // could check the repeat bit, but whatever
                        case 'A': vel.x = 0.; break;
                        case 'D': vel.x = 0.; break;

                        case 'W': vel.y = 0.; break;
                        case 'X': // fallthrough, because I love you enough to make both x and s work for you, dc
                        case 'S': vel.y = 0.; break;

                        default:
                            KeyUsedByCamera = false;
                            break;
                    }
                    break;
                }
            }
            if(KeyUsedByCamera) {
                break;
            }

            // the non-camera keyboard stuff
            switch(wParam)
            {
                case VK_LEFT:
                    break;
            }
            break;
        }

        case WM_LBUTTONDOWN:
        {
            if ((GlobalCameraMode == SEAN_SLEW_MODE) || KeyState(VK_SHIFT)) {
               startdrag(LOWORD(lParam), HIWORD(lParam));
            }
            break;
        }

        case WM_LBUTTONUP:
        {
            if (rect_dragging)
               enddrag(LOWORD(lParam), HIWORD(lParam));
            break;
        }

        case WM_MOUSEMOVE:
        {
            RECT ClientRectangle;
            GetClientRect(hwnd, &ClientRectangle);

            if (rect_dragging) {
               movedrag(LOWORD(lParam), HIWORD(lParam));
               break;
            }

            real32 X = (real32)GET_X_LPARAM(lParam);
            real32 Y = ((real32)ClientRectangle.bottom - 1) -
                (real32)GET_Y_LPARAM(lParam);
            
            static real32 LastX = X;
            static real32 LastY = Y;

            real32 dX = X - LastX;
            real32 dY = Y - LastY;
            switch(GlobalCameraMode) {
                case SEAN_SLEW_MODE: {
                    break;
                }
                case FOLLOW_ORBIT_MODE:
                case MOUSE_ORBIT_MODE:
                    // these are handled with dinput for mouse capture
                    // in the orbit sim loop
                    break;
            }

            LastX = X;
            LastY = Y;
            
            break;
        }

        // probably don't need to go below here
        
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        case WM_CLOSE:
            break;

        case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd,&ps);
            EndPaint(hwnd,&ps);
            return 0L;
        }

        case WM_SIZE:
        {
            RECT Rect;
            GetClientRect(hwnd, &Rect);
            WindowWidth = Rect.right;
            WindowHeight = Rect.bottom;
            break;
        }
    }
    return DefWindowProc(hwnd,msg,wParam,lParam);
}

// WinMain - create everything

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    AppCameraInit(DEFAULT_CAMERA_MODE);
    EngineInit();

    // register the window class
    if (!hPrev)
    {
        WNDCLASS cls;
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
        return FALSE;
    }

    // size the window correctly for the border
    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    
    AdjustWindowRect(&WindowRect,Style,TRUE);

    // create window
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                CW_USEDEFAULT,0,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);
    assert(WindowHandle);

	// Initialize networking
	NetInit( szCmdLine );

    // initialize DirectInput
    HRESULT hr;
    hr = DirectInputCreate(GetModuleHandle(0),DIRECTINPUT_VERSION,(LPDIRECTINPUT*)&dinput,0);    assert(dinput && !FAILED(hr));
    hr = dinput->CreateDevice(GUID_Joystick,(LPDIRECTINPUTDEVICE*)&di_gamepad,0);
    if(di_gamepad) {
        hr = di_gamepad->SetDataFormat(&c_dfDIJoystick2);  assert(!FAILED(hr));
        hr = di_gamepad->SetCooperativeLevel(WindowHandle,DISCL_EXCLUSIVE|DISCL_FOREGROUND);  assert(!FAILED(hr));
    }
    hr = dinput->CreateDevice(GUID_SysMouse,(LPDIRECTINPUTDEVICE*)&di_mouse,0);
    if(di_mouse) {
        hr = di_mouse->SetDataFormat(&c_dfDIMouse);  assert(!FAILED(hr));
        hr = di_mouse->SetCooperativeLevel(WindowHandle,DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);  assert(!FAILED(hr));
    }

    // create RC
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor =
    {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };

    // get a pixel format
    // @todo fullscreen
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    assert(WindowRC);
    wglMakeCurrent(WindowDC,WindowRC);

    // opengl init
    openglInit(WindowWidth, WindowHeight);

    // game init
    SeedRand(rand());

    // @todo: do real init count and update callback setting here
    if (SystemHooks.GameInit)
       (*SystemHooks.GameInit)();
    
    // show the window
    ShowWindow(WindowHandle,sw);

    // message loop
    prevTime = g_server_time;

    MSG msg;
    for(;;)
    {
        if (PeekMessage(&msg, NULL, 0, 0,PM_REMOVE))
        {
            if (msg.message == WM_QUIT) break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
        {
            AppIdle();
        }
    }

    // clean up

    if(di_gamepad) {
        di_gamepad->Unacquire();
        di_gamepad->Release();
    }
    if(di_mouse) {
        di_mouse->Unacquire();
        di_mouse->Release();
    }
    if(dinput) {
        dinput->Release();
    }
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);

    // show off the counters
    PrintCounters();

    EngineShutdown();

    return msg.wParam;
}

#define GL_TEXTURE0_ARB                   0x84C0

typedef void (APIENTRY *PFNGLACTIVETEXTUREARBPROC)(GLenum);
typedef void (APIENTRY *PFNGLCLIENTACTIVETEXTUREARBPROC)(GLenum);
typedef void (APIENTRY *PFNGLMULTITEXCOORD2FARBPROC) (GLenum target, GLfloat s, GLfloat t);

PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;

#define GPA(x)   do_gpa((void (**)(void)) &x, #x)

static void do_gpa(void (**func)(void), char *str)
{
   *func = (void (*)(void)) wglGetProcAddress(str);
   assert(func != NULL);
}

void glClientActiveTexture(int n)
{
   glClientActiveTextureARB(GL_TEXTURE0_ARB+n);
}

void glActiveTexture(int n)
{
   glActiveTextureARB(GL_TEXTURE0_ARB+n);
}

void glMultiTexCoord2f(int n, GLfloat s, GLfloat t)
{
   glMultiTexCoord2fARB(GL_TEXTURE0_ARB+n, s, t);
}

void openglInit(int width, int height)
{
   glClearColor(0.5f, 0.5f, 0.5f, 0.5f);
   glClearDepth(1.0);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glShadeModel(GL_SMOOTH);

   glAlphaFunc(GL_GEQUAL, 0.1f);
   glDisable(GL_ALPHA_TEST);

   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glEnable(GL_BLEND);

   glEnable(GL_DEPTH_TEST);
   glDepthMask(TRUE);
   glDepthFunc(GL_LEQUAL);

   glEnable(GL_TEXTURE_2D);

   glFrontFace(GL_CCW);
   glEnable(GL_CULL_FACE);

   GPA(glActiveTextureARB);
   GPA(glClientActiveTextureARB);
   GPA(glMultiTexCoord2fARB);
}

#define PI     3.141592

void computeRelativeTranslation(vec3f *out, vec3f *src, vec3f *angle)
{
  GLdouble matrix[16];
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glRotatef(-angle->y,  0.0f, 1.0f, 0.0f);
  glRotatef(-angle->x,  1.0f, 0.0f, 0.0f);
  glRotatef(-angle->z,  0.0f, 0.0f, 1.0f);
  glGetDoublev(GL_MODELVIEW_MATRIX, matrix);
  glPopMatrix();

  out->x = matrix[0]*src->x + matrix[1]*src->y + matrix[2]*src->z;
  out->y = matrix[4]*src->x + matrix[5]*src->y + matrix[6]*src->z;
  out->z = matrix[8]*src->x + matrix[9]*src->y + matrix[10]*src->z;
}

void setupCamera(camera &Camera)
{
    SetCameraAspectRatios(Camera,
                          StandardTVPhysicalAspectRatio,
                          (float)ScreenWidth, (float)ScreenHeight,
                          (float)WindowWidth, (float)WindowHeight);
    BuildCameraMatrices(Camera);    

    // Set up the projection
    glMatrixMode(GL_PROJECTION);
    glViewport(0, 0, WindowWidth, WindowHeight);
    glLoadMatrixf((float *)Camera.Projection4x4);

    // Set up the camera
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf((float *)Camera.View4x4);
}


