/*
  Indy Game Jam

*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include "globals.h"

double GetTime( void )
{
    static __int64 Frequency;
    static __int64 StartTime;
    static int Init = 1;
    __int64 Timer;

    if(Init)
    {
        QueryPerformanceFrequency((LARGE_INTEGER*)&Frequency);
        QueryPerformanceCounter((LARGE_INTEGER*)&StartTime);
        Init = 0;
        return 0.0;
    }
    else
    {
        QueryPerformanceCounter((LARGE_INTEGER*)&Timer);
        return ((double)(Timer-StartTime)/(double)(Frequency));
    }
}


struct timing_bar {
    enum { NUM_SAMPLES = 64 };
    char *Label;
    double FrameTotal;
    int FrameIdx;
    int ValidData;
    double LastTime;
    double CurrentTime;
    int y_idx;
    double RecentFrames[NUM_SAMPLES];
//    DrawTimingBar( void );
} FPSBar = {
    "FPS",  // Label
    0.0,    // FrameTotal
};    

void DrawSegmentedBar( float x, float y, float xbar, float ybar,
                       float TotalLength, float SegmentLength,
                       float aaColors[2][3], int ShadeCount = 0 )
{
    float TotalDivSegment = TotalLength / SegmentLength;
    float TotalModSegmentScale =
        float(fmod(TotalLength,SegmentLength)) / SegmentLength;
    int unsigned NumberOfSegments = (int unsigned)TotalDivSegment;
    int unsigned ColorIndex = 0;
    float z = -0.2f;

    for(int unsigned i = 0;i < NumberOfSegments;i++)
    {
        glBegin(GL_QUADS);
        float s = 1.f;
        if(ShadeCount && ((i / ShadeCount) & 1))
        {
            s = 0.5f;
        }
        glColor3f(s*aaColors[ColorIndex][0],
                  s*aaColors[ColorIndex][1],
                  s*aaColors[ColorIndex][2]);
        glVertex3f(x     ,y     ,z);
        glVertex3f(x+xbar,y     ,z);
        glVertex3f(x+xbar,y+ybar,z);
        glVertex3f(x     ,y+ybar,z);
        glEnd();
        ColorIndex = !ColorIndex;
        x += xbar;
    }

    xbar *= TotalModSegmentScale;
    glBegin(GL_QUADS);
    float s = 1.f;
    if(ShadeCount && ((i / ShadeCount) & 1))
    {
        s = 0.5f;
    }
    glColor3f(s*aaColors[ColorIndex][0],
              s*aaColors[ColorIndex][1],
              s*aaColors[ColorIndex][2]);
    glVertex3f(x     ,y     ,z);
    glVertex3f(x+xbar,y     ,z);
    glVertex3f(x+xbar,y+ybar,z);
    glVertex3f(x     ,y+ybar,z);
    glEnd();
}

void DrawTimingBars( void )
{
   static double LastTime = 0.0;
   static double CurrentTime;

   int const NUM_SAMPLES = 64;
   static double RecentFrames[NUM_SAMPLES], FrameTotal=0.0; // secretly im counting on the array being Zeroed as well
	static int    FrameIdx=0;                                // where in the Frame sample storage array we are
	static int    ValidData=0;                               // is the data valid (i.e. have we looped through it all once?)

   CurrentTime = GetTime();
   double ElapsedTime = 1.0/(CurrentTime - LastTime);
   LastTime = CurrentTime;

	FrameTotal-=RecentFrames[FrameIdx];
	RecentFrames[FrameIdx]=ElapsedTime;
	FrameTotal+=ElapsedTime;
	FrameIdx=(FrameIdx+1)%NUM_SAMPLES;
	if (FrameIdx==0) ValidData=1; // once we have wrapped once, we have valid hyst data

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.f,1.f,0.f,1.f,0.f,10.f);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    // @todo need to glPushAttrib and ensure the correct state
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);

    float const border = 0.01f;
    float const width = 1.0f - 2.0f * border;
    float const height = 0.0125f;
    float const d_background = 0.005f;
    float const FPSAcross = 100.f;

    float w = float(ElapsedTime) / FPSAcross;
    float s = (float)ceil(w);
    float e = width * w;
    float t = width * (float(FrameTotal/NUM_SAMPLES) / FPSAcross);
    float back_seg_length = width / 4.f;
    float front_seg_length = back_seg_length / 5.f;

    static float s_current = s;
    static float s_desired = s;
    static double s_time;
    if(s >= s_current)
    {
        // we need to rescale FPS bar immediately
        s_current = s;
        s_time = GetTime();
    }
    else
    if(s < s_current)
    {
        // we'd like to rescale, is it time?
        if(s_desired == s)
        {
            if((GetTime() - s_time) > 3.0f)
            {
                // we've been stable for a bit
                s_current = s;
            }
        }
        else
        {
            s_time = GetTime();
            s_desired = s;
        }
    }
    
    e /= s_current;
    t /= s_current;
    back_seg_length /= s_current;
    front_seg_length /= s_current;

    float x = border, y = border, xbar = front_seg_length, ybar = height;
    float aaColors[2][2][3] =
    {
        { { 1.0f, 0.0f, 0.0f }, { 0.0f, 1.0f, 0.0f } },
        { { 1.0f, 1.0f, 1.0f }, { 0.0f, 0.0f, 1.0f } }
    };

    if(0)
    {
        float x = 0, y = 0;
        float xxbar = xbar * 5.f, yybar = ybar+2*d_background;
        glBegin(GL_QUADS);
        glColor3f(0,0,0);

        glVertex3f(x,y,0);
        glVertex3f(x+xxbar,y,0);
        glVertex3f(x+xxbar,y+yybar,0);
        glVertex3f(x,y+yybar,0);

        glVertex3f(0,+.5,0);
        glVertex3f(0+.1,+.5,0);
        glVertex3f(0+.1,+.5+.1,0);
        glVertex3f(0,+.5+.1,0);
            
        glEnd();
    }
    
    DrawSegmentedBar(x,y-d_background,
                     xbar*5.f,ybar+2.f*d_background,
                     e,back_seg_length,
                     aaColors[0],4);

    if(ValidData)
    {
        DrawSegmentedBar(x,y,
                         xbar,ybar,
                         t,front_seg_length,
                         aaColors[1],20);
    }

//    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
}

float GenerateUnitRandomReal( void )
{
    float Random = ((float)rand())/((float)RAND_MAX);
    assert((Random >= 0) && (Random <= 1));
    return Random;
}

// Rand250, a fast, and reasonably good random number generator
unsigned int  RandIndex1 = 0;
unsigned int  RandIndex2 = 103;
unsigned int  RandTable[256];

int Rand250(void)
{
   int ret = (RandTable[RandIndex1] ^= RandTable[RandIndex2]);
   RandIndex1++; if (RandIndex1 >= 249) RandIndex1 = 0;
   RandIndex2++; if (RandIndex2 >= 249) RandIndex2 = 0;
   return ret;
}

void SeedRand(int seed)
{
   // init our fat rand-table using another rand generator.
   for (int i=0; i < 256; ++i)
   {
      int rand_lo = ((seed = seed * 1103515245 + 12345) >> 16) & 0x7fff;
      int rand_hi = ((seed = seed * 1103515245 + 12345) >> 16) & 0x7fff;
      RandTable[i] = rand_lo * rand_hi;
   }

   RandIndex1 = 0;
   RandIndex2 = 103;
}

// return random number in range [low, high)
int random (int low, int high)
{
   assert (high > low);
   return low + (Rand250() % (high - low));
}

// return random float in range [0.0, 1.0]
#define TWO31 (float(32768.0) * float(32768.0))

float frandom()
{
   return (float(Rand250()) / TWO31);
}
