#ifndef INC_VEC3f_H
#define INC_VEC3f_H

struct vec3f
{
	float x,y,z;
	
	vec3f() : x(0), y(0), z(0) {}
	vec3f(float _x, float _y, float _z) : x(_x), y(_y), z(_z) {}

	vec3f	operator+(const vec3f& v) const { return vec3f(x + v.x, y + v.y, z + v.z); }
	vec3f	operator-(const vec3f& v) const { return vec3f(x - v.x, y - v.y, z - v.z); }
	vec3f	operator*(float f) const { return vec3f(x*f, y*f, z*f); }
	float	operator*(const vec3f& v) const { return x * v.x + y * v.y + z * v.z; }
	void	operator+=(const vec3f& v) { x += v.x; y += v.y; z += v.z; }
	void	operator-=(const vec3f& v) { x -= v.x; y -= v.y; z -= v.z; }
	void	operator*=(float f) { x *= f; y *= f; z *= f; }

	operator const float*() const { return &x; }
};

extern void vec3f_add(vec3f *d, vec3f *v0, vec3f *v1);
extern void vec3f_sub(vec3f *d, vec3f *v0, vec3f *v1);
extern void vec3f_scale(vec3f *d, vec3f *v, float scale);
extern void vec3f_scaleeq(vec3f *vec, float scale);

extern void vec3f_add_scale(vec3f *dest, vec3f *v0, vec3f *v1, float scale);
extern void vec3f_addeq_scale(vec3f *dest, vec3f *src, float scale);

// weight=0 -> a, weight=1 -> b
extern void vec3f_interpolate(vec3f *dest, vec3f *a, vec3f *b, float weight);

extern void vec3f_cross(vec3f *cross, vec3f *v0, vec3f *v1);
extern float vec3f_dot(vec3f *v0, vec3f*v1);
extern float vec3f_mag2(vec3f *v);
extern float vec3f_mag(vec3f *v);
extern float vec3f_one_over_mag(vec3f *v);
extern void vec3f_norm(vec3f *dest, vec3f *src);
extern void vec3f_normeq(vec3f *v);

extern void mat4_mul(float out[4][4], float mat1[4][4], float mat2[4][4]);

#endif
